#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""Do something for installing the landisk-remote-support."""

import os

from iobase import xmldata
from iobase import cmdexec
from iobase.exception import (IoBaseException,)
from ionas.request import (Request, Entry)
from product_conf import (BIN_DIR, PACKAGES_DIR, XML_CONF_DIR,)
import fal_msginfo
from fal_info import (FAL_FEATURES_PATH, Results, ReqTypes)

import manualctrl
from _service import append_service


ENCODING = "UTF-8"
CATEGORYLIST_PATH = "categorylist"
CATEGORY_NAME = "category"
CATEGORY_KEY_PATH = "code"
CATEGORY_FIND_XPATH_FMT = """category[code="%s"]"""

MESSAGEINFOLIST_PATH = "messageinfolist"
MESSAGEINFO_NAME = "messageinfo"
MESSAGEINFO_KEY_PATH = "code"
MESSAGEINFO_FIND_XPATH_FMT = """messageinfo[code="%s"]"""

MESSAGES_FILE = os.path.join(XML_CONF_DIR, "messages.xml")
PACKAGE_MESSAGES_FILE = os.path.join(PACKAGES_DIR, "landisk-remote-support",
                                        "messages.xml")
BUILD_MSGINFO_CMD = os.path.join(BIN_DIR, "build_msginfo")

PYC_SUFFIX = ".pyc"
PY_SUFFIX = ".py"

FEATURES_REQ_XML = \
    "<features><functionlist>" + \
    "<function>remote_support</function>"+ \
    "</functionlist></features>"


class RequestError(IoBaseException):
    """Exception Class for request error."""

    pass


def __send_request(target, type, body="", err_msg=""):
    """Send request entry to "target"."""

    entry = Entry(type, body)
    result_entry = Request(target).request(entry).get_entry()
    if (result_entry.get_result() != Results.OK and
        result_entry.get_result() != Results.FILE_NOT_FOUND):
        raise RequestError("%s - (%s)" % (err_msg, result_entry.get_result()))

    try:
        return result_entry.get()
    except:
        return ""


def __merge_list(new_messages_xml, messages_xml, list_xpath,
        item_name, key_in_item_xpath, find_xpath_fmt):
    """Merge a list in "new_mssages_xml" into "messages_xml"."""

    new_list_node = new_messages_xml.get_node(list_xpath)
    list_node = messages_xml.get_node(list_xpath)

    for item_node in new_list_node.get_node_list(item_name):
        key = item_node.get_value(key_in_item_xpath)
        try:
            cur_node = list_node.get_node(find_xpath_fmt % (key,))
        except:
            list_node.append_child(item_node.get_clone())


def __merge_categorylist(new_messages_xml, messages_xml):
    """Merge "categorylist" tree in "new_mssages_xml" into "messages_xml"."""

    __merge_list(new_messages_xml, messages_xml, CATEGORYLIST_PATH,
        CATEGORY_NAME, CATEGORY_KEY_PATH, CATEGORY_FIND_XPATH_FMT)


def __merge_messageinfolist(new_messages_xml, messages_xml):
    """Merge "messageinfolist" tree in "new_mssages_xml"
       into "messages_xml".
    """

    __merge_list(new_messages_xml, messages_xml, MESSAGEINFOLIST_PATH,
        MESSAGEINFO_NAME, MESSAGEINFO_KEY_PATH, MESSAGEINFO_FIND_XPATH_FMT)


def __update_msginfo(messages_file):
    """Update a "fal_msginfo" for "message_file"."""

    fal_msginfo_path = fal_msginfo.__file__

    (base, suffix) = os.path.splitext(fal_msginfo_path)
    if suffix == PYC_SUFFIX:
        fal_msginfo_path = base + PY_SUFFIX

    cmd_args = [BUILD_MSGINFO_CMD, messages_file, fal_msginfo_path]
    cmdexec.CmdExec(cmd_args).execute()


def update_messages(messages_file, package_messages_file):
    """Update "messages_file" by "package_messages_file"."""

    messages_xml = xmldata.XmlConf(messages_file, encoding=ENCODING)
    package_messages_xml = xmldata.XmlConf(package_messages_file,
                                encoding=ENCODING)

    __merge_categorylist(package_messages_xml, messages_xml)
    __merge_messageinfolist(package_messages_xml, messages_xml)
    tmp_path = messages_file + "." + str(os.getpid())
    messages_xml.save(tmp_path)
    os.rename(tmp_path, messages_file)

    __update_msginfo(messages_file)


def add_package_feature():
    """Add feature info for the package."""

    try:
        __send_request(FAL_FEATURES_PATH, ReqTypes.ADD, FEATURES_REQ_XML)
    except Exception:
        # ignore error
        pass


#
# Main
#

add_package_feature()
update_messages(MESSAGES_FILE, PACKAGE_MESSAGES_FILE)
append_service('cbot-agent')

manualctrl.install()
