"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.App = void 0;
const child_process_1 = require("child_process");
const config = require("config");
const electron = require("electron");
const events_1 = require("events");
const fs = require("fs");
const i18n = require("i18n");
const i18n_1 = require("i18n");
const nodeUuid = require("node-uuid");
const os = require("os");
const utils_1 = require("./utils");
const manager_1 = require("./controlSession/manager");
const desktop_1 = require("./gui/desktop");
const http_1 = require("./servers/http/http");
const localStorageCleaner_1 = require("./localStorageCleaner");
const httpClient_1 = require("./httpClient");
class App {
    constructor() {
        this.profile = null;
        this.proxy = null;
        this.version = null;
        this.hardware = null;
        this.lang = null;
        this.browsers = {};
        this.connectors = {};
        this.detectTerminateProcessCount = 0;
        httpClient_1.default.init(() => {
            return this.getProxy();
        });
    }
    /**
     * インスタンス初期化
     *
     * @returns App
     */
    init() {
        return __awaiter(this, void 0, void 0, function* () {
            // バージョン取得
            console.log('version:', yield this.getVersion());
            // ハードウェア取得
            console.log('hardware:', yield this.getHardware());
            // プロキシ情報取得
            console.log('proxy:', yield this.getProxy());
            this.controlSessionManager = new manager_1.ControlSessionManager(this);
            // エージェントファイルが存在しない場合に、未登録エージェントを作成する
            yield this.getAgent();
            this.eventEmitter = new events_1.EventEmitter;
            // Browser起動通知
            this.eventEmitter.on('start_browser', (browser, ack) => __awaiter(this, void 0, void 0, function* () {
                console.log('start browser notification');
                this.browsers[browser.getUuid()] = browser;
                this.desktopNotification(i18n_1.__('ブラウザを起動しました。'));
                yield this.httpServer.socketIoServer.broadcastBoots();
                try {
                    yield this.eventNotification('start_virtual_browser', {
                        'execute_id': browser.getUuid()
                    });
                }
                catch (e) {
                    console.log(e.message);
                }
                ack();
            }));
            // Browser停止通知
            this.eventEmitter.on('stop_browser', (browser, ack) => __awaiter(this, void 0, void 0, function* () {
                console.log('stop browser notification');
                delete this.browsers[browser.getUuid()];
                this.desktopNotification(i18n_1.__('ブラウザを停止しました。'));
                yield this.httpServer.socketIoServer.broadcastBoots();
                try {
                    yield this.eventNotification('stop_virtual_browser', {
                        'execute_id': browser.getUuid()
                    });
                }
                catch (e) {
                    console.log(e.message);
                }
                ack();
                yield this.processKillIfAllDisconnected();
            }));
            // Connector起動通知
            this.eventEmitter.on('start_connector', (connector, ack) => __awaiter(this, void 0, void 0, function* () {
                console.log('start connector notification');
                this.connectors[connector.getId()] = connector;
                this.desktopNotification(i18n_1.__('コネクタを起動しました。'));
                yield this.httpServer.socketIoServer.broadcastBoots();
                try {
                    yield this.eventNotification('start_virtual_connector', {
                        'execute_id': connector.getId()
                    });
                }
                catch (e) {
                    console.log(e.message);
                }
                ack();
            }));
            // Connector停止通知
            this.eventEmitter.on('stop_connector', (connector, ack) => __awaiter(this, void 0, void 0, function* () {
                console.log('stop connector notification');
                delete this.connectors[connector.getId()];
                this.desktopNotification(i18n_1.__('コネクタを停止しました。'));
                yield this.httpServer.socketIoServer.broadcastBoots();
                try {
                    yield this.eventNotification('stop_virtual_connector', {
                        'execute_id': connector.getId()
                    });
                }
                catch (e) {
                    console.log(e.message);
                }
                ack();
                yield this.processKillIfAllDisconnected();
            }));
            // devicePixelRatio更新
            this.eventEmitter.on('update_device_pixel_ratio', (devicePixelRatio) => __awaiter(this, void 0, void 0, function* () {
                const profile = yield this.getProfile();
                if (profile.device_pixel_ratio != devicePixelRatio) {
                    profile.device_pixel_ratio = devicePixelRatio;
                    try {
                        const result = yield this.notifyProfile(profile, yield this.getAgent());
                        if (result.code == 200 || result.code == 401) {
                            this.setProfile(profile);
                        }
                        else {
                            throw result;
                        }
                    }
                    catch (e) {
                        console.log(e.message);
                    }
                }
            }));
            // 多言語対応
            i18n.configure({
                locales: ['ja', 'en'],
                defaultLocale: (yield this.getLang()).value,
                directory: `${__dirname}/../resources/lang`
            });
            return this;
        });
    }
    /**
     * Agent開始
     */
    start() {
        var _a, _b;
        return __awaiter(this, void 0, void 0, function* () {
            try {
                // SIGINTシグナルでプロセス終了
                let sigint = false;
                process.on("SIGINT", () => __awaiter(this, void 0, void 0, function* () {
                    if (!sigint) {
                        console.log('\nReceived SIGINT. Close server.');
                        sigint = true;
                        yield this.exit();
                    }
                }));
                // SIGTERMシグナルでプロセス終了
                process.on("SIGTERM", () => __awaiter(this, void 0, void 0, function* () {
                    console.log('\nReceived SIGTERM. Close server.');
                    yield this.exit();
                }));
                process.on('exit', () => {
                    console.log('exit now');
                });
                // プロセス終了するべきか監視する(不正にプロセスが残留することを防ぐ)
                this.watchTerminateProcess();
                // ストレージクリーナーを起動
                this.storageCleaner();
                // GUI準備
                if (electron.app) {
                    if (!electron.app.requestSingleInstanceLock()) {
                        yield this.exit();
                        return;
                    }
                    this.desktop = new desktop_1.Desktop(this);
                }
                // HTTPサーバ起動
                this.httpServer = new http_1.HTTPServer(this);
                try {
                    yield this.httpServer.serve();
                }
                catch (e) {
                    // HTTPサーバ起動失敗の場合はプロセス終了する
                    yield ((_a = this.desktop) === null || _a === void 0 ? void 0 : _a.exit('ERROR', e.message));
                    return;
                }
                // API中継クライアントのライフサイクル開始
                this.controlSessionManager.lifeCycle();
                // GUI表示
                yield ((_b = this.desktop) === null || _b === void 0 ? void 0 : _b.start());
                console.log('Agent started.');
            }
            catch (e) {
                console.log(e.message);
            }
        });
    }
    /**
     * プロセス終了
     */
    exit() {
        return __awaiter(this, void 0, void 0, function* () {
            console.log('request exit');
            this.controlSessionManager.end();
            yield this.disconnectBrowsers();
            clearTimeout(this.detectTerminateProcessTimer);
            clearTimeout(this.storageCleanerTimer);
            yield this.cleanStorage();
            if (electron.app) {
                electron.app.exit();
            }
            else {
                process.exit();
            }
        });
    }
    /**
     * すべての接続が切断していればプロセス終了
     * (ワンタイムエージェントの場合のみ)
     */
    processKillIfAllDisconnected() {
        return __awaiter(this, void 0, void 0, function* () {
            if (config.feature.one_time_connectable) {
                if (yield this.isAllDisconnected()) {
                    console.log('Quit the agent.');
                    setTimeout(() => __awaiter(this, void 0, void 0, function* () {
                        yield this.httpServer.app.exit();
                    }));
                }
            }
        });
    }
    /**
     * すべての接続が切断しているか
     */
    isAllDisconnected() {
        return __awaiter(this, void 0, void 0, function* () {
            const status = yield this.getStatus();
            return Object.keys(this.browsers).length == 0 && Object.keys(this.connectors).length == 0 && (status.code != 201);
        });
    }
    /**
     * プロセス終了するべきか監視
     */
    watchTerminateProcess() {
        // ワンタイムエージェントの場合、3回(5分間隔)全切断を検知したらプロセス終了する
        if (config.feature.one_time_connectable) {
            this.detectTerminateProcessTimer = setTimeout(() => __awaiter(this, void 0, void 0, function* () {
                if (yield this.isAllDisconnected()) {
                    this.detectTerminateProcessCount++;
                    if (this.detectTerminateProcessCount >= 3) {
                        console.log('detect terminate process');
                        yield this.exit();
                    }
                }
                else {
                    this.detectTerminateProcessCount = 0;
                }
                this.watchTerminateProcess();
            }), 300000);
        }
    }
    /**
     * ストレージクリーナー
     */
    storageCleaner() {
        return __awaiter(this, void 0, void 0, function* () {
            const storageCleaner = new localStorageCleaner_1.LocalStorageCleaner([
                {
                    path: config.storage.path,
                    timeout: 7200000 // 2時間
                }
            ]);
            storageCleaner.exec();
            this.storageCleanerTimer = setTimeout(() => {
                this.storageCleaner();
            }, 300000);
        });
    }
    /**
     * 終了時ストレージ削除
     */
    cleanStorage() {
        return __awaiter(this, void 0, void 0, function* () {
            yield fs.promises.rmdir(config.storage.path, {
                recursive: true
            });
        });
    }
    /**
     * ポート番号取得
     */
    getPort() {
        return this.httpServer.getPort();
    }
    /**
     * エージェント接続
     */
    connectAgent() {
        return __awaiter(this, void 0, void 0, function* () {
            let status = yield this.getStatus();
            if (status.code != 201) {
                status = yield new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
                    this.eventEmitter.once('updateStatus', (status) => {
                        resolve(status);
                    });
                    yield this.updateStatus(101);
                }));
            }
            return status;
        });
    }
    /**
     * エージェント切断
     */
    disconnectAgent(processKillIfAllDisconnected = true) {
        return __awaiter(this, void 0, void 0, function* () {
            const status = yield new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
                this.eventEmitter.once('updateStatus', (status) => {
                    resolve(status);
                });
                yield this.updateStatus(0);
                yield this.controlSessionManager.disconnect(processKillIfAllDisconnected);
            }));
            return status;
        });
    }
    /**
     * エージェント情報リセット
     */
    resetAgent() {
        this.controlSessionManager.resetAgent();
    }
    /**
     * エージェント情報取得
     */
    getAgent() {
        return this.controlSessionManager.getAgent();
    }
    /**
     * バージョン情報取得
     */
    getVersionInfo() {
        return this.controlSessionManager.getVersionInfo();
    }
    /**
     * バージョン情報更新
     */
    updateVersionInfo(versionInfo) {
        return this.controlSessionManager.updateVersionInfo(versionInfo);
    }
    /**
     * ステータス取得
     */
    getStatus() {
        return this.controlSessionManager.getStatus();
    }
    /**
     * ステータス更新
     */
    updateStatus(code) {
        return this.controlSessionManager.updateStatus(code);
    }
    /**
     * ステータス削除
     */
    deleteStatus() {
        return this.controlSessionManager.deleteStatus();
    }
    /**
     * ステータスをブロードキャスト
     */
    broadcastStatus() {
        return this.httpServer.socketIoServer.broadcastStatus();
    }
    /**
     * 起動数をブロードキャスト
     */
    broadcastBoots() {
        return this.httpServer.socketIoServer.broadcastBoots();
    }
    /**
     * プロフィール設定
     */
    setProfile(profile) {
        profile = this.correctProfile(profile);
        utils_1.writeFile(`${config.profile.fileName}`, JSON.stringify(profile), config.profile.encoding);
        this.profile = profile;
        console.log(this.profile);
        return this.profile;
    }
    /**
     * プロフィール取得
     */
    getProfile() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.profile) {
                return Object.assign({}, this.profile);
            }
            else {
                try {
                    const data = yield fs.promises.readFile(`${config.profile.fileName}`, {
                        encoding: config.profile.encoding
                    });
                    let profile = JSON.parse(data.toString());
                    profile = this.correctProfile(profile);
                    this.profile = profile;
                    return profile;
                }
                catch (e) {
                    let profile = {
                        uuid: nodeUuid.v1(),
                        name: `${os.userInfo().username}'s PC`,
                        device_pixel_ratio: null,
                        max_values: {
                            1: config.browser.default_max_boot,
                            2: config.connector.default_max_boot
                        },
                        allow_file_transfer: {
                            inbound: true,
                            outbound: true
                        },
                        registered: false
                    };
                    profile = this.correctProfile(profile);
                    return profile;
                }
            }
        });
    }
    /**
     * プロフィール補正
     */
    correctProfile(profile) {
        if (profile.registered === undefined) {
            profile.registered = true;
        }
        if (!profile.allow_file_transfer) {
            profile.allow_file_transfer = {
                inbound: true,
                outbound: true
            };
        }
        if (profile.max_values[1] === undefined) {
            profile.max_values[1] = config.browser.default_max_boot;
        }
        if (profile.max_values[2] === undefined) {
            profile.max_values[2] = config.connector.default_max_boot;
        }
        if (!config.feature.browser) {
            profile.max_values[1] = 0;
        }
        if (!config.feature.connector) {
            profile.max_values[2] = 0;
        }
        return profile;
    }
    /**
     * プロフィール削除
     */
    deleteProfile() {
        if (fs.existsSync(config.profile.fileName)) {
            fs.unlinkSync(config.profile.fileName);
        }
        this.profile = null;
    }
    /**
     * プロフィール通知
     */
    notifyProfile(profile, agent) {
        var _a, _b, _c;
        return __awaiter(this, void 0, void 0, function* () {
            if (agent === null || agent === void 0 ? void 0 : agent.registered) {
                const payload = yield this.getFullProfile(profile);
                const profileHash = utils_1.md5Hash(JSON.stringify([payload, agent]));
                if (this.NotifiedProfileHash != profileHash) {
                    const lang = (yield this.getLang()).value;
                    try {
                        yield httpClient_1.default.put(`${config.appServer.endpoint}:${config.appServer.port}/api/agents/${agent.id}`, payload, {
                            headers: {
                                'content-type': 'application/json',
                                'content-language': lang,
                                'secret': agent.secret
                            }
                        });
                        this.NotifiedProfileHash = profileHash;
                        console.log('Notify profile.');
                    }
                    catch (e) {
                        if ((_b = (_a = e.response) === null || _a === void 0 ? void 0 : _a.data) === null || _b === void 0 ? void 0 : _b.code) {
                            return {
                                code: e.response.data.code,
                                message: e.response.data.message
                            };
                        }
                        else {
                            if ((_c = e.response) === null || _c === void 0 ? void 0 : _c.status) {
                                return {
                                    code: e.response.status,
                                    message: e.response.statusText
                                };
                            }
                            else {
                                return {
                                    code: 500,
                                    message: e.message
                                };
                            }
                        }
                    }
                }
            }
            return {
                code: 200
            };
        });
    }
    /**
     * フルプロフィール取得
     */
    getFullProfile(profile = null) {
        return __awaiter(this, void 0, void 0, function* () {
            const version = yield this.getVersion();
            const hardware = yield this.getHardware();
            if (!profile) {
                profile = yield this.getProfile();
            }
            const fullProfile = {
                name: profile.name,
                platform: hardware.platform,
                platform_version: hardware.platform_version,
                agent_version: version.version,
                agent_edition: version.edition,
                serial_number: hardware.serial_number ? hardware.serial_number : profile.uuid,
                detail: hardware.detail,
                device_pixel_ratio: profile.device_pixel_ratio,
                max_values: [
                    {
                        value: profile.max_values[1],
                        type: 1
                    },
                    {
                        value: profile.max_values[2],
                        type: 2
                    }
                ],
                architecture: this.getArchitecture(),
                allow_file_transfer: {
                    inbound: profile.allow_file_transfer.inbound,
                    outbound: profile.allow_file_transfer.outbound
                }
            };
            return fullProfile;
        });
    }
    /**
     * プロキシ更新
     */
    setProxy(proxy) {
        utils_1.writeFile(`${config.proxy.fileName}`, JSON.stringify(proxy), config.profile.encoding);
        this.proxy = proxy;
        console.log(this.proxy);
        return this.proxy;
    }
    /**
     * プロキシ取得
     */
    getProxy() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.proxy) {
                return this.proxy;
            }
            else {
                try {
                    let data;
                    if (config.proxy.command) {
                        const options = {
                            env: {
                                'HOME': process.env.HOME,
                                'NODE_ENV': process.env.NODE_ENV
                            },
                        };
                        data = yield new Promise((resolve, reject) => {
                            child_process_1.exec(config.proxy.command, options, (err, stdout) => {
                                if (err) {
                                    reject(err);
                                }
                                else {
                                    resolve(stdout);
                                }
                            });
                        });
                    }
                    else {
                        data = (yield fs.promises.readFile(`${config.proxy.fileName}`, {
                            encoding: config.proxy.encoding
                        })).toString();
                    }
                    this.proxy = JSON.parse(data);
                    return this.proxy;
                }
                catch (e) {
                    return {
                        enable: false,
                        url: null,
                        pac: null
                    };
                }
            }
        });
    }
    /**
     * プロキシ削除
     */
    deleteProxy() {
        if (fs.existsSync(config.proxy.fileName)) {
            fs.unlinkSync(config.proxy.fileName);
        }
        this.proxy = null;
    }
    /**
     * プロキシ設定をリロード
     */
    reloadProxy() {
        return __awaiter(this, void 0, void 0, function* () {
            this.proxy = null;
            return yield this.getProxy();
        });
    }
    /**
     * 言語更新
     */
    setLang(lang) {
        var _a, _b;
        utils_1.writeFile(`${config.lang.fileName}`, JSON.stringify(lang), config.profile.encoding);
        this.lang = {
            value: lang,
            defined: true
        };
        i18n.setLocale(this.lang.value);
        (_a = this.desktop) === null || _a === void 0 ? void 0 : _a.initApplicationMenu();
        (_b = this.desktop) === null || _b === void 0 ? void 0 : _b.initContextMenu();
        console.log(this.lang);
        return this.lang;
    }
    /**
     * 言語取得
     */
    getLang() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.lang) {
                return this.lang;
            }
            else {
                try {
                    const data = yield fs.promises.readFile(`${config.lang.fileName}`, {
                        encoding: config.lang.encoding
                    });
                    this.lang = {
                        value: JSON.parse(data.toString()),
                        defined: true
                    };
                    return this.lang;
                }
                catch (e) {
                    return {
                        value: 'en',
                        defined: false
                    };
                }
            }
        });
    }
    /**
     * 言語削除
     */
    deleteLang() {
        if (fs.existsSync(config.lang.fileName)) {
            fs.unlinkSync(config.lang.fileName);
        }
        this.lang = null;
    }
    /**
     * バージョン取得
     */
    getVersion() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.version) {
                return Object.assign({}, this.version);
            }
            else {
                const data = yield fs.promises.readFile(`${__dirname}/../resources/version`, {
                    encoding: config.proxy.encoding
                });
                this.version = JSON.parse(data.toString());
                return this.version;
            }
        });
    }
    /**
     * ハードウェア情報のクリア
     */
    clearHardware() {
        this.hardware = null;
    }
    /**
     * ハードウェア情報取得
     */
    getHardware() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.hardware) {
                return Object.assign({}, this.hardware);
            }
            else {
                const hardware = {
                    platform: null,
                    platform_version: null,
                    serial_number: null,
                    detail: null
                };
                try {
                    const data = yield fs.promises.readFile(`${config.hardware.fileName}`, {
                        encoding: config.hardware.encoding
                    });
                    Object.assign(hardware, JSON.parse(data.toString()));
                }
                catch (e) { }
                if (hardware.platform === null) {
                    hardware.platform = this.getPlatform();
                }
                if (hardware.platform_version === null) {
                    hardware.platform_version = this.getPlatformVersion();
                }
                this.hardware = hardware;
                return this.hardware;
            }
        });
    }
    /**
     * プラットフォーム取得
     */
    getPlatform() {
        return os.platform();
    }
    /**
     * プラットフォームバージョン取得
     */
    getPlatformVersion() {
        return os.version();
    }
    /**
     * アーキテクチャ取得
     */
    getArchitecture() {
        return os.arch();
    }
    /**
     * ブラウザリスト取得
     */
    getBrowsers() {
        return this.browsers;
    }
    /**
     * コネクタリスト取得
     */
    getConnectors() {
        return this.connectors;
    }
    /**
     * 全ブラウザ切断
     */
    disconnectBrowsers() {
        return __awaiter(this, void 0, void 0, function* () {
            // 全ブラウザのソケットを切断
            const promises = [];
            for (const uuid in this.browsers) {
                const promise = this.browsers[uuid].stop();
                promises.push(promise);
            }
            yield Promise.all(promises);
        });
    }
    /**
     * イベント通知
     */
    eventNotification(event, args) {
        return __awaiter(this, void 0, void 0, function* () {
            console.log(`eventNotification[${event}]`);
            const eventFileName = `${config.events.folderName}/${event}`;
            if (fs.existsSync(eventFileName)) {
                const data = yield fs.promises.readFile(eventFileName, {
                    encoding: config.events.encoding
                });
                let command = data.toString();
                if (args) {
                    // 引数内のプレースホルダーのサニタイズ
                    const arg_keys = Object.keys(args);
                    arg_keys.forEach(key => {
                        arg_keys.forEach(_key => {
                            args[key] = args[key].toString().replace(`{${_key}}`, '');
                        });
                    });
                    Object.keys(args).forEach(key => {
                        command = command.replace(`{${key}}`, JSON.stringify(args[key]));
                    });
                }
                try {
                    yield new Promise((resolve, reject) => {
                        child_process_1.exec(command, (err, stdout, stderr) => {
                            if (err) {
                                console.log(`stderr: ${stderr}`);
                                reject({
                                    code: 500,
                                    message: 'Command failed'
                                });
                            }
                            else {
                                console.log(`stdout: ${stdout}`);
                                resolve();
                            }
                        });
                    });
                }
                catch (e) {
                    throw e;
                }
            }
            else {
                throw {
                    code: 404,
                    message: 'Event not found'
                };
            }
        });
    }
    /**
     * desktop通知
     */
    desktopNotification(message) {
        var _a;
        (_a = this.desktop) === null || _a === void 0 ? void 0 : _a.notification(message);
    }
    /**
     * 外部リンク
     */
    externalLink(url) {
        var _a;
        return (_a = this.desktop) === null || _a === void 0 ? void 0 : _a.externalLink(url);
    }
}
exports.App = App;
