"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ControlSessionClient = void 0;
const config = require("config");
const i18n_1 = require("i18n");
const SocketIOTunnel = require("socket.io-tunnel");
const socket_io_apiRelay_1 = require("../socket-io.apiRelay");
const httpClient_1 = require("../httpClient");
class ControlSessionClient {
    constructor(manager, relayServerUrl) {
        this.manager = manager;
        this.relayServerUrl = relayServerUrl;
    }
    /**
     * 接続シーケンス開始
     */
    start() {
        return new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
            const agent = yield this.manager.getAgent();
            yield this.manager.updateStatus(102);
            // バージョンチェック
            const ret = yield this.validateVersion();
            console.log(ret);
            if (ret.code != 200) {
                console.log(`Failed validate version.: ${JSON.stringify(ret)}`);
                resolve(yield this.manager.updateStatus(400));
                return;
            }
            this.manager.setVersionInfo({
                latestVersion: ret.latest_version,
                latestUrl: ret.latest_url,
                announcement: ret.announcement
            });
            if (!ret.can_connect) {
                console.log(`Connection refused by version verification: ${JSON.stringify(ret)}`);
                resolve(yield this.manager.updateStatus(426));
                return;
            }
            const httpAgent = yield httpClient_1.default.getHttpAgentForSocketIo(`${this.relayServerUrl}/socket.io/socket.io.js`);
            this.socketIoRelay = new socket_io_apiRelay_1.SocketIoRelay(this.relayServerUrl, httpAgent, {
                transportsWsOnly: config.webSocket.transports_ws_only,
                timeout: 15000,
                reconnectionAttempts: 4
            });
            this.socketIoRelay.on('connect_error', (err) => __awaiter(this, void 0, void 0, function* () {
                resolve(yield this.manager.updateStatus(400));
            }));
            if (!agent.registered) {
                try {
                    // 中継サーバへのエージェント登録
                    const { agentId, secret } = yield this.socketIoRelay.register();
                    agent.registered = true;
                    agent.id = agentId;
                    agent.secret = secret;
                }
                catch (e) {
                    this.socketIoRelay.close();
                    resolve(yield this.manager.updateStatus(500));
                    return;
                }
                try {
                    // エージェントID登録完了
                    console.log(this.socketIoRelay.agentId, this.socketIoRelay.agentSecret);
                    const ret = yield this.agentIdRegistered({
                        registered: true,
                        id: this.socketIoRelay.agentId,
                        secret: this.socketIoRelay.agentSecret
                    });
                    if (ret.code != 200) {
                        console.log(`Failed notify agent Id registered.: ${JSON.stringify(ret)}`);
                        this.socketIoRelay.close();
                        resolve(yield this.manager.updateStatus(500));
                        return;
                    }
                }
                catch (e) {
                    console.log(`Failed notify agent Id registered.: ${JSON.stringify(e)}`);
                    this.socketIoRelay.close();
                    resolve(yield this.manager.updateStatus(500));
                    return;
                }
            }
            try {
                // 中継サーバへのエージェント認証
                yield this.socketIoRelay.authenticate(agent.id, agent.secret);
            }
            catch (e) {
                console.log(`Failed Agent authentication.: ${JSON.stringify(e)}`);
                this.socketIoRelay.close();
                resolve(yield this.manager.updateStatus((e.code == 401 || e.code == 409 || e.code == 419 || e.code == 429) ? e.code : 500));
                return;
            }
            // 端末ID,認証キー情報をファイルに保存
            try {
                this.manager.saveAgent(agent);
            }
            catch (e) {
                console.log('Unable to save agent information.');
                this.socketIoRelay.close();
                resolve(yield this.manager.updateStatus(500));
                return;
            }
            try {
                // WS中継Listen要求
                yield this.socketIoRelay.listen('127.0.0.1', this.manager.app.getPort());
            }
            catch (e) {
                console.log(`Failed WS relay Listen request.: ${JSON.stringify(e)}`);
                this.socketIoRelay.close();
                resolve(yield this.manager.updateStatus(500));
                return;
            }
            try {
                // エージェント接続完了
                const ret = yield this.agentConnected(agent);
                if (ret.code != 200) {
                    console.log(`Failed notify agent connected.: ${JSON.stringify(ret)}`);
                    this.socketIoRelay.close();
                    resolve(yield this.manager.updateStatus(500));
                    return;
                }
            }
            catch (e) {
                console.log(`failed notify agent connected.: ${JSON.stringify(e)}`);
                this.socketIoRelay.close();
                resolve(yield this.manager.updateStatus(500));
                return;
            }
            this.socketIoRelay.on('connect', (socket) => {
                console.log('Connected to WS relay server.');
                this.tunnelBuilder = new SocketIOTunnel.TunnelBuilder(socket);
            });
            this.socketIoRelay.on('relay_disconnect', () => __awaiter(this, void 0, void 0, function* () {
                console.log('Disconnect from WS relay server.');
                if ((yield this.manager.getStatus()).code > 0) {
                    resolve(yield this.manager.updateStatus(101));
                }
                else {
                    resolve(yield this.manager.getStatus());
                }
            }));
            const status = yield this.manager.updateStatus(201);
            this.manager.app.desktopNotification(`${i18n_1.__('ステータス')}: ${status.message}`);
        }));
    }
    /**
     * エージェントバージョン検証
     */
    validateVersion() {
        var _a, _b, _c;
        return __awaiter(this, void 0, void 0, function* () {
            const payload = yield this.manager.app.getFullProfile();
            const lang = (yield this.manager.app.getLang()).value;
            try {
                const response = yield httpClient_1.default.post(`${config.appServer.endpoint}:${config.appServer.port}/api/agents/version`, payload, {
                    headers: {
                        'content-type': 'application/json',
                        'content-language': lang
                    }
                });
                return response.data;
            }
            catch (e) {
                if ((_b = (_a = e.response) === null || _a === void 0 ? void 0 : _a.data) === null || _b === void 0 ? void 0 : _b.code) {
                    return {
                        code: e.response.data.code,
                        message: e.response.data.message
                    };
                }
                else {
                    if ((_c = e.response) === null || _c === void 0 ? void 0 : _c.status) {
                        return {
                            code: e.response.status,
                            message: e.response.statusText
                        };
                    }
                    else {
                        return {
                            code: 500,
                            message: e.message
                        };
                    }
                }
            }
        });
    }
    /**
     * エージェントID発行完了
     */
    agentIdRegistered(agent) {
        return __awaiter(this, void 0, void 0, function* () {
            console.log('agentIdRegistered');
            // エージェント更新通知
            return yield this.manager.app.notifyProfile(yield this.manager.app.getProfile(), agent);
        });
    }
    /**
     * エージェント接続完了
     */
    agentConnected(agent) {
        return __awaiter(this, void 0, void 0, function* () {
            // エージェント更新通知
            return yield this.manager.app.notifyProfile(yield this.manager.app.getProfile(), agent);
        });
    }
    /**
     * 終了
     */
    end() {
        var _a;
        (_a = this.socketIoRelay) === null || _a === void 0 ? void 0 : _a.close();
    }
}
exports.ControlSessionClient = ControlSessionClient;
