"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ControlSessionManager = void 0;
const config = require("config");
const fs = require("fs");
const i18n_1 = require("i18n");
const moment = require("moment");
const nanoid_1 = require("nanoid");
const nodeUuid = require("node-uuid");
const utils_1 = require("../utils");
const client_1 = require("./client");
class ControlSessionManager {
    constructor(app) {
        this.ended = false;
        this.agent = null;
        this.versionInfo = {
            latestVersion: null,
            latestUrl: null,
            announcement: null
        };
        /**
         * ステータスコード
         */
        this.statusCodes = {
            '0': 'オフライン',
            '101': '接続要求',
            '102': '処理中',
            '201': 'オンライン',
            '400': '接続エラー',
            '401': '認証失敗',
            '409': '接続の競合',
            '419': '接続期限切れ',
            '426': '要アップデート',
            '429': 'リクエスト回数超過',
            '500': '内部エラー'
        };
        this.status = null;
        /**
         * 接続リトライのインターバル
         */
        this.retryIntervals = [
            10000,
            30000,
            60000,
            300000,
            900000,
            1800000
        ];
        this.retryIntervalsIndex = null;
        this.retryTimestamp = 0;
        this.app = app;
    }
    /**
     * ライフサイクル
     */
    lifeCycle() {
        return __awaiter(this, void 0, void 0, function* () {
            const status = yield this.getStatus();
            const profile = yield this.app.getProfile();
            // プロフィールが未設定の場合はオフラインに変更する
            if (!profile.registered && status.code) {
                yield this.updateStatus(0);
            }
            // 接続リトライ判定
            if (status.code >= 400 && status.code != 401 && status.code != 419 && status.code != 426 && this.retryTimestamp <= moment().valueOf()) {
                yield this.updateStatus(101);
            }
            // 接続処理
            if (status.code >= 100 && status.code < 400) {
                // 接続開始
                this.client = new client_1.ControlSessionClient(this, `${config.apiRelayServer.endpoint}:${config.apiRelayServer.port}`);
                const status = yield this.client.start();
                // ステータスをトースト通知
                if (status.code != 101) {
                    this.app.desktopNotification(`${i18n_1.__('ステータス')}: ${status.message}`);
                }
                // 400,500系エラーの場合
                if (status.code >= 400) {
                    // ワンタイムエージェントの場合、プロセス終了する
                    if (config.feature.one_time_connectable) {
                        yield this.app.processKillIfAllDisconnected();
                    }
                    // 401(認証エラー),419(接続期限切れ),426(要アップデート)以外の場合
                    if (status.code != 401 && status.code != 419 && status.code != 426) {
                        // リトライを設定
                        this.retryTimestamp = moment().valueOf() + this.getRetryInterval();
                    }
                }
                this.client = null;
            }
            // ループ
            if (!this.ended) {
                this.lifeCycleTimer = setTimeout(() => __awaiter(this, void 0, void 0, function* () {
                    this.lifeCycle();
                }), 1000);
            }
        });
    }
    /**
     * リトライインターバル取得
     */
    getRetryInterval() {
        if (this.retryIntervalsIndex === null) {
            this.retryIntervalsIndex = -1;
        }
        if (this.retryIntervals.length - 1 > this.retryIntervalsIndex) {
            this.retryIntervalsIndex++;
        }
        return this.retryIntervals[this.retryIntervalsIndex];
    }
    /**
     * 最新バージョン取得
     */
    getVersionInfo() {
        return this.versionInfo;
    }
    /**
     * 最新バージョン更新
     */
    setVersionInfo(versionInfo) {
        this.versionInfo = versionInfo;
        return this.versionInfo;
    }
    /**
     * ステータス取得
     */
    getStatus() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.status) {
                return Object.assign({}, this.status, {
                    message: yield this.getStatusMessage(this.status.code) // 言語を切り替えた場合のためメッセージは再取得する
                });
            }
            else {
                try {
                    const data = yield fs.promises.readFile(`${config.apiRelayClient.status.fileName}`, {
                        encoding: config.apiRelayClient.status.encoding
                    });
                    const parsed = JSON.parse(data.toString());
                    this.status = {
                        code: parsed.code,
                        message: yield this.getStatusMessage(parsed.code),
                        registered: (yield this.getAgent()).registered,
                        versionInfo: this.getVersionInfo()
                    };
                    return this.status;
                }
                catch (e) {
                    return {
                        code: 0,
                        message: yield this.getStatusMessage(0),
                        registered: (yield this.getAgent()).registered,
                        versionInfo: this.getVersionInfo()
                    };
                }
            }
        });
    }
    /**
     * ステータス更新
     */
    updateStatus(code) {
        return __awaiter(this, void 0, void 0, function* () {
            const data = {
                code: code
            };
            utils_1.writeFile(`${config.apiRelayClient.status.fileName}`, JSON.stringify(data), config.apiRelayClient.status.encoding);
            this.status = {
                code: data.code,
                message: yield this.getStatusMessage(data.code),
                registered: (yield this.getAgent()).registered,
                versionInfo: this.getVersionInfo()
            };
            if (code < 100 || code >= 200) {
                this.app.eventEmitter.emit('updateStatus', this.status);
            }
            if (code == 201) {
                this.retryIntervalsIndex = null;
            }
            yield this.app.broadcastStatus();
            console.log(this.status);
            return this.status;
        });
    }
    /**
     * バージョン情報更新
     */
    updateVersionInfo(versionInfo) {
        return __awaiter(this, void 0, void 0, function* () {
            this.versionInfo = versionInfo;
            this.status.versionInfo = versionInfo;
            yield this.app.broadcastStatus();
            return this.versionInfo;
        });
    }
    /**
     * ステータス削除
     */
    deleteStatus() {
        if (fs.existsSync(config.apiRelayClient.status.fileName)) {
            fs.unlinkSync(config.apiRelayClient.status.fileName);
        }
        this.status = null;
    }
    /**
     * ステータスメッセージ取得
     */
    getStatusMessage(code) {
        return __awaiter(this, void 0, void 0, function* () {
            if (code === undefined) {
                code = (yield this.getStatus()).code;
            }
            return i18n_1.__(this.statusCodes[code.toString()]);
        });
    }
    /**
     * エージェント情報取得
     */
    getAgent() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.agent) {
                return Object.assign({}, this.agent);
            }
            else {
                try {
                    const data = yield fs.promises.readFile(`${config.apiRelayClient.agent.fileName}`, {
                        encoding: config.apiRelayClient.agent.encoding
                    });
                    console.log(`Loaded agent info: ${data.toString()}`);
                    this.agent = JSON.parse(data.toString());
                    return this.agent;
                }
                catch (e) {
                    const agent = this.createUnregisteredAgent();
                    console.log(`created agent info: ${JSON.stringify(agent)}`);
                    return this.saveAgent(agent);
                }
            }
        });
    }
    /**
     * エージェント情報保存
     */
    saveAgent(agent) {
        utils_1.writeFile(`${config.apiRelayClient.agent.fileName}`, JSON.stringify(agent), config.apiRelayClient.agent.encoding);
        this.agent = agent;
        return agent;
    }
    /**
     * エージェント情報リセット
     */
    resetAgent() {
        this.saveAgent(this.createUnregisteredAgent());
    }
    /**
     * 未登録エージェント情報生成
     */
    createUnregisteredAgent() {
        return {
            registered: false,
            id: nodeUuid.v4(),
            secret: nanoid_1.nanoid(80)
        };
    }
    /**
     * 切断
     */
    disconnect(processKillIfAllDisconnected = true) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.client) {
                this.client.end();
            }
            const status = yield this.getStatus();
            this.app.desktopNotification(`${i18n_1.__('ステータス')}: ${status.message}`);
            if (processKillIfAllDisconnected) {
                yield this.app.processKillIfAllDisconnected();
            }
        });
    }
    /**
     * 終了処理
     */
    end() {
        this.ended = true;
        clearTimeout(this.lifeCycleTimer);
        if (this.client) {
            this.client.end();
        }
    }
}
exports.ControlSessionManager = ControlSessionManager;
