"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Desktop = void 0;
const config = require("config");
const electron = require("electron");
const i18n_1 = require("i18n");
const os = require("os");
const path = require("path");
class Desktop {
    constructor(app) {
        this.window = null;
        this.app = app;
        electron.app.setAppUserModelId(config.appId);
    }
    start() {
        return __awaiter(this, void 0, void 0, function* () {
            electron.ipcMain.handle("getPort", () => __awaiter(this, void 0, void 0, function* () {
                return this.app.getPort();
            }));
            electron.app.on('activate', () => __awaiter(this, void 0, void 0, function* () {
                yield this.initWindow();
            }));
            electron.app.on('window-all-closed', () => { });
            electron.app.on('before-quit', (event) => __awaiter(this, void 0, void 0, function* () {
                event.preventDefault();
                yield this.app.exit();
            }));
            yield this.initWindow();
            this.initApplicationMenu();
            this.initTray();
            this.initContextMenu();
            yield this.initAbout();
        });
    }
    /**
     * 終了
     */
    exit(title, content) {
        return __awaiter(this, void 0, void 0, function* () {
            electron.dialog.showErrorBox(title, content);
            yield this.app.exit();
        });
    }
    /**
     * ウインドウ初期化
     */
    initWindow() {
        return new Promise((resolve, reject) => {
            // ウインドウがすでにある場合
            if (this.window) {
                this.window.show();
                resolve(this.window);
                return;
            }
            // ウインドウ作成
            electron.app.whenReady().then(() => {
                this.window = new electron.BrowserWindow({
                    title: config.productName,
                    webPreferences: {
                        nodeIntegration: true,
                        webviewTag: true
                    }
                });
                this.window.on('closed', () => {
                    this.window = null;
                });
                this.window.loadFile(`resources/desktop/webview.html`);
                resolve(this.window);
            });
        });
    }
    /**
     * アプリケーションメニュー初期化
     */
    initApplicationMenu() {
        const menu = electron.Menu.buildFromTemplate([
            {
                label: i18n_1.__('ファイル'),
                submenu: [
                    {
                        role: 'about', label: i18n_1.__('%s について', config.productName)
                    },
                    {
                        type: 'separator'
                    },
                    {
                        role: 'quit', label: i18n_1.__('%s を終了する', config.productName)
                    }
                ]
            },
            {
                label: i18n_1.__('編集'),
                submenu: [
                    { role: 'undo', label: i18n_1.__('元に戻す') },
                    { role: 'redo', label: i18n_1.__('やり直す') },
                    { type: 'separator' },
                    { role: 'cut', label: i18n_1.__('切り取り') },
                    { role: 'copy', label: i18n_1.__('コピー') },
                    { role: 'paste', label: i18n_1.__('貼り付け') },
                    { role: 'selectAll', label: i18n_1.__('すべてを選択') }
                ]
            },
            {
                label: i18n_1.__('ウィンドウ'),
                submenu: [
                    {
                        role: 'close', label: i18n_1.__('ウィンドウを閉じる')
                    }
                ]
            }
        ]);
        electron.Menu.setApplicationMenu(menu);
    }
    /**
     * トレイ初期化
     */
    initTray() {
        this.tray = new electron.Tray(path.resolve(`${__dirname}/../../resources/desktop/${os.platform() == 'win32' ? 'icon.ico' : 'tray-icon.png'}`));
        this.tray.setToolTip(config.productName);
        return this.tray;
    }
    /**
     * コンテキストメニュー初期化
     */
    initContextMenu() {
        this.tray.setContextMenu(electron.Menu.buildFromTemplate([
            {
                label: i18n_1.__('ダッシュボード'),
                click: () => __awaiter(this, void 0, void 0, function* () {
                    yield this.initWindow();
                })
            },
            {
                type: 'separator'
            },
            {
                role: 'quit', label: i18n_1.__('%s を終了する', config.productName)
            }
        ]));
    }
    /**
     * About初期化
     */
    initAbout() {
        return __awaiter(this, void 0, void 0, function* () {
            const version = yield this.app.getVersion();
            electron.app.setAboutPanelOptions({
                applicationName: config.productName,
                applicationVersion: `${version.version} (${version.edition})`,
                version: '',
                copyright: require('../../package.json').author
            });
        });
    }
    /**
     * 通知
     */
    notification(message) {
        new electron.Notification({
            title: config.productName,
            body: message,
            silent: true
        }).show();
    }
    /**
     * 外部リンク
     */
    externalLink(url) {
        electron.shell.openExternal(url);
        return url;
    }
}
exports.Desktop = Desktop;
