"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.HTTPServer = void 0;
const bodyParser = require("body-parser");
const bcrypt = require("bcryptjs");
const config = require("config");
const crypto = require("crypto");
const express = require("express");
const session = require("express-session");
const fs = require("fs");
const http = require("http");
const passport = require("passport");
const passportCustom = require("passport-custom");
const utils_1 = require("../../utils");
const api_1 = require("./api");
const web_1 = require("./web");
const RelayTokenStore_1 = require("../../socket-io.apiRelay/RelayTokenStore");
const socketIo_1 = require("../socketIo/socketIo");
class HTTPServer {
    constructor(app) {
        this.credentials = null;
        this.app = app;
        this.express = express();
        this.express.set("view engine", "ejs");
        this.express.set('views', `${__dirname}/../../../resources/views`);
        this.express.use(express.static(`${__dirname}/../../../public`));
        this.express.use('/storage', (req, res, next) => {
            RelayTokenStore_1.default.authRequest(req, res, next);
        });
        this.express.use('/storage', express.static(config.storage.path));
        this.express.use(bodyParser.urlencoded({ extended: true }));
        this.express.use(bodyParser.json());
        // セッション
        this.session = session({ secret: crypto.randomBytes(30).toString("hex") });
        this.express.use(this.session);
        this.express.use(passport.initialize());
        this.express.use(passport.session());
        this.express.use(require('express-flash')());
        this.express.use(this.csrfProtect);
        // 認証
        passport.use(new passportCustom.Strategy((req, done) => __awaiter(this, void 0, void 0, function* () {
            const credential = yield this.getCredential('root');
            if (credential.secret !== null && (yield bcrypt.compare(req.body.password, credential.secret))) {
                done(null, 'root');
            }
            else {
                done(null, false);
            }
        })));
        passport.serializeUser(function (user, done) {
            done(null, user);
        });
        passport.deserializeUser(function (user, done) {
            done(null, user);
        });
        this.express.use((req, res, next) => __awaiter(this, void 0, void 0, function* () {
            res.locals.lang = (yield this.app.getLang()).value;
            next();
        }));
    }
    /**
     * サーバ起動
     */
    serve() {
        return __awaiter(this, void 0, void 0, function* () {
            new api_1.API(this);
            if (config.feature.web_ui) {
                new web_1.Web(this);
            }
            // HTTPサーバを起動する
            this.server = new http.Server(this.express);
            this.socketIoServer = new socketIo_1.SocketIoServer(this);
            this.socketIoServer.preparate();
            this.port = yield this.listen();
            console.log(`Http server listening at http://127.0.0.1:${this.port}`);
        });
    }
    /**
     * listen開始
     */
    listen() {
        return new Promise((resolve, reject) => {
            let [startPort, endPort] = config.port.httpServer.split('-').map((port) => {
                return parseInt(port);
            });
            if (!endPort) {
                endPort = startPort;
            }
            const attempt = (port) => {
                const listeningHandler = () => {
                    this.server.off('error', errorHandler);
                    resolve(port);
                };
                const errorHandler = (error) => {
                    this.server.off('listening', listeningHandler);
                    console.log(error.message);
                    if (error.message.match('EADDRINUSE') && port < endPort) {
                        attempt(++port);
                    }
                    else {
                        reject(error);
                    }
                };
                this.server.once('listening', listeningHandler);
                this.server.once('error', errorHandler);
                this.server.listen(port);
            };
            attempt(startPort);
        });
    }
    /**
     * ポート番号取得
     */
    getPort() {
        return this.port;
    }
    /**
     * CSRF保護
     */
    csrfProtect(req, res, next) {
        if (req.path.match(/^\/api\//i) && req.header('x-agent-id') !== undefined) {
            return next();
        }
        if (req.method == 'GET') {
            if (!req.session['csrfToken']) {
                const token = crypto.randomBytes(30).toString('hex');
                req.session['csrfToken'] = token;
            }
            res.locals.csrfToken = req.session['csrfToken'];
            res.locals.csrf = `<input type="hidden" name="_token" value="${req.session['csrfToken']}">`;
        }
        else if (req.method == 'POST' || req.method == 'PUT' || req.method == 'PATCH' || req.method == 'DELETE') {
            if (req.header('x-requested-with') == 'XMLHttpRequest') {
                if (req.header('x-csrf-token') === undefined || req.session['csrfToken'] !== req.header('x-csrf-token')) {
                    res.status(419).send('Page Expired');
                    return;
                }
            }
            else {
                if (req.body._token === undefined || req.session['csrfToken'] !== req.body._token) {
                    res.status(419).send('Page Expired');
                    return;
                }
            }
        }
        return next();
    }
    /**
     * 認証情報更新
     */
    setCredential(userName, credential) {
        return __awaiter(this, void 0, void 0, function* () {
            const credentials = yield this.getCredentials();
            credential.secret = yield bcrypt.hash(credential.secret, 8);
            credentials[userName] = credential;
            utils_1.writeFile(`${config.credentials.fileName}`, JSON.stringify(credentials), config.credentials.encoding);
            this.credentials = credentials;
            return credential;
        });
    }
    /**
     * 認証情報取得
     */
    getCredential(userName) {
        return __awaiter(this, void 0, void 0, function* () {
            let credential;
            if (this.credentials) {
                credential = this.credentials[userName];
            }
            else {
                credential = (yield this.getCredentials())[userName];
            }
            if (credential === undefined) {
                credential = {
                    secret: null
                };
            }
            return credential;
        });
    }
    /**
     * すべての認証情報取得
     */
    getCredentials() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.credentials) {
                return Object.assign({}, this.credentials);
            }
            else {
                try {
                    const data = yield fs.promises.readFile(`${config.credentials.fileName}`, {
                        encoding: config.credentials.encoding
                    });
                    this.credentials = JSON.parse(data.toString());
                    return this.credentials;
                }
                catch (e) {
                    return {};
                }
            }
        });
    }
    /**
     * すべての認証情報削除
     */
    deleteCredentials() {
        if (fs.existsSync(config.credentials.fileName)) {
            fs.unlinkSync(config.credentials.fileName);
        }
        this.credentials = null;
    }
}
exports.HTTPServer = HTTPServer;
