"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Web = void 0;
const config = require("config");
const express = require("express");
class Web {
    constructor(httpServer) {
        this.httpServer = httpServer;
        this.express = this.httpServer.express;
        this.router();
    }
    /**
     * 認可
     */
    isAuthenticated(req, res, next) {
        if (req.isAuthenticated()) {
            res.locals.isAuth = true;
            res.locals.username = req.user;
            return next();
        }
        res.locals.isAuth = false;
        res.locals.username = null;
        res.redirect('/login');
    }
    /**
     * 認証済みの場合のリダイレクト
     */
    redirectIfauthenticated(req, res, next) {
        if (!req.isAuthenticated()) {
            next();
            return;
        }
        res.redirect('/dashboard');
    }
    /**
     * 言語未設定時のリダイレクト
     */
    redirectIfLangUndefined(req, res, next) {
        return __awaiter(this, void 0, void 0, function* () {
            if ((yield this.httpServer.app.getLang()).defined) {
                next();
                return;
            }
            res.redirect('wizard/lang');
        });
    }
    /**
     * パスワード未設定時のリダイレクト
     */
    redirectIfPasswordUndefined(req, res, next) {
        return __awaiter(this, void 0, void 0, function* () {
            if ((yield this.httpServer.getCredential('root')).secret !== null) {
                next();
                return;
            }
            res.redirect('wizard/password');
        });
    }
    /**
     * プロフィール未設定時のリダイレクト
     */
    redirectIfProfileUndefined(req, res, next) {
        return __awaiter(this, void 0, void 0, function* () {
            const excludePaths = [
                '/wizard/profile',
                '/logout'
            ];
            if (excludePaths.find((path) => path == req.path) || (yield this.httpServer.app.getProfile()).registered) {
                next();
                return;
            }
            res.redirect('/wizard/profile');
        });
    }
    /**
     * Router
     */
    router() {
        return __awaiter(this, void 0, void 0, function* () {
            this.express.use((req, res, next) => {
                res.locals.feature = config.feature;
                res.locals.productName = config.productName;
                res.locals.serviceName = config.serviceName;
                res.locals.serviceUrl = config.serviceUrl;
                res.locals.path = req.path;
                next();
            });
            this.guestRoute();
            this.authedRoute();
        });
    }
    /**
     * ゲストルート
     */
    guestRoute() {
        // ルータ作成
        const guest = express.Router();
        this.express.use(guest);
        // ルーティング
        guest.get('/', (req, res) => {
            res.redirect('/login');
        });
        guest.get('/wizard/lang', this.redirectIfauthenticated, (req, res) => {
            this.initLang(req, res);
        });
        guest.get('/wizard/password', this.redirectIfauthenticated, (req, res) => {
            this.initPassword(req, res);
        });
        guest.get('/login', this.redirectIfauthenticated, (req, res, next) => __awaiter(this, void 0, void 0, function* () {
            yield this.redirectIfLangUndefined(req, res, next);
        }), (req, res, next) => __awaiter(this, void 0, void 0, function* () {
            yield this.redirectIfPasswordUndefined(req, res, next);
        }), (req, res) => {
            this.login(req, res);
        });
    }
    /**
     * 認証済みルート
     */
    authedRoute() {
        // ルータ作成
        const authed = express.Router();
        this.express.use(authed);
        // ミドルウェア設定
        authed.use(this.isAuthenticated);
        authed.use((req, res, next) => __awaiter(this, void 0, void 0, function* () {
            yield this.redirectIfProfileUndefined(req, res, next);
        }));
        // ルーティング
        authed.post('/logout', (req, res) => {
            this.logout(req, res);
        });
        authed.get('/dashboard', (req, res) => __awaiter(this, void 0, void 0, function* () {
            yield this.dashboard(req, res);
        }));
        authed.get('/logs', (req, res) => {
            this.logs(req, res);
        });
        authed.get('/settings', (req, res) => __awaiter(this, void 0, void 0, function* () {
            yield this.settings(req, res);
        }));
        authed.get('/password', (req, res) => {
            this.password(req, res);
        });
        authed.get('/wizard/profile', (req, res) => __awaiter(this, void 0, void 0, function* () {
            yield this.wizardProfile(req, res);
        }));
    }
    /**
     * ログイン
     */
    login(req, res) {
        res.render('login');
    }
    /**
     * ログアウト
     */
    logout(req, res) {
        req.logout();
        req.session.destroy(() => {
            res.redirect('/login');
        });
    }
    /**
     * Agentホーム
     */
    dashboard(req, res) {
        return __awaiter(this, void 0, void 0, function* () {
            res.render('dashboard', {
                profile: yield this.httpServer.app.getProfile(),
                hardware: yield this.httpServer.app.getHardware(),
                version: yield this.httpServer.app.getVersion()
            });
        });
    }
    /**
     * ログ
     */
    logs(req, res) {
        res.render('logs');
    }
    /**
     * 設定
     */
    settings(req, res) {
        return __awaiter(this, void 0, void 0, function* () {
            res.render('settings', {
                lang: (yield this.httpServer.app.getLang()).value,
                profile: yield this.httpServer.app.getProfile(),
                proxy: Object.assign(Object.assign({}, yield this.httpServer.app.getProxy()), { readonly: !!config.proxy.command }),
                edition: (yield this.httpServer.app.getVersion()).edition
            });
        });
    }
    /**
     * パスワード変更
     */
    password(req, res) {
        res.render('password');
    }
    /**
     * 言語初期化(ウィザード)
     */
    initLang(req, res) {
        return __awaiter(this, void 0, void 0, function* () {
            res.render('wizard/lang');
        });
    }
    /**
     * パスワード初期化(ウィザード)
     */
    initPassword(req, res) {
        return __awaiter(this, void 0, void 0, function* () {
            res.render('wizard/password');
        });
    }
    /**
     * プロフィール設定(ウィザード)
     */
    wizardProfile(req, res) {
        return __awaiter(this, void 0, void 0, function* () {
            res.render('wizard/profile', {
                profile: yield this.httpServer.app.getProfile()
            });
        });
    }
}
exports.Web = Web;
