"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.SocketIoRelay = exports.STATUS_LISTEN = exports.STATUS_READY = exports.STATUS_NONE = void 0;
const socketIo = require("socket.io-client");
const ClientSocket_1 = require("./ClientSocket");
const events_1 = require("events");
exports.STATUS_NONE = 0;
exports.STATUS_READY = 1; // registered or authenticated
exports.STATUS_LISTEN = 2;
class SocketIoRelay extends events_1.EventEmitter {
    constructor(relayUrl, agent, options) {
        super();
        this.socket = {};
        this.clients = {};
        this.agent_id = '';
        this.secret = '';
        this.relay_url = '';
        this.url = relayUrl;
        this.status = exports.STATUS_NONE;
        let connectOpts = {};
        connectOpts.reconnection = false;
        if (options.transportsWsOnly) {
            connectOpts.transports = ["websocket"];
        }
        if (options.timeout) {
            connectOpts.timeout = options.timeout;
        }
        if (options.reconnectionAttempts) {
            connectOpts.reconnectionAttempts = options.reconnectionAttempts;
        }
        connectOpts.agent = agent;
        this.socket = socketIo.connect(this.url, connectOpts);
        this.socket.io.on("connect_error", (err) => {
            super.emit('connect_error', err);
        });
    }
    get agentId() {
        return this.agent_id;
    }
    get agentSecret() {
        return this.secret;
    }
    get relayUrl() {
        return this.relay_url;
    }
    get disconnected() {
        return this.socket.disconnected;
    }
    register() {
        return new Promise((resolve, reject) => {
            if (this.status == exports.STATUS_READY) {
                // already registerd
                resolve({ code: 200 });
                return;
            }
            this.socket.emit('register', '', '', (code, registeredAgentId, registeredSecret) => {
                if (code === 200 && registeredAgentId && registeredSecret) {
                    this.agent_id = registeredAgentId;
                    this.secret = registeredSecret;
                    this.status = exports.STATUS_READY;
                    console.log(`Agent is registered. agent_id=${registeredAgentId} secret=${registeredSecret}`);
                    resolve({
                        code: 200,
                        agentId: this.agent_id,
                        secret: this.secret
                    });
                }
                else {
                    console.log(`Failed register. code=${code}`);
                    this.socket.close();
                    reject({ code });
                }
            });
        });
    }
    authenticate(agentId, secret) {
        this.agent_id = agentId;
        this.secret = secret;
        return new Promise((resolve, reject) => {
            if (this.status == exports.STATUS_READY) {
                // already registerd
                resolve({ code: 200 });
                return;
            }
            this.socket.emit('register', agentId, secret, (code) => {
                if (code === 200) {
                    this.status = exports.STATUS_READY;
                    console.log(`Agent is authenticated. agent_id=${agentId}`);
                    resolve({ code: 200 });
                }
                else {
                    console.log(`Failed authenticate. code=${code}`);
                    this.socket.close();
                    reject({ code });
                }
            });
        });
    }
    listen(apiHost, apiPort, tokenId) {
        return new Promise((resolve, reject) => {
            if (!tokenId && this.status != exports.STATUS_READY) {
                // invalid status
                reject({ code: 400 });
                return;
            }
            if (!tokenId) {
                this.socket.emit('listen', apiHost, apiPort, (code, relayUrl) => {
                    callback(code, relayUrl);
                });
            }
            else {
                this.socket.emit('onetime_listen', apiHost, apiPort, tokenId, (code, relayUrl) => {
                    callback(code, relayUrl);
                });
            }
            const callback = (code, relayUrl) => {
                if (code === 200 && relayUrl) {
                    this.relay_url = relayUrl;
                    console.log(`Listening from url=${relayUrl}`);
                    resolve({ code: 200 });
                }
                else {
                    console.log(`Failed Listening. code=${code}`);
                    this.socket.close();
                    reject({ code });
                }
            };
            this.socket.on('client_connect', (clientId) => {
                console.log('Connected ws client[#%s]', clientId);
                const clientSocket = new ClientSocket_1.ClientSocket(clientId, this);
                this.clients[clientId] = clientSocket;
                super.emit('connect', clientSocket);
                this.send(clientId, 'server_connect');
            });
            this.socket.on('client_disconnect', (clientId) => {
                if (clientId in this.clients) {
                    console.log('Disconnected ws client[#%s]', clientId);
                    this.clients[clientId].removeAllSocketListeners();
                    this.clients[clientId].emit('disconnect', clientId);
                    this.clients[clientId].disconnected = true;
                    delete this.clients[clientId];
                }
            });
            this.socket.on('connect', () => {
                console.log('Connected to relay server(#%s)', this.agentId);
                super.emit('relay_connect');
            });
            this.socket.on('disconnect', () => {
                console.log('Disconnected from relay server(#%s)', this.agentId);
                super.emit('relay_disconnect');
            });
        });
    }
    on(event, listener) {
        super.on(event, listener);
        return this;
    }
    once(event, listener) {
        super.once(event, listener);
        return this;
    }
    off(event, listener) {
        super.off(event, listener);
        return this;
    }
    emit(event, ...datas) {
        if (typeof (event) === 'string') {
            this.send('', event, ...datas);
            return true;
        }
        return false;
    }
    send(clientId, eventName, ...datas) {
        const emitData = {
            clientId,
            eventName,
        };
        this.socket.emit('emit', emitData, ...datas);
    }
    getSocket() {
        return this.socket;
    }
    close() {
        this.socket.close();
    }
}
exports.SocketIoRelay = SocketIoRelay;
