"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ClientSocket = exports.EmitWithAckTimeoutException = void 0;
const config = require("config");
const events_1 = require("events");
class EmitWithAckTimeoutException extends Error {
    constructor(...args) {
        super(...args);
        this.name = this.constructor.name;
        this.message = 'ack timed out';
    }
}
exports.EmitWithAckTimeoutException = EmitWithAckTimeoutException;
class ClientSocket extends events_1.EventEmitter {
    constructor(id, socketIoRelay, headers = []) {
        super();
        this._id = id;
        this.socketIoRelay = socketIoRelay;
        this.events = [];
        this._handshake = {
            headers: headers
        };
        this._disconnected = false;
        this.setMaxListeners(128);
    }
    get id() {
        return this._id;
    }
    get connected() {
        return !this._disconnected;
    }
    get disconnected() {
        return this._disconnected;
    }
    get serverSocket() {
        return this.socketIoRelay;
    }
    get handshake() {
        return this._handshake;
    }
    set disconnected(v) {
        this._disconnected = v;
    }
    clientEvent(event) {
        return this._id + '/' + event;
    }
    isSocketRelayEvent(event) {
        if (event === 'disconnect') {
            return false;
        }
        else {
            return true;
        }
    }
    on(event, listener) {
        if (this.isSocketRelayEvent(event)) {
            const cEvent = this.clientEvent(event);
            const socket = this.socketIoRelay.getSocket();
            socket.emit('attach', cEvent);
            socket.on(cEvent, (_serverId, ..._args) => {
                if (this.socketIoRelay.serverId === _serverId) {
                    // 自分のネームスペースのみを処理する。
                    listener(..._args);
                }
            });
            this.events.push(cEvent);
        }
        else {
            super.on(event, listener);
        }
        return this;
    }
    once(event, listener) {
        if (this.isSocketRelayEvent(event)) {
            const cEvent = this.clientEvent(event);
            const socket = this.socketIoRelay.getSocket();
            socket.emit('attach', cEvent);
            socket.once(cEvent, (_serverId, ..._args) => {
                if (this.socketIoRelay.serverId === _serverId) {
                    // 自分のネームスペースのみを処理する。
                    listener(..._args);
                }
            });
        }
        else {
            super.once(event, listener);
        }
        return this;
    }
    off(event, listener) {
        if (this.isSocketRelayEvent(event)) {
            const cEvent = this.clientEvent(event);
            const socket = this.socketIoRelay.getSocket();
            socket.off(cEvent, listener);
        }
        else {
            super.off(event, listener);
        }
        return this;
    }
    emit(event, ...args) {
        if (typeof (event) === 'string') {
            if (this.isSocketRelayEvent(event)) {
                this.socketIoRelay.send(this._id, event, ...args);
            }
            else {
                super.emit(event, ...args);
            }
            return true;
        }
        return false;
    }
    emitWithAckTimeout(event, ...args) {
        return new Promise((_resolve, _reject) => {
            let timeout;
            const disconnect = () => {
                clearTimeout(timeout);
                _reject(new EmitWithAckTimeoutException);
            };
            const resolve = (value) => {
                this.off('disconnect', disconnect);
                clearTimeout(timeout);
                _resolve(value);
            };
            const reject = (reason) => {
                this.off('disconnect', disconnect);
                clearTimeout(timeout);
                _reject(reason);
            };
            if (this.disconnected) {
                reject(new EmitWithAckTimeoutException);
                return;
            }
            if (typeof (event) === 'string') {
                let hasAck = false;
                args.forEach((arg, index) => {
                    if (typeof arg == 'function') {
                        args[index] = (..._args) => {
                            arg(..._args);
                            resolve(true);
                        };
                        hasAck = true;
                    }
                });
                if (this.isSocketRelayEvent(event)) {
                    this.socketIoRelay.send(this._id, event, ...args);
                }
                else {
                    super.emit(event, ...args);
                }
                if (!hasAck) {
                    resolve(true);
                }
                else {
                    this.once('disconnect', disconnect);
                    timeout = setTimeout(() => {
                        reject(new EmitWithAckTimeoutException);
                    }, config.relayServer.ackTimeout);
                }
            }
            else {
                resolve(false);
            }
        });
    }
    removeAllSocketListeners() {
        const socket = this.socketIoRelay.getSocket();
        this.events.forEach((event) => {
            socket.removeEventListener(event);
        });
    }
    disconnect() {
        const socket = this.socketIoRelay.getSocket();
        socket.emit('client_disconnect', this._id, this.socketIoRelay.serverId);
    }
}
exports.ClientSocket = ClientSocket;
