"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.startClientSocket = exports.createProxyUrl = exports.md5Hash = exports.countKeys = exports.sleep = exports.makeStorageDirAndTryCallback = exports.writeFile = void 0;
const config = require("config");
const crypto = require("crypto");
const moment = require("moment");
const io = require("socket.io-client");
const fs = require("fs");
const path = require("path");
/**
 * ファイル書き込み
 */
const writeFile = (filepath, data, encode) => {
    if (!fs.existsSync(path.dirname(filepath))) {
        fs.mkdirSync(path.dirname(filepath), {
            recursive: true
        });
    }
    fs.writeFileSync(filepath, data, {
        encoding: encode
    });
};
exports.writeFile = writeFile;
/**
 * ストレージディレクトリを作成する
 */
const makeStorageDir = (path) => __awaiter(void 0, void 0, void 0, function* () {
    if (!fs.existsSync(config.storage.path)) {
        try {
            yield fs.promises.mkdir(config.storage.path, {
                recursive: true
            });
        }
        catch (e) {
            console.log(e.message);
        }
    }
    const currentTime = new Date();
    const parsedPath = path.split('/');
    let targetPath = config.storage.path;
    for (const _path of parsedPath) {
        targetPath += `/${_path}`;
        try {
            fs.utimesSync(targetPath, currentTime, currentTime);
        }
        catch (e) {
            try {
                yield fs.promises.mkdir(targetPath, {
                    recursive: true
                });
            }
            catch (e) {
                console.log(e.message);
            }
        }
    }
    if (!fs.existsSync(targetPath)) {
        throw new Error(`makeStorageDir failed. [${targetPath}]`);
    }
    return targetPath;
});
/**
 * ストレージディレクトリを作成してからコールバックを処理する
 */
const makeStorageDirAndTryCallback = (path, callback) => __awaiter(void 0, void 0, void 0, function* () {
    for (let count = 0; count < 3; count++) {
        try {
            return yield callback(yield makeStorageDir(path));
        }
        catch (e) {
            console.log(e.message);
        }
        yield exports.sleep(100);
    }
    throw new Error(`makeStorageDirAndTryCallback failed. [${path}]`);
});
exports.makeStorageDirAndTryCallback = makeStorageDirAndTryCallback;
/**
 * スリープ
 */
const sleep = (timeout) => {
    return new Promise((resolve, reject) => {
        setTimeout(() => {
            resolve();
        }, timeout);
    });
};
exports.sleep = sleep;
/**
 * 要素数カウント
 */
const countKeys = (object) => {
    let count = 0;
    for (const key in object) {
        count++;
    }
    return count;
};
exports.countKeys = countKeys;
/**
 * md5ハッシュ
 */
const md5Hash = (data) => {
    const md5 = crypto.createHash('md5');
    return md5.update(data).digest('hex');
};
exports.md5Hash = md5Hash;
/**
 * プロキシURL生成
 */
const createProxyUrl = (host, port, userName = null, password = null) => {
    let protocol = '';
    host = encodeURIComponent(host);
    userName = encodeURIComponent(userName);
    password = encodeURIComponent(password);
    let auth = userName !== null ? userName : '';
    if (password) {
        auth += `:${password}`;
    }
    if (auth !== '') {
        auth = `${auth}@`;
    }
    const matches = host.match(/^(https?):\/\//);
    if (matches) {
        protocol = `${matches[1]}://`;
        host = host.replace(/^https?:\/\//, '');
    }
    else {
        protocol = 'http://';
    }
    return `${protocol}${auth}${host}:${port}`;
};
exports.createProxyUrl = createProxyUrl;
/**
 * socket.io-client開始
 */
class EmitWithAckTimeoutException extends Error {
    constructor(...args) {
        super(...args);
        this.name = this.constructor.name;
        this.message = 'ack timed out';
    }
}
class PingPongTimeoutException extends Error {
    constructor(...args) {
        super(...args);
        this.name = this.constructor.name;
        this.message = 'ping-pong timed out';
    }
}
const startClientSocket = (uri, opts) => {
    const socket = io(uri, opts);
    // 未定義イベントにack要求が届いたらエラーを返す
    require('socketio-wildcard')(io.Manager)(socket);
    socket.on('*', (packet) => {
        if (!socket.hasListeners(packet.data[0])) {
            packet.data.forEach((arg) => {
                if (typeof arg == 'function') {
                    arg({
                        name: 'Error',
                        message: 'Undefined event'
                    });
                }
            });
        }
    });
    socket.emitWithAckTimeout = function (event, ...args) {
        return new Promise((_resolve, _reject) => {
            let timeout;
            const disconnect = () => {
                clearTimeout(timeout);
                _reject(new EmitWithAckTimeoutException);
            };
            let rejected = false;
            const resolve = (value) => {
                this.off('disconnect', disconnect);
                clearTimeout(timeout);
                _resolve(value);
            };
            const reject = (reason) => {
                rejected = true;
                this.off('disconnect', disconnect);
                clearTimeout(timeout);
                _reject(reason);
            };
            if (this.disconnected) {
                reject(new EmitWithAckTimeoutException);
                return;
            }
            if (typeof (event) === 'string') {
                let hasAck = false;
                args.forEach((arg, index) => {
                    if (typeof arg == 'function') {
                        args[index] = (..._args) => {
                            if (!rejected) {
                                arg(..._args);
                                resolve(true);
                            }
                        };
                        hasAck = true;
                    }
                });
                this.emit(event, ...args);
                if (!hasAck) {
                    resolve(true);
                }
                else {
                    this.once('disconnect', disconnect);
                    timeout = setTimeout(() => {
                        reject(new EmitWithAckTimeoutException);
                    }, 60000);
                }
            }
            else {
                resolve(false);
            }
        });
    };
    socket.waitUntilCanPingPong = function () {
        return new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
            let pong = false;
            this.once('_pong', () => {
                resolve();
                pong = true;
            });
            const timestamp = moment().valueOf();
            while (!pong) {
                this.emit('_ping');
                yield exports.sleep(100);
                if (timestamp + 60000 < moment().valueOf() || !this.connected) {
                    this.removeEventListener('_pong');
                    reject(new PingPongTimeoutException);
                    break;
                }
            }
        }));
    };
    return socket;
};
exports.startClientSocket = startClientSocket;
