"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Browser = void 0;
const config = require("config");
const css_selector_parser_1 = require("css-selector-parser");
const eventEmitter = require("events");
const fs = require("fs");
const jimp = require("jimp");
const moment = require("moment");
const puppeteer = require("puppeteer-core");
const FileChooser_1 = require("puppeteer-core/lib/cjs/puppeteer/common/FileChooser");
const nodeUuid = require("node-uuid");
const utils_1 = require("../../utils");
const downloader_1 = require("./downloader");
const fileManager_1 = require("./fileManager");
const authStore_1 = require("./authStore");
class Browser extends eventEmitter {
    constructor(_config) {
        super();
        this.browser = null; // ブラウザインスタンス
        this.windows = {}; // ウインドウ(タブ)リスト
        this.screenshotCasting = null; // スクリーンショットをキャストするか
        this.allowScreenshot = true; // スクリーンショットを許可するか
        this.interrupt = { screenshot: null }; // 割込処理管理のジョブIDを格納する
        this.switchedCurrentWindow = false; // ウインドウ切替検知
        this.newCastConnection = false; // 新規キャスト先接続検知
        this.windowsDiff = true; // ウインドウ差分検知
        this.newPage = false; // 新規ページ検知
        this.scrolled = false; // スクロールイベント発生検知
        this.scrollIntoView = false; // scrollIntoView発生検知
        this.dialogHandling = true; // ダイアログ発生時にハンドリングするか
        this.fileChooserHandling = true; // FileChooser発生時にハンドリングするか
        this.dialogHandlingFunction = null; // ダイアログハンドリング用Function
        this.fileChooserHandlingFunction = null; // FileChooserハンドリング用Function
        this.dialog = null;
        this.monitoringEnabled = false;
        this.fileChooser = null;
        this.authStore = new authStore_1.AuthStore;
        /**
         * ブラウザ操作コマンド
         */
        this.commands = {
            eval: (params) => __awaiter(this, void 0, void 0, function* () {
                const result = yield this.currentWindow.page.evaluate((script) => {
                    try {
                        const result = eval(script);
                        return Promise.resolve(result);
                    }
                    catch (e) {
                        return Promise.reject(e.message);
                    }
                }, params.script);
                return result;
            }),
            openWindow: () => __awaiter(this, void 0, void 0, function* () {
                const windowNumber = this.getNewWindowNumber();
                yield this.currentWindow.page.evaluate(() => {
                    window.open('about:blank');
                    Promise.resolve();
                });
                yield this.commands.waitFindWindow({ windowNumber: windowNumber, timeout: 10000 });
            }),
            switchWindow: (params) => __awaiter(this, void 0, void 0, function* () {
                const window = this.windows[params.windowNumber];
                if (window) {
                    yield window.page.bringToFront();
                    this.currentWindow = window;
                    this.switchedCurrentWindow = true;
                }
            }),
            closeWindow: (params) => __awaiter(this, void 0, void 0, function* () {
                const window = this.windows[params.windowNumber];
                if (window && utils_1.countKeys(this.windows) > 1) {
                    yield window.page.close();
                }
            }),
            waitFindWindow: (params) => __awaiter(this, void 0, void 0, function* () {
                let startTime = moment();
                const timeout = params.timeout;
                while (true) {
                    if (this.windows[params.windowNumber]) {
                        break;
                    }
                    let currentTime = moment();
                    if (currentTime.diff(startTime) > timeout) {
                        throw new Error('waitFindWindow timed out ' + timeout + 'ms');
                    }
                    yield utils_1.sleep(100);
                }
            }),
            goto: (params) => __awaiter(this, void 0, void 0, function* () {
                let url = params.url;
                if (!url.match(/^https?:\/\//)) {
                    url = 'http://' + url;
                }
                this.authStore.push(url);
                yield this.currentWindow.page.goto(url); // page.goto()はpage.waitForNavigation()も同時に行うため、page.waitForPageTransition()は不要
            }),
            goBack: () => __awaiter(this, void 0, void 0, function* () {
                // await currentWindow.page.goBack(); // page.goBack()はpage.waitForNavigation()も同時に行うため、フレーム内が遷移するとずっと待機してしまう
                yield this.currentWindow.page.evaluate(() => {
                    window.history.back();
                });
                yield this.currentWindow.page.waitForPageTransition();
            }),
            goForward: () => __awaiter(this, void 0, void 0, function* () {
                // await currentWindow.page.goForward(); // page.goForward()はpage.waitForNavigation()も同時に行うため、フレーム内が遷移するとずっと待機してしまう
                yield this.currentWindow.page.evaluate(() => {
                    window.history.forward();
                });
                yield this.currentWindow.page.waitForPageTransition();
            }),
            reload: () => __awaiter(this, void 0, void 0, function* () {
                yield this.currentWindow.page.reload();
            }),
            scroll: (params) => __awaiter(this, void 0, void 0, function* () {
                yield this.currentWindow.page.evaluate((params) => {
                    window.scrollTo(params.left, params.top);
                }, params);
                this.currentWindow.scrollLeft = params.left;
                this.currentWindow.scrollTop = params.top;
                this.scrolled = true;
            }),
            waitFindSelector: (params) => __awaiter(this, void 0, void 0, function* () {
                // 新規ウインドウの場合、遷移完了していない場合があるので待機
                yield this.currentWindow.page._waitForPageTransition();
                const [frame, selector] = yield this.getTargetFrameAndSelector(params.selector);
                // timeoutを100msで仕掛けると稀に検知できないことがある。puppeteerの不具合か？
                // await frame.waitForSelector(selector,{
                //     visible: true,
                //     timeout: params.timeout,
                // });
                const waitForSelector = (selector, timeout) => {
                    return new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
                        let startTime = moment();
                        while (true) {
                            try {
                                if (yield frame.evaluate((selector) => {
                                    // puppeteerのwaitForSelector()を再現
                                    const element = document.querySelector(selector);
                                    if (element) {
                                        const style = window.getComputedStyle(element);
                                        function hasVisibleBoundingBox(element) {
                                            const rect = element.getBoundingClientRect();
                                            return !!(rect.top || rect.bottom || rect.width || rect.height);
                                        }
                                        const isVisible = style && style.visibility !== 'hidden' && hasVisibleBoundingBox(element);
                                        if (isVisible) {
                                            return Promise.resolve(true);
                                        }
                                    }
                                    return Promise.resolve(false);
                                }, selector)) {
                                    break;
                                }
                            }
                            catch (e) {
                                // evaluate中にリダイレクトが発生すると例外発生するためリトライする
                                console.log(e);
                            }
                            let currentTime = moment();
                            if (currentTime.diff(startTime) > timeout) {
                                reject(new Error('waitForSelector timed out ' + timeout + 'ms'));
                                return;
                            }
                            yield utils_1.sleep(10);
                        }
                        // 画面内に存在しているか判定
                        try {
                            const inSight = yield frame.evaluate((selector) => {
                                const element = document.querySelector(selector);
                                if (!element) {
                                    return Promise.reject(new Error('element not found'));
                                }
                                const rect = element.getBoundingClientRect();
                                const screenTop = window.scrollY;
                                const screenBottom = screenTop + window.innerHeight;
                                const elementTop = rect.y + window.scrollY;
                                const elementBottom = elementTop + rect.height;
                                let inSight = false;
                                if (screenBottom > elementTop && screenTop < elementBottom) {
                                    inSight = true;
                                }
                                return Promise.resolve(inSight);
                            }, selector);
                            resolve(inSight);
                        }
                        catch (e) {
                            this.log(e.message);
                            reject(new Error('waitForSelector timed out ' + timeout + 'ms'));
                        }
                    }));
                };
                const inSight = yield waitForSelector(selector, params.timeout);
                if (params.scrollIntoView) {
                    if (!inSight) {
                        const beforeScroll = yield frame.evaluate(() => {
                            return Promise.resolve({
                                x: window.scrollX,
                                y: window.scrollY,
                            });
                        });
                        yield frame.evaluate((selector) => {
                            const element = document.querySelector(selector);
                            element.scrollIntoView({
                                behavior: 'instant',
                                block: 'center',
                                inline: 'center',
                            });
                        }, selector);
                        // scrollIntoViewによってスムーススクロールが発生していた場合を想定してスクロール完了するまで待機する
                        const waitForSmoothScroll = () => {
                            return new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
                                let timeout = 10000;
                                let startTime = moment();
                                const proc = (x, y) => __awaiter(this, void 0, void 0, function* () {
                                    let currentTime = moment();
                                    if (currentTime.diff(startTime) > timeout) {
                                        reject(new Error('waitForSmoothScroll timed out ' + timeout + 'ms'));
                                        return;
                                    }
                                    const scroll = yield frame.evaluate(() => {
                                        return Promise.resolve({
                                            x: window.scrollX,
                                            y: window.scrollY,
                                        });
                                    });
                                    // if(x !== null && y !== null){
                                    //     console.log('waitForSmoothScroll',scroll);
                                    // }
                                    const _x = scroll.x;
                                    const _y = scroll.y;
                                    if (x !== _x || y !== _y) {
                                        x = _x;
                                        y = _y;
                                        setTimeout(() => {
                                            proc(x, y);
                                        }, 20);
                                        return;
                                    }
                                    resolve({ x: x, y: y });
                                });
                                yield utils_1.sleep(20);
                                proc(null, null);
                            }));
                        };
                        const afterScroll = yield waitForSmoothScroll();
                        if (beforeScroll.x != afterScroll.x || beforeScroll.y != afterScroll.y) {
                            this.scrollIntoView = true;
                        }
                    }
                    yield frame.focus(selector);
                }
            }),
            clickElement: (params) => __awaiter(this, void 0, void 0, function* () {
                let frame, selector, left, top, button;
                if (params.selector) {
                    [frame, selector] = yield this.getTargetFrameAndSelector(params.selector);
                }
                left = params.left;
                top = params.top;
                if (params.button == 'right') {
                    button = 'right';
                }
                else if (params.button == 'middle') {
                    button = 'middle';
                }
                else {
                    button = 'left';
                }
                // // スクリプトでclick()を実行
                // await frame.$eval(selector,element => element.click()); // スクリプトでclick()するとロボット判定されることがある
                // // エレメントの座標を取得してマウスクリック(クリック座標がブラウザ表示範囲から見切れている場合クリックできない。調整が必要)
                // const element = await frame.$(selector);
                // const boxModel = await element.boxModel();
                // const x: number = boxModel.content.reduce((accumulator,current) => {
                //     return accumulator + current.x;
                // },0) / boxModel.content.length;
                // const y: number = boxModel.content.reduce((accumulator,current) => {
                //     return accumulator + current.y;
                // },0) / boxModel.content.length;
                // await currentWindow.page.mouse.click(x,y);
                // await element.dispose();
                // // エレメントが表示される位置までスクロールしないとクリックできないことがある
                // const element = await frame.$(selector);
                // await frame.evaluate((element) => {
                //     element.scrollIntoView({
                //         behavior: 'instant',
                //         block: 'center',
                //         inline: 'center',
                //     });
                //     return Promise.resolve();
                // },element);
                // await element.click();
                // dialog割り込みが発生中はframe.click(), this.currentWindow.page.waitForPageTransition()共にresolveしないため、
                // 割り込み発生している場合は待たない。
                let timeout;
                const dialogInterrupt = () => {
                    return new Promise((resolve, reject) => {
                        const checkDialogInterrupt = () => {
                            if (this.dialog) {
                                resolve('dialog');
                                return;
                            }
                            timeout = setTimeout(() => {
                                checkDialogInterrupt();
                            }, 100);
                        };
                        checkDialogInterrupt();
                    });
                };
                const clickAndWaitForPageTransition = () => __awaiter(this, void 0, void 0, function* () {
                    // 注意!! deviceScaleFactorの値が適切でないと、スクリーンショットを撮るタイミングでviewportの変更が発生して指定座標をクリックできなくなる
                    if (frame && selector) {
                        yield frame.click(selector, { button: button });
                    }
                    else {
                        yield this.currentWindow.page.mouse.click(left - this.currentWindow.scrollLeft, top - this.currentWindow.scrollTop, {
                            button: button
                        });
                    }
                    yield this.currentWindow.page.waitForPageTransition();
                });
                yield Promise.race([dialogInterrupt(), clickAndWaitForPageTransition()]);
                clearTimeout(timeout);
            }),
            typeElement: (params) => __awaiter(this, void 0, void 0, function* () {
                const [frame, selector] = yield this.getTargetFrameAndSelector(params.selector);
                const element = yield frame.$(selector);
                const tagName = yield frame.evaluate((element) => {
                    return Promise.resolve(element.tagName);
                }, element);
                if (typeof params.value == 'number') {
                    params.value = params.value.toString();
                }
                // プルダウンの場合
                if (tagName == 'SELECT') {
                    const disabled = yield (yield element.getProperty('disabled')).jsonValue();
                    if (!disabled) {
                        // select-multipleの場合
                        if (typeof params.value == 'object') {
                            yield frame.evaluate((element, params) => {
                                const options = element.querySelectorAll('option');
                                options.forEach((option) => {
                                    if (!option.disabled) {
                                        const _selected = option.selected;
                                        const selected = params.value.findIndex((value) => {
                                            let _value = option.getAttribute('value');
                                            if (_value === null) {
                                                _value = '';
                                            }
                                            return value == _value;
                                        }) !== -1;
                                        option.selected = selected;
                                        if (selected != _selected) {
                                            element.dispatchEvent(new Event('input', { 'bubbles': true }));
                                            element.dispatchEvent(new Event('change', { 'bubbles': true }));
                                        }
                                    }
                                });
                                return Promise.resolve();
                            }, element, params);
                        }
                        // select-oneの場合
                        else {
                            yield frame.evaluate((element, params) => {
                                const _selectedIndex = element.selectedIndex;
                                if (params.value !== '') {
                                    const options = element.querySelectorAll('option');
                                    options.forEach((option) => {
                                        if (!option.disabled && option.value == params.value) {
                                            option.selected = true;
                                        }
                                    });
                                }
                                else {
                                    let option = element.querySelector('option[value=""]');
                                    if (!option) {
                                        option = element.querySelector('option:not([value])');
                                    }
                                    if (option) {
                                        option.selected = true;
                                    }
                                }
                                if (_selectedIndex != element.selectedIndex) {
                                    element.dispatchEvent(new Event('input', { 'bubbles': true }));
                                    element.dispatchEvent(new Event('change', { 'bubbles': true }));
                                }
                                return Promise.resolve();
                            }, element, params);
                        }
                    }
                }
                // INPUTタグの場合
                else if (tagName == 'INPUT' || tagName == 'TEXTAREA') {
                    const type = yield (yield element.getProperty('type')).jsonValue();
                    // ラジオボタン・チェックボックスの場合
                    if (type == 'radio' || type == 'checkbox') {
                        const tagNames = selector.split(' > ');
                        const inputTagName = tagNames.pop().replace(/:nth-of-type\([0-9]+\)$/, '');
                        // 属しているフォームのセレクタ生成
                        let index;
                        for (index = 0; index < tagNames.length; index++) {
                            if (tagNames[index].indexOf('form') === 0) {
                                break;
                            }
                        }
                        let formSelector = '';
                        if (index < tagNames.length) {
                            formSelector = tagNames.slice(0, index + 1).join(' > ');
                        }
                        if (!formSelector) {
                            formSelector = 'html';
                        }
                        yield frame.evaluate((formSelector, inputTagName, params, type) => {
                            const form = document.querySelector(formSelector);
                            const options = form.querySelectorAll(inputTagName);
                            options.forEach((option) => {
                                if (!option.disabled) {
                                    const _checked = option.checked;
                                    let checked;
                                    if (type == 'radio') {
                                        let _value = option.getAttribute('value');
                                        if (_value === null) {
                                            _value = '';
                                        }
                                        checked = params.value == _value;
                                        if (checked) {
                                            option.checked = checked;
                                        }
                                    }
                                    else if (type == 'checkbox') {
                                        checked = params.value.findIndex((value) => {
                                            let _value = option.getAttribute('value');
                                            if (_value === null) {
                                                _value = '';
                                            }
                                            return value == _value;
                                        }) !== -1;
                                        option.checked = checked;
                                    }
                                    if (checked != _checked) {
                                        if (type != 'radio' || checked) { // radioの場合、イベント発火するのはcheckedになった時だけ。
                                            option.dispatchEvent(new Event('input', { 'bubbles': true }));
                                            option.dispatchEvent(new Event('change', { 'bubbles': true }));
                                            option.dispatchEvent(new Event('click', { 'bubbles': true }));
                                        }
                                    }
                                }
                            });
                            return Promise.resolve();
                        }, formSelector, inputTagName, params, type);
                    }
                    // テキストボックス・テキストエリアの場合
                    else {
                        yield frame.evaluate((element) => {
                            element.value = '';
                            return Promise.resolve();
                        }, element);
                        yield element.type(params.value);
                    }
                }
                yield element.dispose();
                yield this.currentWindow.page.waitForPageTransition();
            }),
            scrollElement: (params) => __awaiter(this, void 0, void 0, function* () {
                const [frame, selector] = yield this.getTargetFrameAndSelector(params.selector);
                const element = yield frame.$(selector);
                yield frame.evaluate((element, direction) => {
                    if (direction == 'up') {
                        element.scrollTop -= element.clientHeight / 2;
                    }
                    else if (direction == 'down') {
                        element.scrollTop += element.clientHeight / 2;
                    }
                    return Promise.resolve();
                }, element, params.direction);
                yield element.dispose();
            }),
            getSelectOptions: (params) => __awaiter(this, void 0, void 0, function* () {
                const [frame] = yield this.getTargetFrameAndSelector(params.selector);
                const options = yield frame.evaluate((params) => {
                    const options = [];
                    const _pulldown = document.querySelector(params.selectSelector);
                    const _options = _pulldown.querySelectorAll('option');
                    _options.forEach((_option) => {
                        let value = _option.getAttribute('value');
                        if (value == null) {
                            value = '';
                        }
                        options.push({
                            value: value,
                            label: _option.innerHTML,
                            selected: _option.selected,
                            disabled: _option.disabled
                        });
                    });
                    return Promise.resolve(options);
                }, params);
                return options;
            }),
            getInputOptions: (params) => __awaiter(this, void 0, void 0, function* () {
                const [frame] = yield this.getTargetFrameAndSelector(params.selector);
                const offsetLeft = params.frameStack.reduce((offset, rect) => offset += rect.left - rect.scrollLeft, 0);
                const offsetTop = params.frameStack.reduce((offset, rect) => offset += rect.top - rect.scrollTop, 0);
                const options = yield frame.evaluate((params, offsetLeft, offsetTop) => {
                    const options = [];
                    let formElement;
                    if (params.formSelector != '') {
                        formElement = document.querySelector(params.formSelector);
                    }
                    else {
                        formElement = document;
                    }
                    const _options = formElement.querySelectorAll(params.inputTagName);
                    _options.forEach((_option) => {
                        let value = _option.getAttribute('value');
                        if (value == null) {
                            value = '';
                        }
                        const rect = _option.getBoundingClientRect();
                        options.push({
                            name: _option.getAttribute('name'),
                            value: value,
                            width: rect.width,
                            height: rect.height,
                            left: rect.left + window.scrollX + offsetLeft,
                            top: rect.top + window.scrollY + offsetTop,
                            checked: _option.checked,
                            disabled: _option.disabled
                        });
                    });
                    return Promise.resolve(options);
                }, params, offsetLeft, offsetTop);
                return options;
            }),
            getUrl: () => __awaiter(this, void 0, void 0, function* () {
                return this.currentWindow.page.url();
            }),
            getTitle: () => __awaiter(this, void 0, void 0, function* () {
                return yield this.currentWindow.page.title();
            }),
            getText: (params) => __awaiter(this, void 0, void 0, function* () {
                const [frame, selector] = yield this.getTargetFrameAndSelector(params.selector);
                const element = yield frame.$(selector);
                const value = yield (yield element.getProperty('value')).jsonValue();
                const textContent = yield (yield element.getProperty('textContent')).jsonValue();
                const title = yield (yield element.getProperty('title')).jsonValue();
                yield element.dispose();
                let text = value ? value : (textContent ? textContent : (title ? title : null));
                if (text !== null) {
                    text = text.trim();
                }
                return text;
            }),
            getHref: (params) => __awaiter(this, void 0, void 0, function* () {
                const [frame, selector] = yield this.getTargetFrameAndSelector(params.selector);
                const element = yield frame.$(selector);
                const href = yield frame.evaluate((element) => {
                    while (element && element.tagName != 'A') {
                        element = element.parentNode;
                    }
                    if (element.tagName == 'A') {
                        if (element.href !== undefined) {
                            return Promise.resolve(element.href);
                        }
                    }
                    return Promise.resolve(null);
                }, element);
                yield element.dispose();
                return href;
            }),
            getElement: (params) => __awaiter(this, void 0, void 0, function* () {
                const [frame, selector] = yield this.getTargetFrameAndSelector(params.selector);
                const element = yield frame.$(selector);
                const html = yield frame.evaluate((element) => {
                    return Promise.resolve(element.outerHTML);
                }, element);
                yield element.dispose();
                return html;
            }),
            getElementScreenshot: () => __awaiter(this, void 0, void 0, function* () {
                // エディタ側で未対応のため未実装
                return '';
            }),
            getAttribute: (params) => __awaiter(this, void 0, void 0, function* () {
                const [frame, selector] = yield this.getTargetFrameAndSelector(params.selector);
                const element = yield frame.$(selector);
                const attributeName = params.attributeName;
                const value = yield frame.evaluate((element, attributeName) => {
                    return Promise.resolve(element.getAttribute(attributeName));
                }, element, attributeName);
                yield element.dispose();
                return value;
            }),
            getAttributes: (params) => __awaiter(this, void 0, void 0, function* () {
                const [frame, selector] = yield this.getTargetFrameAndSelector(params.selector);
                const element = yield frame.$(selector);
                const attributes = yield frame.evaluate((element) => {
                    const attributes = {};
                    const _attributes = element.attributes;
                    for (const index in _attributes) {
                        const attribute = _attributes[index];
                        if (attribute.nodeName !== undefined) {
                            attributes[attribute.nodeName] = attribute.value;
                        }
                    }
                    return Promise.resolve(attributes);
                }, element);
                yield element.dispose();
                return attributes;
            }),
            matchFiles: (params) => __awaiter(this, void 0, void 0, function* () {
                let timedout = false;
                const timeoutTimer = setTimeout(() => {
                    timedout = true;
                }, params.timeout);
                const matchFiles = () => __awaiter(this, void 0, void 0, function* () {
                    const downloadPath = `${this.contractId}/downloads/${this.uuid}`;
                    // ダウンロードファイルリスト取得
                    const files = [];
                    for (let file of yield fs.promises.readdir(`${config.storage.path}/${downloadPath}`)) {
                        let stats;
                        try {
                            stats = yield fs.promises.stat(`${config.storage.path}/${downloadPath}/${file}`);
                        }
                        catch (e) {
                            // readdirからstatまでの間にダウンロード完了した可能性がある
                            file = file.replace(/\.crdownload$/, '');
                            stats = yield fs.promises.stat(`${config.storage.path}/${downloadPath}/${file}`);
                        }
                        if (file != `${this.uuid}.lock`) {
                            files.push({
                                name: file,
                                createdAt: stats.mtimeMs,
                                adopted: false
                            });
                        }
                    }
                    files.sort((a, b) => {
                        return a.createdAt > b.createdAt ? -1 : 1;
                    });
                    // ファイルマッチング
                    const targetFiles = params.targetFiles.map((targetFile) => {
                        return {
                            name: targetFile.name,
                            required: targetFile.required,
                            matched: false
                        };
                    });
                    const adoptedFiles = [];
                    for (let targetFileIndex = 0; targetFileIndex < targetFiles.length; targetFileIndex++) {
                        const targetFile = targetFiles[targetFileIndex];
                        for (let fileIndex = 0; fileIndex < files.length; fileIndex++) {
                            const file = files[fileIndex];
                            file.name = file.name.replace(/\.crdownload$/, '');
                            if (!file.adopted && isMatch(file.name, targetFile.name)) {
                                targetFile.matched = true;
                                file.adopted = true;
                                adoptedFiles.push({
                                    name: file.name
                                });
                                break;
                            }
                        }
                    }
                    // マッチング判定
                    let notMatchedTargetFiles;
                    if (!timedout) {
                        // タイムアウトしていない場合は、すべてマッチしたかを評価
                        notMatchedTargetFiles = targetFiles.filter((targetFile) => {
                            return !targetFile.matched;
                        });
                    }
                    else {
                        // タイムアウトしている場合は、必須ファイルがすべてマッチしたかを評価
                        notMatchedTargetFiles = targetFiles.filter((targetFile) => {
                            return targetFile.required && !targetFile.matched;
                        });
                    }
                    if (notMatchedTargetFiles.length) {
                        throw new Error('No matching file');
                    }
                    return adoptedFiles;
                });
                const isMatch = (str, pattern) => {
                    ['[', '\\', '^', '$', '.', '|', '+', '(', ')'].forEach((symbol) => {
                        const regExp = new RegExp(`\\${symbol}`, 'g');
                        pattern = pattern.replace(regExp, `\\${symbol}`);
                    });
                    pattern = pattern.replace(/\*/g, '.*').replace(/\?/g, '.');
                    const regExp = new RegExp(`^${pattern}$`);
                    return !!str.match(regExp);
                };
                while (true) {
                    try {
                        const files = yield matchFiles();
                        clearTimeout(timeoutTimer);
                        return files;
                    }
                    catch (e) {
                        console.log(e.message);
                        if (e.message == 'No matching file' && !timedout) {
                            yield utils_1.sleep(100);
                        }
                        else {
                            clearTimeout(timeoutTimer);
                            throw e;
                        }
                    }
                }
            }),
            getFiles: (params) => __awaiter(this, void 0, void 0, function* () {
                const downloadPath = `${this.contractId}/downloads/${this.uuid}`;
                const copiedPathWithoutContractId = `copied/${this.uuid}`;
                const copiedPath = `${this.contractId}/${copiedPathWithoutContractId}`;
                const targetFiles = params.targetFiles.map((targetFile) => {
                    return {
                        name: targetFile.name,
                        id: null,
                        size: null,
                        timestamp: null
                    };
                });
                let timedout = false;
                const timeoutTimer = setTimeout(() => {
                    timedout = true;
                }, params.timeout);
                const waitFiles = () => {
                    return new Promise((resolve, reject) => {
                        let remaining = targetFiles.length;
                        for (const targetFile of targetFiles) {
                            if (fs.existsSync(`${config.storage.path}/${downloadPath}/${targetFile.name}`)) {
                                remaining--;
                            }
                        }
                        if (remaining) {
                            reject(new Error('No matching file'));
                        }
                        else {
                            resolve();
                        }
                    });
                };
                while (true) {
                    try {
                        if (timedout) {
                            throw new Error('Wait timed out');
                        }
                        yield waitFiles();
                        clearTimeout(timeoutTimer);
                        break;
                    }
                    catch (e) {
                        console.log(e.message);
                        if (e.message == 'No matching file') {
                            yield utils_1.sleep(100);
                        }
                        else {
                            clearTimeout(timeoutTimer);
                            throw e;
                        }
                    }
                }
                // 採用ファイルをコピーディレクトリに移動
                for (const targetFile of targetFiles) {
                    targetFile.id = nodeUuid.v1();
                    const copiedFullPath = yield utils_1.makeStorageDirAndTryCallback(`${copiedPath}/${targetFile.id}`, (fullPath) => __awaiter(this, void 0, void 0, function* () {
                        yield fs.promises.rename(`${config.storage.path}/${downloadPath}/${targetFile.name}`, `${fullPath}/${targetFile.name}`);
                        return fullPath;
                    }));
                    const stats = yield fs.promises.stat(`${copiedFullPath}/${targetFile.name}`);
                    targetFile.size = stats.size;
                    targetFile.timestamp = Math.floor(stats.mtimeMs);
                    this.emit('copy_files', [{
                            name: targetFile.name,
                            size: stats.size,
                            createdAt: stats.mtimeMs,
                            ino: stats.ino
                        }]);
                }
                return targetFiles.map((targetFile) => {
                    return {
                        meta: {
                            name: targetFile.name,
                            size: targetFile.size,
                            timestamp: targetFile.timestamp
                        },
                        ref: `${copiedPathWithoutContractId}/${targetFile.id}`
                    };
                });
            }),
            download: (params) => __awaiter(this, void 0, void 0, function* () {
                let selector = params.selector;
                let src;
                if (selector) {
                    // ダウンロード対象タグを探して、URLを取得する属性を決定する
                    const selectors = selector.split(' >> ');
                    const parsedSelector = selectors[selectors.length - 1].split(' > ');
                    let attributeName;
                    for (let i = parsedSelector.length - 1; i >= 0; i--) {
                        const matches = parsedSelector[i].match(/^[a-z0-9]+/);
                        if (matches[0] == 'img') {
                            attributeName = 'src';
                            break;
                        }
                        else if (matches[0] == 'a') {
                            attributeName = 'href';
                            parsedSelector.splice(i + 1);
                            break;
                        }
                    }
                    selectors[selectors.length - 1] = parsedSelector.join(' > ');
                    selector = selectors.join(' >> ');
                    // 属性からURLを取得する
                    src = yield this.command({
                        command: 'getAttribute',
                        params: {
                            selector: selector,
                            attributeName: attributeName
                        }
                    });
                }
                else {
                    src = this.currentWindow.page.url();
                }
                // ダウンロード実行
                const cookies = yield this.currentWindow.page.cookies();
                const url = yield this.currentWindow.page.evaluate(() => {
                    return Promise.resolve(document.location.href);
                });
                const downloader = new downloader_1.Downloader(this.uuid, this.contractId, cookies);
                yield downloader.download(src, url, {
                    userAgent: this.meta.userAgent,
                    lang: this.meta.lang,
                    proxy: this.proxy,
                    auth: this.authStore.search(url)
                });
            }),
            mouseMove: (params) => __awaiter(this, void 0, void 0, function* () {
                let left, top;
                if (params.selector) {
                    const position = yield this.commands.getPagePositionFromSelector({ selector: params.selector });
                    left = position.left;
                    top = position.top;
                    left += params.left < position.elementInfo.width ? params.left : position.elementInfo.width;
                    top += params.top < position.elementInfo.height ? params.top : position.elementInfo.height;
                }
                else {
                    left = params.left;
                    top = params.top;
                }
                yield this.currentWindow.page.mouse.move(left - this.currentWindow.scrollLeft, top - this.currentWindow.scrollTop);
                this.currentWindow.mouseLeft = left;
                this.currentWindow.mouseTop = top;
                return {
                    windowNumber: this.currentWindow.number,
                    left: left,
                    top: top
                };
            }),
            getPagePositionFromSelector: (params) => __awaiter(this, void 0, void 0, function* () {
                let left = 0, top = 0;
                const result = yield this.commands.elementFromSelector({ selector: params.selector });
                const elementInfo = result.target;
                const frameStack = [] = result.frameStack;
                frameStack.forEach((frame) => {
                    left += frame.left - frame.scrollLeft;
                    top += frame.top - frame.scrollTop;
                });
                left += elementInfo.left;
                top += elementInfo.top;
                return {
                    left: left,
                    top: top,
                    elementInfo: elementInfo
                };
            }),
            elementFromPoint: (params, frame = null, frameStack = []) => __awaiter(this, void 0, void 0, function* () {
                if (!frame) {
                    frame = this.currentWindow.page.mainFrame();
                }
                const selector = yield frame.evaluate((params, selectorFromElementScript) => {
                    const getSelectorFromElement = eval(selectorFromElementScript);
                    const element = document.elementFromPoint(params.left, params.top);
                    if (params.currentSelector) {
                        let currentElement = null;
                        try {
                            currentElement = document.querySelector(params.currentSelector);
                        }
                        catch (e) {
                        }
                        if (element === currentElement) {
                            return Promise.resolve(params.currentSelector);
                        }
                    }
                    const selector = getSelectorFromElement(element, params.selectorRule == 'id');
                    return Promise.resolve(selector);
                }, params, this.getSelectorFromElementScript());
                // エレメントサイズ・位置取得
                const elementInfo = yield this.getElementInfo(frame, selector);
                if (elementInfo.tag == 'IFRAME' || elementInfo.tag == 'FRAME') {
                    const _frame = yield this.pushFrameInfo(frame, elementInfo, frameStack);
                    params.left = params.left - elementInfo.left + elementInfo.scrollLeft;
                    params.top = params.top - elementInfo.top + elementInfo.scrollTop;
                    return yield this.commands.elementFromPoint(params, _frame, frameStack);
                }
                return {
                    target: elementInfo,
                    frameStack: frameStack,
                };
            }),
            elementFromSelector: (params, frame = null, frameStack = []) => __awaiter(this, void 0, void 0, function* () {
                if (!frame) {
                    frame = this.currentWindow.page.mainFrame();
                }
                const selectors = params.selector.split(' >> ');
                const selector = selectors.shift();
                // エレメントサイズ・位置取得
                const elementInfo = yield this.getElementInfo(frame, selector);
                if ((elementInfo.tag == 'IFRAME' || elementInfo.tag == 'FRAME') && selectors.length) {
                    const _frame = yield this.pushFrameInfo(frame, elementInfo, frameStack);
                    params.selector = selectors.join(' >> ');
                    return yield this.commands.elementFromSelector(params, _frame, frameStack);
                }
                return {
                    target: elementInfo,
                    frameStack: frameStack,
                };
            }),
            elementFromFocus: (params, frame = null, frameStack = []) => __awaiter(this, void 0, void 0, function* () {
                if (!frame) {
                    frame = this.currentWindow.page.mainFrame();
                }
                const selectors = params.selector.split(' >> ');
                const selector = selectors.shift();
                // エレメントサイズ・位置取得
                const elementInfo = yield this.getElementInfo(frame, selector);
                if ((elementInfo.tag == 'IFRAME' || elementInfo.tag == 'FRAME') && selectors.length) {
                    const _frame = yield this.pushFrameInfo(frame, elementInfo, frameStack);
                    params.selector = selectors.join(' >> ');
                    return yield this.commands.elementFromFocus(params, _frame, frameStack);
                }
                yield frame.focus(selector);
                if (params.behavior == 'back') {
                    yield this.currentWindow.page.keyboard.down('Shift');
                }
                yield this.currentWindow.page.keyboard.press('Tab');
                if (params.behavior == 'back') {
                    yield this.currentWindow.page.keyboard.up('Shift');
                }
                const _selector = yield frame.evaluate((params, selectorFromElementScript) => {
                    const getSelectorFromElement = eval(selectorFromElementScript);
                    const selector = getSelectorFromElement(document.activeElement, params.selectorRule == 'id');
                    return Promise.resolve(selector);
                }, params, this.getSelectorFromElementScript());
                const _elementInfo = yield this.getElementInfo(frame, _selector);
                return {
                    target: _elementInfo,
                    frameStack: frameStack,
                };
            }),
            siblingFromSelector: (params) => __awaiter(this, void 0, void 0, function* () {
                const [frame, _selector, padding] = yield this.getTargetFrameAndSelector(params.selector, true);
                const baseSelector = params.selector.replace(_selector, '');
                const selector = this.strictNthOfType(_selector);
                // 兄弟要素のエレメント取得
                const siblings = yield frame.evaluate((baseSelector, selector, padding, selectorFromElementScript) => {
                    const getSelectorFromElement = eval(selectorFromElementScript);
                    const siblingElements = document.querySelectorAll(selector);
                    const siblingRects = [];
                    siblingElements.forEach((siblingElement) => {
                        const siblingRect = siblingElement.getBoundingClientRect();
                        siblingRects.push({
                            selector: baseSelector + getSelectorFromElement(siblingElement),
                            width: siblingRect.width,
                            height: siblingRect.height,
                            left: siblingRect.left + padding.left,
                            top: siblingRect.top + padding.top,
                        });
                    });
                    return Promise.resolve(siblingRects);
                }, baseSelector, selector, padding, this.getSelectorFromElementScript());
                return siblings;
            }),
            siblingSelectorFromGroupSelector: (params) => __awaiter(this, void 0, void 0, function* () {
                const [frame, _groupSelector] = yield this.getTargetFrameAndSelector(params.groupSelector);
                const baseSelector = params.groupSelector.replace(_groupSelector, '');
                const groupSelector = this.strictNthOfType(_groupSelector);
                const siblingSelector = yield frame.evaluate((groupSelector, index, selectorFromElementScript) => {
                    const getSelectorFromElement = eval(selectorFromElementScript);
                    const elements = document.querySelectorAll(groupSelector);
                    const element = elements[index];
                    if (element) {
                        return Promise.resolve(getSelectorFromElement(element));
                    }
                    return Promise.resolve(null);
                }, groupSelector, params.index, this.getSelectorFromElementScript());
                if (siblingSelector) {
                    return this.strictNthOfType(baseSelector + siblingSelector);
                }
                return null;
            }),
            convertSelectorRule: (params) => __awaiter(this, void 0, void 0, function* () {
                const useId = params.rule == 'id' ? true : false;
                const [frame, selector] = yield this.getTargetFrameAndSelector(params.selector);
                const baseSelector = params.selector.replace(selector, '');
                return baseSelector + (yield frame.evaluate((selector, useId, selectorFromElementScript) => {
                    const getSelectorFromElement = eval(selectorFromElementScript);
                    const element = document.querySelector(selector);
                    return Promise.resolve(getSelectorFromElement(element, useId));
                }, selector, useId, this.getSelectorFromElementScript()));
            }),
            getWindowInfo: () => __awaiter(this, void 0, void 0, function* () {
                // await currentWindow.page._waitForPageTransition();
                const scroll = yield this.currentWindow.page.evaluate(() => {
                    return Promise.resolve({
                        screenWidth: window.innerWidth,
                        screenHeight: window.innerHeight,
                        scrollLeft: window.scrollX,
                        scrollTop: window.scrollY,
                    });
                });
                return scroll;
            }),
            getWindowNumber: () => {
                return this.currentWindow.number;
            },
            getScreenshot: () => __awaiter(this, void 0, void 0, function* () {
                // await currentWindow.page._waitForPageTransition();
                if (!(yield this.currentWindow.page.isInTransition())) {
                    let binary = null;
                    if (this.currentWindow.page.url().indexOf('http') === 0) {
                        binary = (yield this.currentWindow.page.screenshot({
                            type: 'png',
                            encoding: 'binary',
                        }));
                    }
                    // スクリーンショット取得後のスクロール位置取得
                    const scroll = yield this.currentWindow.page.evaluate(() => {
                        return Promise.resolve({
                            left: window.scrollX,
                            top: window.scrollY,
                        });
                    });
                    // ハッシュ値取得
                    if (!binary) {
                        throw new Error('buffer is empty');
                    }
                    const binaryHash = utils_1.md5Hash(binary);
                    this.currentWindow.hash = binaryHash;
                    if (!this.screenshotCasting || this.devicePixelRatio != 1) {
                        const image = yield this.buffer2jimp(binary);
                        // スクリーンショットをキャスト
                        if (!this.screenshotCasting) {
                            this.castScreenshot(image, scroll);
                        }
                        // devicePixelRatioが1以外の場合はjimpによりリサイズされたbufferを返却する
                        if (this.devicePixelRatio != 1) {
                            return yield new Promise((resolve, reject) => {
                                image.getBuffer(jimp.MIME_PNG, (err, gridBinary) => __awaiter(this, void 0, void 0, function* () {
                                    resolve(gridBinary);
                                }));
                            });
                        }
                    }
                    // devicePixelRatioが1の場合は無加工のbinaryを返却する
                    return binary;
                }
                return null;
            }),
            startScreenshot: () => {
                this.allowScreenshot = true;
                this.screenshot();
            },
            stopScreenshot: () => {
                clearTimeout(this.screenshotJobTimer);
                this.allowScreenshot = false;
            },
            screenshot: () => __awaiter(this, void 0, void 0, function* () {
                this.screenshot();
            }),
            disableDialogHandling: () => {
                this.dialogHandling = false;
            },
            enableDialogHandling: () => {
                this.dialogHandling = true;
            },
            onceDialogHandling: (params) => {
                const handling = (dialog) => __awaiter(this, void 0, void 0, function* () {
                    this.dialogHandlingFunction = null;
                    this.log('onceDialogHandling');
                    this.log(`command: ${params.command}, value: ${params.value}`);
                    if (params.command == 'accept') {
                        if (params.value === null) {
                            yield dialog.accept();
                        }
                        else {
                            yield dialog.accept(params.value);
                        }
                    }
                    else {
                        yield dialog.dismiss();
                    }
                });
                this.dialogHandlingFunction = handling;
                this.currentWindow.page.once('dialog', handling);
            },
            disableFileChooserHandling: () => {
                this.fileChooserHandling = false;
            },
            enableFileChooserHandling: () => {
                this.fileChooserHandling = true;
            },
            onceFileChooserHandling: (params) => {
                const handling = () => __awaiter(this, void 0, void 0, function* () {
                    try {
                        const fileManager = new fileManager_1.FileManager(params.tokenId, this.contractId);
                        const files = yield fileManager.get(params.files);
                        console.log(files);
                        if (params.command == 'accept') {
                            if (!this.fileChooser.isMultiple()) {
                                // マルチプルではない場合は最初の1ファイルだけ使用する
                                files.splice(1);
                            }
                            yield this.fileChooser.accept(files);
                        }
                        else {
                            yield this.fileChooser.cancel();
                        }
                        this.fileChooser = null;
                        this.fileChooserHandlingFunction = null;
                    }
                    catch (e) {
                        console.log(e.message);
                        this.emit('error', new Error('file chooser handling failed'));
                    }
                });
                this.fileChooserHandlingFunction = handling;
            },
            waitUntilFileChooserHanding: (params) => __awaiter(this, void 0, void 0, function* () {
                let timedout = false;
                const timeoutTimer = setTimeout(() => {
                    timedout = true;
                }, params.timeout);
                while (true) {
                    if (timedout) {
                        throw new Error('Wait timed out');
                    }
                    if (!this.fileChooserHandlingFunction) {
                        break;
                    }
                    yield utils_1.sleep(100);
                }
                clearTimeout(timeoutTimer);
            }),
            waitDialog: (params) => {
                return new Promise((resolve, reject) => {
                    let startTime = moment();
                    const proc = () => {
                        if (this.dialogHandlingFunction) {
                            let currentTime = moment();
                            if (currentTime.diff(startTime) > params.timeout) {
                                this.dialogHandlingFunction = null;
                                reject(new Error('Wait Dialog Timeout.'));
                                return;
                            }
                            else {
                                setTimeout(proc, 100);
                                return;
                            }
                        }
                        resolve();
                    };
                    proc();
                });
            },
            castConnect: () => __awaiter(this, void 0, void 0, function* () {
                this.log('castConnect.');
                this.newCastConnection = true;
            })
        };
        this.uuid = _config.uuid;
        this.contractId = _config.contractId;
        this.meta = _config.meta;
        this.proxy = _config.proxy;
        this.useConnector = _config.useConnector;
        this.allowLoopback = _config.allowLoopback;
    }
    /**
     * ログ出力
     */
    log(message) {
        const dateTime = moment().format('YYYY-MM-DD HH:mm:ss');
        console.log(`[${dateTime}] Browser: (${this.uuid}) ${message}`);
    }
    /**
     * 起動
     */
    start() {
        return __awaiter(this, void 0, void 0, function* () {
            this.log('Start browser.');
            // 使用するchromiumを指定
            let chromium;
            if (process.env.LAMBDA_TASK_ROOT) {
                process.env.HOME = '/opt/';
                chromium = require('chrome-aws-lambda');
            }
            else {
                chromium = {
                    headless: config.browser.headless,
                    args: [],
                    executablePath: config.browser.executablePath ? config.browser.executablePath : require('chromium').path,
                };
                chromium.executablePath = chromium.executablePath.replace('app.asar', 'app.asar.unpacked');
                console.log(chromium);
            }
            // 起動オプション設定
            const params = {
                headless: chromium.headless,
                args: Object.assign([], chromium.args, [
                    `--user-agent=${this.meta.userAgent}`,
                    `--lang=${this.meta.lang}`,
                    `--window-size=${this.meta.screenWidth},${this.meta.screenHeight + (config.browser.headless ? 0 : config.browser.headPanelHeight)}`,
                    '--ignore-certificate-errors'
                ]),
                handleSIGINT: false,
                defaultViewport: {
                    width: this.meta.screenWidth,
                    height: this.meta.screenHeight
                },
                ignoreHTTPSErrors: true,
                executablePath: yield chromium.executablePath,
            };
            if (!config.browser.sandbox) {
                params.args.push('--no-sandbox');
            }
            if (this.useConnector && this.allowLoopback) {
                params.args.push('--host-rules=MAP localhost localhost.agent.local, MAP 127.0.0.1 127.0.0.1.agent.local');
            }
            if (this.proxy.host) {
                params.args.push(`--proxy-server=${this.proxy.host}:${this.proxy.port}`);
            }
            // 起動
            try {
                this.browser = true;
                this.browser = yield puppeteer.launch(params);
                // ダウンロード設定
                let downloadPath = yield utils_1.makeStorageDirAndTryCallback(`${this.contractId}/downloads/${this.uuid}`, (fullPath) => __awaiter(this, void 0, void 0, function* () {
                    fs.closeSync(fs.openSync(`${fullPath}/${this.uuid}.lock`, 'w'));
                    return fullPath;
                }));
                if (process.platform == "win32") {
                    downloadPath = downloadPath.replace(/\//g, '\\');
                }
                const client = yield this.browser.target().createCDPSession();
                client.send('Browser.setDownloadBehavior', {
                    behavior: 'allow',
                    downloadPath: downloadPath
                });
            }
            catch (e) {
                console.log(e);
                this.browser = false;
                throw e;
            }
            // devicePixelRatio取得
            this.devicePixelRatio = yield this.getDevicePixelRatio(this.browser);
            console.log('devicePixelRatio:', this.devicePixelRatio);
            this.emit('update_device_pixel_ratio', this.devicePixelRatio);
            // 新規ウインドウが開いた時
            this.browser.on('targetcreated', (target) => __awaiter(this, void 0, void 0, function* () {
                if (target.type() == 'page') {
                    this.log('browser targetcreated');
                    const page = yield target.page();
                    yield initNewWindow(page);
                    this.switchedCurrentWindow = true;
                    this.windowsDiff = true;
                    this.newPage = true;
                    // ページロード完了まで待機する
                    if (page.url().indexOf('http') === 0) {
                        try {
                            yield page.waitFor(() => {
                                return document.readyState == 'complete';
                            });
                        }
                        catch (e) {
                            this.log(e.message);
                        }
                    }
                    yield addUnloadEvent(page);
                }
            }));
            // 新規ページ初期化
            const initNewWindow = (page) => __awaiter(this, void 0, void 0, function* () {
                yield page.setViewport({
                    width: this.meta.screenWidth,
                    height: this.meta.screenHeight,
                    deviceScaleFactor: this.devicePixelRatio
                });
                // ループバックリクエストを禁止する
                if (!this.allowLoopback) {
                    yield page.setRequestInterception(true);
                    page.on('request', (request) => __awaiter(this, void 0, void 0, function* () {
                        const urlObj = new URL(request.url());
                        if (urlObj.hostname == 'localhost' || urlObj.hostname == '127.0.0.1') {
                            yield request.abort('addressunreachable');
                        }
                        else {
                            yield request.continue();
                        }
                    }));
                }
                // ナビゲートリクエストが失敗した場合は遷移完了とする
                page.on('requestfailed', (request) => __awaiter(this, void 0, void 0, function* () {
                    var _a;
                    if (request.resourceType() == 'document' && request.isNavigationRequest()) {
                        this.log('Navigation request failed.');
                        try {
                            // プロキシ認証エラーの場合
                            if (((_a = request.response()) === null || _a === void 0 ? void 0 : _a.status()) == 407) {
                                this.log('Proxy Authentication Required');
                                yield page.evaluate(() => {
                                    const message = '407 Proxy Authentication Required';
                                    document.write(message);
                                    document.title = message;
                                });
                            }
                            // 遷移完了
                            yield page.evaluate(() => {
                                window['cbot-loaded'] = true;
                            });
                        }
                        catch (e) {
                            console.log(e);
                        }
                    }
                }));
                // ページ遷移中か
                page.isInTransition = () => __awaiter(this, void 0, void 0, function* () {
                    try {
                        // ページ遷移中にevaluateがロックされていることがあるため、1秒経過しても状態取得できない場合は遷移中と判断する
                        const evaluate = page.evaluate(() => {
                            return Promise.resolve(!window['cbot-loaded']);
                        });
                        const timeout = new Promise((resolve, reject) => {
                            setTimeout(() => {
                                resolve(true);
                            }, 1000);
                        });
                        return yield Promise.race([evaluate, timeout]);
                    }
                    catch (e) {
                        // evaluateで例外が発生する場合は遷移中と判断する
                        return true;
                    }
                });
                // ページ遷移待機
                page.waitForPageTransition = () => {
                    return new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
                        try {
                            yield utils_1.sleep(50);
                            yield page._waitForPageTransition();
                            resolve();
                        }
                        catch (e) {
                            reject(e);
                        }
                    }));
                };
                page._waitForPageTransition = () => {
                    return new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
                        let startTime = moment();
                        const timeout = 30000;
                        while (true) {
                            if (!(yield page.isInTransition())) {
                                this.screenshot();
                                break;
                            }
                            let currentTime = moment();
                            if (currentTime.diff(startTime) > timeout) {
                                reject(new Error('waitForPageTransition timed out ' + timeout + 'ms'));
                                break;
                            }
                            yield utils_1.sleep(100);
                        }
                        resolve();
                    }));
                };
                // loadイベントで遷移完了とする
                page.on('load', () => __awaiter(this, void 0, void 0, function* () {
                    this.log('page loaded');
                    yield addUnloadEvent(page);
                    this.newPage = true;
                    this.currentWindow.watchElements = {};
                    this.currentWindow.watchElementsHash = null;
                    this.screenshot();
                    // FileChooser検知とイベント発火
                    // page.waitForFileChooser()では隙間なく検知することができないため、
                    // CDPSessionでfileCHooserOpenedイベントを直接フックしている
                    // ページ遷移でドメインが変わると設定が無効になるためページ遷移毎に再設定する
                    yield page._client.send('Page.setInterceptFileChooserDialog', {
                        enabled: true
                    });
                    page._client.removeAllListeners('Page.fileChooserOpened');
                    page._client.on('Page.fileChooserOpened', (event) => __awaiter(this, void 0, void 0, function* () {
                        const frame = page._frameManager.frame(event.frameId);
                        const context = yield frame.executionContext();
                        const element = yield context._adoptBackendNodeId(event.backendNodeId);
                        const fileChooser = new FileChooser_1.FileChooser(element, event);
                        page.emit('fileChooser', fileChooser);
                    }));
                }));
                // ウインドウを閉じた時のハンドリング
                page.on('close', () => {
                    this.log('page close');
                    let windowNumber = page.windowNumber;
                    if (this.windows[page.windowNumber] == this.currentWindow) {
                        // windowNumberが一番近いウインドウに切り替え
                        const diff = {};
                        for (const _windowNumber in this.windows) {
                            if (parseInt(_windowNumber) == page.windowNumber) {
                                continue;
                            }
                            diff[_windowNumber] = Math.abs(page.windowNumber - parseInt(_windowNumber));
                            windowNumber = (diff[windowNumber] < diff[_windowNumber]) ? windowNumber : parseInt(_windowNumber);
                        }
                        this.currentWindow = this.windows[windowNumber];
                    }
                    delete this.windows[page.windowNumber];
                    this.windowsDiff = true;
                    this.switchedCurrentWindow = true;
                });
                // ダイアログのハンドリング
                page.on('dialog', (dialog) => __awaiter(this, void 0, void 0, function* () {
                    if (this.dialogHandling) {
                        this.log('Notify alert.');
                        const text = dialog.message();
                        let type = dialog.type();
                        this.dialog = dialog;
                        this.emit('dialog', text, type);
                    }
                    else {
                        // execute_bot中に予期せずアラートがあがった場合は強制切断する
                        if (!this.dialogHandlingFunction) {
                            this.emit('error', new Error('unexpected dialog'));
                            return;
                        }
                    }
                }));
                // FileChooserのハンドリング
                page.on('fileChooser', (fileChooser) => __awaiter(this, void 0, void 0, function* () {
                    if (!this.fileChooser) {
                        this.fileChooser = fileChooser;
                        if (this.fileChooserHandling) {
                            this.emit('file_chooser', this.fileChooser.isMultiple());
                        }
                        else {
                            if (this.fileChooserHandlingFunction) {
                                this.fileChooserHandlingFunction();
                            }
                            else {
                                // execute_bot中に予期せずFileChooserがあがった場合は無視する
                                yield this.fileChooser.cancel();
                                this.fileChooser = null;
                            }
                        }
                    }
                }));
                const number = this.getNewWindowNumber();
                page.windowNumber = number;
                this.windows[number] = this.windowObject(number);
                this.windows[number].page = page;
                this.currentWindow = this.windows[number];
            });
            const addUnloadEvent = (page) => __awaiter(this, void 0, void 0, function* () {
                try {
                    yield page.evaluate(() => {
                        window['cbot-loaded'] = true;
                        window.addEventListener('beforeunload', () => {
                            window['cbot-loaded'] = false;
                        });
                    });
                }
                catch (e) {
                    console.log(e);
                }
            });
            // ページオブジェクト格納
            const page = (yield this.browser.pages())[0];
            yield initNewWindow(page);
            yield addUnloadEvent(page);
            // Proxy認証
            if (this.proxy.user) {
                yield page.authenticate({ username: this.proxy.user, password: this.proxy.password });
            }
            // 遷移監視開始
            this.watchTransition();
        });
    }
    /**
     * 停止
     */
    stop() {
        return new Promise((resolve, reject) => {
            this.log('Stop browser.');
            // browserが使用可能になる前に呼ばれた場合に実行待機する
            const proc = () => __awaiter(this, void 0, void 0, function* () {
                if (this.browser === true) {
                    setTimeout(() => {
                        proc();
                    }, 500);
                    return;
                }
                if (this.browser) {
                    yield this.browser.close();
                    this.browser = null;
                }
                clearTimeout(this.watchTransitionTimer);
                clearTimeout(this.screenshotJobTimer);
                this.screenshotCasting = false;
                this.removeAllListeners();
                resolve();
            });
            proc();
        });
    }
    /**
     * devicePixelRatio取得
     */
    getDevicePixelRatio(browser) {
        return __awaiter(this, void 0, void 0, function* () {
            // window.devicePixelRatioはEmulationによってdeviceScaleFactorと同じになる
            // CDPメソッドから直接スクリーンショットを取得することでEmulationの影響を受けていないイメージを取り出し、
            // サイズを測定することでdevicePixelRatioの値とする
            const page = (yield browser.pages())[0];
            const viewport = page.viewport();
            yield page.setViewport({
                width: 1,
                height: 1
            });
            const client = yield page.target().createCDPSession();
            const result = yield client.send('Page.captureScreenshot', {
                format: 'png',
                quality: 0,
                clip: {
                    x: 0,
                    y: 0,
                    width: 1,
                    height: 1,
                    scale: 1
                }
            });
            yield page.setViewport(viewport);
            const buffer = Buffer.from(result.data, 'base64');
            const image = yield jimp.read(buffer);
            return image.getWidth();
        });
    }
    /**
     * 新しいウインドウ番号取得
     */
    getNewWindowNumber() {
        let number = 0;
        for (const _number in this.windows) {
            if (number < parseInt(_number)) {
                number = parseInt(_number);
            }
        }
        number++;
        return number;
    }
    ;
    /**
     * 新規ページオブジェクト
     */
    windowObject(number) {
        return {
            number: number,
            page: null,
            url: null,
            title: '新しいタブ',
            scrollLeft: 0,
            scrollTop: 0,
            documentWidth: this.meta.screenWidth,
            documentHeight: this.meta.screenHeight,
            hash: null,
            hashes: {},
            flush: false,
            latestScrollLeft: null,
            latestScrollTop: null,
            watchElements: {},
            watchElementsHash: null,
            mouseLeft: null,
            mouseTop: null
        };
    }
    /**
     * モニタリングを有効にする
     */
    enableMonitoring() {
        this.monitoringEnabled = true;
    }
    /**
     * モニタリングを無効にする
     */
    disableMonitoring() {
        this.monitoringEnabled = false;
    }
    /**
     * 遷移監視
     */
    watchTransition() {
        return __awaiter(this, void 0, void 0, function* () {
            const _switchedCurrentWindow = this.switchedCurrentWindow;
            const _newCastConnection = this.newCastConnection;
            const _windowsDiff = this.windowsDiff;
            const _newPage = this.newPage;
            try {
                // モニタリング有効でなければスキップ
                if (!this.monitoringEnabled) {
                    return;
                }
                this.switchedCurrentWindow = false;
                this.newCastConnection = false;
                this.windowsDiff = false;
                this.newPage = false;
                // windowsに変化があるまたは新規キャスト先がいる場合はウインドウリスト送信
                if (_windowsDiff || _newCastConnection) {
                    this.log('Notify windows.');
                    const windows = {};
                    for (const number in this.windows) {
                        windows[number] = { title: this.windows[number].title };
                    }
                    yield this.emitAndWaitForAck('windows', windows);
                }
                // ウインドウ状態を取得
                const windowNumber = this.currentWindow.number;
                const windowState = yield this.currentWindow.page.evaluate(() => {
                    const documentWidth = Math.max.apply(null, [
                        document.body.clientWidth,
                        document.body.scrollWidth,
                        document.documentElement.scrollWidth,
                        document.documentElement.clientWidth,
                    ]);
                    const documentHeight = Math.max.apply(null, [
                        document.body.clientHeight,
                        document.body.scrollHeight,
                        document.documentElement.scrollHeight,
                        document.documentElement.clientHeight,
                    ]);
                    let title = document.title;
                    const url = document.location.href;
                    if (url == 'chrome-error://chromewebdata/') {
                        title = 'エラー';
                    }
                    if (url == 'about:blank' || url == 'data:,') {
                        title = '新しいタブ';
                    }
                    return Promise.resolve({
                        title: title,
                        url: url,
                        documentWidth: documentWidth,
                        documentHeight: documentHeight,
                        scrollLeft: window.scrollX,
                        scrollTop: window.scrollY,
                        mouseLeft: null,
                        mouseTop: null
                    });
                });
                windowState.windowNumber = windowNumber;
                windowState.newPage = _newPage;
                windowState.timestamp = moment().valueOf();
                // ページ遷移の場合は画像ハッシュと最新スクロール位置、マウス、エレメント監視をクリア
                if (windowState.newPage) {
                    this.currentWindow.flush = true;
                    this.currentWindow.latestScrollLeft = null;
                    this.currentWindow.latestScrollTop = null;
                    this.currentWindow.mouseLeft = null;
                    this.currentWindow.mouseTop = null;
                    // currentWindow.watchElements = {}; // ページ遷移完了後に初期化するように変更した
                    // currentWindow.watchElementsHash = null;
                    if (!this.allowScreenshot) {
                        windowState.newPage = false;
                        this.currentWindow.flush = false;
                    }
                }
                // マウス情報
                windowState.mouseLeft = this.currentWindow.mouseLeft;
                windowState.mouseTop = this.currentWindow.mouseTop;
                // スクロール中か判定
                let scrolling = false;
                if (this.currentWindow.latestScrollLeft != null && this.currentWindow.latestScrollTop != null) {
                    if (this.currentWindow.latestScrollLeft != windowState.scrollLeft || this.currentWindow.latestScrollTop != windowState.scrollTop) {
                        // スクロール中と判断
                        scrolling = true;
                    }
                }
                this.currentWindow.latestScrollLeft = windowState.scrollLeft;
                this.currentWindow.latestScrollTop = windowState.scrollTop;
                // モニタリング有効かつスクロール中でなければ遷移通知処理
                if (this.monitoringEnabled && !scrolling) {
                    const isInTransition = yield this.currentWindow.page.isInTransition();
                    // ここまでの処理で遷移を検知した場合はクライアントに遷移通知
                    if ((isInTransition ? false : windowState.url != this.currentWindow.url)
                        || (isInTransition ? false : windowState.title != this.currentWindow.title)
                        || (isInTransition ? false : windowState.scrollLeft != this.currentWindow.scrollLeft)
                        || (isInTransition ? false : windowState.scrollTop != this.currentWindow.scrollTop)
                        || (isInTransition ? false : windowState.documentWidth != this.currentWindow.documentWidth)
                        || (isInTransition ? false : windowState.documentHeight != this.currentWindow.documentHeight)
                        || windowState.newPage || _switchedCurrentWindow || _newCastConnection || this.scrolled || this.scrollIntoView) {
                        windowState.excludeOperator = this.scrolled;
                        // 遷移ステータスを送信
                        this.log('Notify transition.');
                        yield this.emitAndWaitForAck('transition', windowState);
                        // カレント情報を更新
                        this.currentWindow.url = windowState.url;
                        this.currentWindow.title = windowState.title;
                        this.currentWindow.documentWidth = windowState.documentWidth;
                        this.currentWindow.documentHeight = windowState.documentHeight;
                        this.currentWindow.scrollLeft = windowState.scrollLeft;
                        this.currentWindow.scrollTop = windowState.scrollTop;
                        if (this.scrolled) {
                            this.scrolled = false;
                        }
                        if (this.scrollIntoView) {
                            this.scrollIntoView = false;
                        }
                    }
                }
                // 新規キャスト先がいる場合はすべての画像ハッシュをクリア
                if (_newCastConnection) {
                    for (const number in this.windows) {
                        this.windows[number].flush = true; // すべてのハッシュを破棄
                    }
                }
                // ウインドウが切り替わっているまたは新規キャスト先がいる場合はスクリーンショット送信
                if (_switchedCurrentWindow || _newCastConnection || windowState.newPage) {
                    this.screenshot();
                }
            }
            catch (e) {
                this.log(`WatchTransition Error. [${e.message}]`);
                this.switchedCurrentWindow = _switchedCurrentWindow;
                this.newCastConnection = _newCastConnection;
                this.windowsDiff = _windowsDiff;
                this.newPage = _newPage;
            }
            finally {
                if (this.browser !== null) {
                    this.watchTransitionTimer = setTimeout(() => {
                        this.watchTransition();
                    }, 100);
                }
            }
        });
    }
    /**
     * スクリーンショット
     */
    screenshot() {
        return __awaiter(this, void 0, void 0, function* () {
            clearTimeout(this.screenshotJobTimer);
            if (!this.allowScreenshot) {
                this.log('Deny screenshot.');
                return;
            }
            // ジョブID発行
            const jobId = new Date().getTime();
            this.interrupt.screenshot = jobId;
            let captureSamplingInterval = config.browser.screenshot.captureSamplingInterval;
            try {
                // // モニタリング有効でなければスキップ
                if (!this.monitoringEnabled) {
                    return;
                }
                if (this.screenshotCasting) {
                    this.log('screenshotCasting');
                    return;
                }
                // ページ遷移中の場合はスキップ
                if (yield this.currentWindow.page.isInTransition()) {
                    return;
                }
                if (this.currentWindow.flush) {
                    this.currentWindow.hash = null;
                    this.currentWindow.hashes = {};
                    this.currentWindow.flush = false;
                }
                // スクリーンショット取得前のスクロール位置取得
                const scrollBeforeScreenshot = yield this.currentWindow.page.evaluate(() => {
                    return Promise.resolve({
                        left: window.scrollX,
                        top: window.scrollY,
                    });
                });
                // 割込がなければスクリーンショット取得
                if (jobId == this.interrupt.screenshot) {
                    // スクリーンショット取得
                    const binary = yield this.currentWindow.page.screenshot({
                        type: 'png',
                        encoding: 'binary',
                    });
                    // スクリーンショット取得後のスクロール位置取得
                    const scrollAfterScreenshot = yield this.currentWindow.page.evaluate(() => {
                        return Promise.resolve({
                            left: window.scrollX,
                            top: window.scrollY,
                        });
                    });
                    // スクリーンショット取得前後でスクロール位置に変化がない場合
                    if (scrollBeforeScreenshot.left == scrollAfterScreenshot.left && scrollBeforeScreenshot.top == scrollAfterScreenshot.top) {
                        // ハッシュ値比較して変わっていればキャスト
                        const binaryHash = utils_1.md5Hash(binary);
                        if (this.currentWindow.hash != binaryHash && !this.screenshotCasting) {
                            this.currentWindow.hash = binaryHash;
                            const image = yield this.buffer2jimp(binary);
                            this.castScreenshot(image, scrollAfterScreenshot);
                        }
                        // 監視エレメント情報を取得
                        const watchElements = {};
                        for (const _selector in this.currentWindow.watchElements) {
                            const rect = yield this.getElementRect(_selector);
                            if (rect) {
                                watchElements[_selector] = Object.assign(rect, { tags: this.currentWindow.watchElements[_selector] });
                            }
                            else {
                                delete this.currentWindow.watchElements[_selector];
                            }
                        }
                        const watchElementsHash = utils_1.md5Hash(JSON.stringify(watchElements));
                        if (watchElementsHash != this.currentWindow.watchElementsHash) {
                            yield this.emitAndWaitForAck('watchElements', {
                                windowNumber: this.currentWindow.number,
                                elements: watchElements,
                            });
                            this.currentWindow.watchElementsHash = watchElementsHash;
                        }
                    }
                    // スクロール位置が変化している場合はスクロール中のため、スクリーンショットと位置情報がずれていて使えない
                    // 次のスクリーンショット実行タイミングを早める
                    else {
                        this.log('Detect scrolling.');
                        captureSamplingInterval /= 3;
                    }
                }
            }
            catch (e) {
                this.log(`Screenshot error. [${e.message}]`);
            }
            finally {
                // 割込がなければタイマー設定
                if (jobId == this.interrupt.screenshot) {
                    if (this.browser !== null) {
                        this.screenshotJobTimer = setTimeout(() => {
                            this.log('Screenshot timer.');
                            this.screenshot();
                        }, captureSamplingInterval);
                    }
                }
                else {
                    this.log('Interrupt screenshot.');
                }
            }
        });
    }
    /**
     * スクリーンショットをキャスト
     */
    castScreenshot(_image, scroll) {
        return __awaiter(this, void 0, void 0, function* () {
            this.screenshotCasting = true;
            const gridWidth = this.meta.screenWidth;
            const gridHeight = config.browser.screenshot.gridHeight;
            const screenWidth = this.meta.screenWidth;
            const screenHeight = this.meta.screenHeight;
            // 切り出し位置の補正(ページ最下部までスクロールした際にスクロール位置かグリッド高さの倍数でない場合がある)
            let _y = -(scroll.top % gridHeight);
            // グリッド画像に分割
            const promises = [];
            for (let y = _y; y < screenHeight; y += gridHeight) {
                for (let x = 0; x < screenWidth; x += gridWidth) {
                    // グリッドサイズにクロップ
                    let image = _image.clone();
                    let width = gridWidth;
                    let height = gridHeight;
                    if (screenWidth - x < gridWidth) {
                        width = Math.floor(screenWidth - x);
                    }
                    if (screenHeight - y < gridHeight) {
                        height = Math.floor(screenHeight - y);
                    }
                    image = image.crop(x, y, width, height);
                    // グリッド位置算出
                    const gridLeft = x + scroll.left;
                    const gridTop = y + scroll.top;
                    // バッファ化
                    const gridBinary = yield new Promise((resolve, reject) => {
                        image.getBuffer(jimp.MIME_PNG, (err, gridBinary) => __awaiter(this, void 0, void 0, function* () {
                            resolve(gridBinary);
                        }));
                    });
                    // 画像ハッシュ算出
                    const hash = utils_1.md5Hash(gridBinary);
                    // 画像が新規または変更されていたらキャスト
                    if (!this.windows[this.currentWindow.number].hashes[`grid-${gridLeft}-${gridTop}`] || this.windows[this.currentWindow.number].hashes[`grid-${gridLeft}-${gridTop}`] != hash) {
                        promises.push(new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
                            this.windows[this.currentWindow.number].hashes[`grid-${gridLeft}-${gridTop}`] = hash;
                            this.log(`Cast grid image: 'grid-${gridLeft}-${gridTop}'`);
                            try {
                                const grid = {
                                    windowNumber: this.currentWindow.number,
                                    width: width,
                                    height: height,
                                    left: gridLeft,
                                    top: gridTop,
                                    binary: gridBinary
                                };
                                yield this.emitAndWaitForAck('screenshot', grid);
                                resolve();
                            }
                            catch (e) {
                                reject(e);
                            }
                        })));
                    }
                }
            }
            try {
                yield Promise.all(promises);
            }
            catch (e) {
                console.log(e.message);
            }
            this.screenshotCasting = false;
        });
    }
    /**
     * bufferからjimpに変換
     */
    buffer2jimp(buffer) {
        return __awaiter(this, void 0, void 0, function* () {
            const screenWidth = this.meta.screenWidth;
            const screenHeight = this.meta.screenHeight;
            return yield new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
                const image = yield jimp.read(buffer);
                if (this.devicePixelRatio == 1) {
                    resolve(image);
                }
                else {
                    image.resize(screenWidth, screenHeight, (err, image) => {
                        resolve(image);
                    });
                }
            }));
        });
    }
    /**
     * getSelectorFromElementScript
     */
    getSelectorFromElementScript() {
        return `
            (el, useId=false, strictNthOfType=false) => {
                const getSiblingElemetsIndex = (el,name) => {
                    let index = 1;
                    let sib = el;
                    while((sib = sib.previousElementSibling)){
                        let sibName = escape(sib.nodeName.toLowerCase());
                        if(sibName == 'input'){
                            if(el.getAttribute("type") !== null){
                                sibName += '[type="' + el.getAttribute("type") + '"]';
                            }
                            if(el.name !== ''){
                                sibName += '[name="' + el.name + '"]';
                            }
                        }
                        if(sibName === name){
                            ++index;
                        }
                    }
                    return index;
                }

                const escape = (selector) => {
                    return selector.replace(/[ !"#$%&'()*+,.\/:;<=>?@\\[\\\\\\]^\`{|}~]/g,'\\\\$&');
                }

                const names = [];
                if(!(el instanceof Element)){
                    return names;
                }
                while(el.nodeType === Node.ELEMENT_NODE){
                    let name = escape(el.nodeName.toLowerCase());
                    if(name == 'input'){
                        if(el.getAttribute("type") !== null){
                            name += '[type="' + el.getAttribute("type") + '"]';
                        }
                        if(el.name !== ''){
                            name += '[name="' + el.name + '"]';
                        }
                    }
                    if(el.hasAttribute('cbot-force-selector-rule')) {
                        useId = el.getAttribute('cbot-force-selector-rule') == 'id';
                    }
                    if(useId){
                        if(el.id) {
                            let id = el.id;
                            id = id.replace(/([:;])/g,'\\\\\\\\$1');
                            id = id.replace(/^([0-9]+)/,'\\\\\\\\00003$1');
                            name += '#' + id;
                            names.unshift(name);
                            break;
                        }
                    }

                    const index = getSiblingElemetsIndex(el,name);
                    if(strictNthOfType || 1 < index){
                        name += ':nth-of-type(' + index + ')';
                    }

                    names.unshift(name);
                    el = el.parentNode;
                }
                return names.join(' > ');
            }
        `;
    }
    /**
     * セレクターの厳格化
     */
    strictNthOfType(selector) {
        const selectors = selector.split(' >> ');
        if (selectors.length > 1) {
            for (const index in selectors) {
                selectors[index] = this.strictNthOfType(selectors[index]);
            }
            return selectors.join(' >> ');
        }
        else {
            const elements = selector.split(' > ');
            const parser = new css_selector_parser_1.CssSelectorParser;
            for (const index in elements) {
                if (elements[index]) {
                    const parsedElement = parser.parse(elements[index]);
                    if (parsedElement.type != 'ruleSet' || !parsedElement.rule.id) {
                        if (!elements[index].match(/:nth-of-type\(.+\)$/)) {
                            elements[index] += ':nth-of-type(1)';
                        }
                    }
                }
            }
            return elements.join(' > ');
        }
    }
    /**
     * 要素rect情報取得
     */
    getElementRect(_selector) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const [frame, selector, padding] = yield this.getTargetFrameAndSelector(_selector, true);
                const rect = yield frame.evaluate((selector, padding) => {
                    const element = document.querySelector(selector);
                    let rect = null;
                    if (element) {
                        rect = element.getBoundingClientRect();
                        return Promise.resolve({
                            width: rect.width,
                            height: rect.height,
                            left: rect.left + padding.left,
                            top: rect.top + padding.top,
                        });
                    }
                    return Promise.resolve(null);
                }, selector, padding);
                return rect;
            }
            catch (e) {
                this.log(`getElementRect failed. ${e.message}`);
                return null;
            }
        });
    }
    /**
     * セレクターから操作対象のフレームを取得
     */
    getTargetFrameAndSelector(selector, withPadding = false) {
        return __awaiter(this, void 0, void 0, function* () {
            let frame = this.currentWindow.page.mainFrame();
            const selectors = selector.split(' >> ');
            let targetSelector = selectors.pop();
            let padding = { left: 0, top: 0 };
            for (const index in selectors) {
                const selector = selectors[index];
                const iframe = yield frame.$(selector);
                if (withPadding) {
                    const _padding = yield iframe.evaluate((iframe) => {
                        const rect = iframe.getBoundingClientRect();
                        return Promise.resolve({
                            left: rect.left,
                            top: rect.top,
                        });
                    }, iframe);
                    padding.left += _padding.left;
                    padding.top += _padding.top;
                }
                frame = yield iframe.contentFrame();
                yield iframe.dispose();
            }
            if (withPadding) {
                const scroll = yield this.currentWindow.page.evaluate(() => {
                    return Promise.resolve({
                        x: window.scrollX,
                        y: window.scrollY,
                    });
                });
                padding.left += scroll.x;
                padding.top += scroll.y;
                return [frame, targetSelector, padding];
            }
            return [frame, targetSelector];
        });
    }
    /**
     * ブラウザ状態をリフレッシュ
     */
    refresh() {
        this.currentWindow.hash = null;
        this.currentWindow.hashes = {};
        this.newCastConnection = true;
    }
    /**
     * エレメント監視登録
     */
    registWatchElement(data) {
        data.selector = this.strictNthOfType(data.selector);
        if (!this.windows[data.windowNumber].watchElements[data.selector]) {
            this.windows[data.windowNumber].watchElements[data.selector] = [];
        }
        if (data.tag) {
            this.windows[data.windowNumber].watchElements[data.selector].push(data.tag);
        }
    }
    /**
     * エレメント監視
     */
    watchElements(data) {
        return __awaiter(this, void 0, void 0, function* () {
            const _watchElements = JSON.parse(JSON.stringify(this.currentWindow.watchElements));
            // 追加監視を設定
            if (data.append.selector) {
                data.append.selector = this.strictNthOfType(data.append.selector);
                if (!_watchElements[data.append.selector]) {
                    _watchElements[data.append.selector] = [];
                }
                if (data.append.tag) {
                    _watchElements[data.append.selector].push(data.append.tag);
                }
            }
            const watchElements = [];
            for (const selector in _watchElements) {
                const rect = yield this.getElementRect(selector);
                if (rect) {
                    watchElements.push(Object.assign(rect, { selector: selector, tags: _watchElements[selector] }));
                    if (data.append.selector == selector) {
                        watchElements[watchElements.length - 1]['target'] = true;
                    }
                }
            }
            return {
                windowNumber: this.currentWindow.number,
                elements: watchElements
            };
        });
    }
    /**
     * アラート操作
     */
    alertCommand(data) {
        return __awaiter(this, void 0, void 0, function* () {
            if (data.command == 'accept') {
                if (this.dialog.type() == 'prompt') {
                    yield this.dialog.accept(data.value);
                }
                else {
                    yield this.dialog.accept();
                }
            }
            else {
                yield this.dialog.dismiss();
            }
            this.dialog = null;
        });
    }
    /**
     * FileChooser操作
     */
    fileChooserCommand(fileManager, targets) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const files = yield fileManager.get(targets);
                if (!this.fileChooser.isMultiple()) {
                    // マルチプルではない場合は最初の1ファイルだけ使用する
                    files.splice(1);
                }
                yield this.fileChooser.accept(files);
                try {
                    yield this.emitAndWaitForAck('file_chooser_ended', {
                        result: 0
                    });
                }
                catch (e) {
                    console.log(e.message);
                }
            }
            catch (e) {
                console.log(e.message);
                try {
                    yield this.fileChooser.cancel();
                    yield this.emitAndWaitForAck('file_chooser_ended', {
                        result: -1,
                        message: e.message
                    });
                }
                catch (e) {
                    console.log(e.message);
                }
            }
            this.fileChooser = null;
        });
    }
    /**
     * FileChooserキャンセル
     */
    fileChooserCancel() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.fileChooser.cancel();
            this.fileChooser = null;
        });
    }
    /**
     * コマンド実行
     */
    command(data) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this.commands[data.command](data.params);
        });
    }
    /**
     * エレメント情報取得
     */
    getElementInfo(frame, selector) {
        return __awaiter(this, void 0, void 0, function* () {
            // エレメントサイズ・位置取得
            const elementInfo = yield frame.evaluate((selector) => {
                const element = document.querySelector(selector);
                const rect = element.getBoundingClientRect();
                return Promise.resolve({
                    selector: selector,
                    width: rect.width,
                    height: rect.height,
                    left: rect.left + window.scrollX,
                    top: rect.top + window.scrollY,
                    scrollLeft: window.scrollX,
                    scrollTop: window.scrollY,
                    tag: element.tagName,
                    type: element.type,
                    value: element.value,
                    placeholder: element.placeholder,
                    title: element.title,
                    text: element.innerText,
                    disabled: element.disabled,
                    readOnly: element.readOnly
                });
            }, selector);
            return elementInfo;
        });
    }
    /**
     * フレームスタック
     */
    pushFrameInfo(frame, elementInfo, frameStack) {
        return __awaiter(this, void 0, void 0, function* () {
            const iframe = yield frame.$(elementInfo.selector);
            const _frame = yield iframe.contentFrame();
            yield iframe.dispose();
            const frameInfo = yield _frame.evaluate(() => {
                return Promise.resolve({
                    scrollLeft: window.scrollX,
                    scrollTop: window.scrollY,
                });
            });
            frameStack.push({
                selector: elementInfo.selector,
                width: elementInfo.width,
                height: elementInfo.height,
                left: elementInfo.left,
                top: elementInfo.top,
                scrollLeft: frameInfo.scrollLeft,
                scrollTop: frameInfo.scrollTop,
            });
            return _frame;
        });
    }
    /**
     * ack待ちemit
     */
    emitAndWaitForAck(event, ...args) {
        return __awaiter(this, void 0, void 0, function* () {
            return new Promise((resolve, reject) => {
                if (this.listenerCount(event)) {
                    this.emit(event, ...args, (err, ...data) => {
                        if (err) {
                            reject(err);
                        }
                        else {
                            resolve(data);
                        }
                    });
                }
                else {
                    reject(new Error('listener is undefined'));
                }
            });
        });
    }
}
exports.Browser = Browser;
