"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Client = void 0;
const config = require("config");
const fs = require("fs");
const moment = require("moment");
const redis = require("redis");
const socketIoStream = require("socket.io-stream");
const httpClient_1 = require("../../httpClient");
const utils_1 = require("../../utils");
const browser_1 = require("./browser");
const connector_1 = require("./connector");
const fileManager_1 = require("./fileManager");
const downloadWatcher_1 = require("./downloadWatcher");
class Client {
    constructor(mode, eventEmitter) {
        this.isReady = false; // 通信可能状態か
        this.connector = null;
        this.storageUsageSize = {};
        this.destructed = false;
        this.mode = mode;
        this.eventEmitter = eventEmitter;
    }
    /**
     * ログ出力
     */
    log(message) {
        const dateTime = moment().format('YYYY-MM-DD HH:mm:ss');
        console.log(`[${dateTime}] BrowserClient: (${this.event.uuid}) ${message}`);
    }
    /**
     * UUID取得
     */
    getUuid() {
        var _a;
        return (_a = this.event) === null || _a === void 0 ? void 0 : _a.uuid;
    }
    /**
     * 起動処理
     */
    start(event) {
        return __awaiter(this, void 0, void 0, function* () {
            this.event = event;
            this.log('start.');
            yield this.startBrowserNotification();
            try {
                yield this.connect();
            }
            catch (e) {
                this.log(e.message);
                throw e;
            }
        });
    }
    /**
     * 終了処理
     */
    stop() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.destructor();
            yield this.disconnect();
        });
    }
    /**
     * ブラウザ準備完了
     */
    ready() {
        this.isReady = true;
        this.browser.enableMonitoring();
    }
    /**
     * ブラウザ準備未完了
     */
    unready() {
        var _a;
        this.isReady = false;
        (_a = this.browser) === null || _a === void 0 ? void 0 : _a.disableMonitoring();
    }
    /**
     * ブラウザ起動通知
     */
    startBrowserNotification() {
        return new Promise((resolve, reject) => {
            if (this.eventEmitter) {
                // ブラウザ起動通知
                this.eventEmitter.emit('start_browser', this, () => {
                    resolve();
                });
            }
            else {
                resolve();
            }
        });
    }
    /**
     * ブラウザ停止通知
     */
    stopBrowserNotification() {
        return new Promise((resolve, reject) => {
            if (this.eventEmitter) {
                // ブラウザ停止通知
                this.eventEmitter.emit('stop_browser', this, () => {
                    resolve();
                });
            }
            else {
                resolve();
            }
        });
    }
    /**
     * 切断
     */
    disconnect() {
        return __awaiter(this, void 0, void 0, function* () {
            clearTimeout(this.reconnectTimer);
            clearTimeout(this.virtualConnectorReconnectTimer);
            try {
                yield this.requestDisconnect();
            }
            catch (e) {
                this.log(e.message);
            }
            this.reconnectId = null;
        });
    }
    /**
     * 切断要求
     */
    requestDisconnect() {
        return new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
            try {
                yield this.socket.emitWithAckTimeout('request_disconnect', () => {
                    resolve();
                });
            }
            catch (e) {
                reject(e);
            }
        }));
    }
    /**
     * BrowserControlServer接続
     */
    connect() {
        return new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
            try {
                const browserControlServerEndpoint = `${this.event.browserControlServer.endpoint}:${this.event.browserControlServer.port}${this.event.browserControlServer.path ? this.event.browserControlServer.path : ''}`;
                /**
                 * オプション設定
                 */
                const options = {
                    reconnection: false,
                    forceNew: true,
                    query: {
                        relay_id: this.event.browserControlServer.relayId,
                    },
                    timeout: 10000
                };
                // この接続はrelayIdによって経路が明確になっているのでhttps→wssの段階接続で問題ないためコメントアウト。
                // 拡張ヘッダを付与する必要があるためhttpsアクセスが必須。
                // if(config.webSocket.transports_ws_only) {
                //     options.transports = ["websocket"];
                // }
                options.agent = yield httpClient_1.default.getHttpAgentForSocketIo(`${this.event.browserControlServer.endpoint}:${this.event.browserControlServer.port}/socket.io/socket.io.js?relay_id=${this.event.browserControlServer.relayId}`);
                /**
                 * BrowserControlServerに接続開始
                 */
                this.socket = utils_1.startClientSocket(browserControlServerEndpoint, options);
                /**
                 * 接続
                 */
                this.socket.on('connect', () => __awaiter(this, void 0, void 0, function* () {
                    this.log('Connected');
                    // BrowserControlServerに初期化通知
                    try {
                        yield this.socket.waitUntilCanPingPong();
                        yield this.socket.emitWithAckTimeout('initialize', {
                            uuid: this.event.uuid
                        }, (data) => __awaiter(this, void 0, void 0, function* () {
                            let exception = null;
                            try {
                                if (data.result !== 0) {
                                    throw data;
                                }
                                let useConnector = false;
                                let allowLoopback = !this.event.proxy.host;
                                if (data.connectorInfo) {
                                    useConnector = true;
                                    allowLoopback = !data.connectorInfo.proxy.host;
                                    this.connector = new connector_1.Connector();
                                    this.connector.on('reconnect', () => __awaiter(this, void 0, void 0, function* () {
                                        if (!this.virtualConnectorDisconnectedTimestamp) {
                                            this.virtualConnectorDisconnectedTimestamp = moment().valueOf();
                                            yield this.attemptReconnectVirtualConnector();
                                        }
                                    }));
                                    this.connector.on('disconnect', () => __awaiter(this, void 0, void 0, function* () {
                                        // コネクタが切断した場合は停止する
                                        if (!this.destructed) {
                                            yield this.stop();
                                        }
                                    }));
                                    const port = yield this.connector.connect(data.connectorInfo);
                                    // コネクタを使う場合は、フォワーディングポートをproxyとして使う
                                    this.event.proxy.host = '127.0.0.1';
                                    this.event.proxy.port = port;
                                }
                                this.browser = new browser_1.Browser({
                                    uuid: this.event.uuid,
                                    contractId: this.event.contractId,
                                    meta: this.event.meta,
                                    proxy: this.event.proxy,
                                    useConnector: useConnector,
                                    allowLoopback: allowLoopback
                                });
                                this.setBrowserEvents();
                                this.ready();
                                yield this.browser.start();
                            }
                            catch (e) {
                                exception = e;
                                return;
                            }
                            finally {
                                // pubsub接続情報がある場合はpublish
                                if (this.event.pubsubChannel) {
                                    const pub = redis.createClient(this.event.pubsubPort, this.event.pubsubHost);
                                    const response = exception ? { message: exception.message } : { result: data.result, message: data.message };
                                    yield new Promise((resolve, reject) => {
                                        pub.publish(this.event.pubsubChannel, JSON.stringify(response), () => {
                                            pub.quit();
                                            resolve();
                                        });
                                    });
                                }
                                if (exception) {
                                    reject(exception);
                                }
                            }
                            this.log(`BrowserControlServer Connected.`);
                            this.handleDownloadWatcher();
                            if (this.mode == 'instance') {
                                resolve();
                            }
                        }));
                    }
                    catch (e) {
                        this.log(e.message);
                        reject(e);
                    }
                }));
                this.setEvents(resolve, reject);
            }
            catch (e) {
                reject(e);
            }
        }));
    }
    /**
     * 再接続
     */
    reconnect(_resolve) {
        return new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
            var _a;
            try {
                const browserControlServerEndpoint = `${this.event.browserControlServer.endpoint}:${this.event.browserControlServer.port}${this.event.browserControlServer.path ? this.event.browserControlServer.path : ''}`;
                /**
                 * オプション設定
                 */
                const options = {
                    reconnection: false,
                    forceNew: true,
                    query: {
                        relay_id: this.event.browserControlServer.relayId,
                    },
                    timeout: 10000,
                    transportOptions: {
                        polling: {
                            extraHeaders: {
                                'x-reconnect-id': this.reconnectId,
                                'x-connector-connected': !!((_a = this.connector) === null || _a === void 0 ? void 0 : _a.connected)
                            }
                        }
                    }
                };
                // この接続はrelayIdによって経路が明確になっているのでhttps→wssの段階接続で問題ないためコメントアウト。
                // 拡張ヘッダを付与する必要があるためhttpsアクセスが必須。
                // if(config.webSocket.transports_ws_only) {
                //     options.transports = ["websocket"];
                // }
                options.agent = yield httpClient_1.default.getHttpAgentForSocketIo(`${this.event.browserControlServer.endpoint}:${this.event.browserControlServer.port}/socket.io/socket.io.js?relay_id=${this.event.browserControlServer.relayId}`);
                /**
                 * BrowserControlServerに接続開始
                 */
                this.socket = utils_1.startClientSocket(browserControlServerEndpoint, options);
                /**
                 * 接続
                 */
                this.socket.on('connect', () => __awaiter(this, void 0, void 0, function* () {
                    this.log('Reconnected');
                    try {
                        yield this.socket.waitUntilCanPingPong();
                        this.browser.refresh();
                        this.ready();
                        resolve();
                    }
                    catch (e) {
                        yield this.disconnect();
                        reject(e);
                    }
                }));
                this.setEvents(_resolve, reject);
            }
            catch (e) {
                reject(e);
            }
        }));
    }
    /**
     * 再接続試行
     */
    attemptReconnect(resolve, reject) {
        return __awaiter(this, void 0, void 0, function* () {
            this.log('attemptReconnect()');
            try {
                yield this.reconnect(resolve);
            }
            catch (e) {
                this.log(`Reconnect failed: ${e.message}`);
                if (this.disconnectedTimestamp + 180000 <= moment().valueOf()) {
                    this.reconnectId = null;
                }
                if (this.reconnectId) {
                    this.reconnectTimer = setTimeout(() => {
                        this.attemptReconnect(resolve, reject);
                    }, 5000);
                }
                else {
                    yield this.destructor(resolve);
                }
            }
        });
    }
    /**
     * 仮想コネクタ再接続試行
     */
    attemptReconnectVirtualConnector() {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.destructed) {
                this.log('attemptReconnectVirtualConnector()');
                const retry = (err) => {
                    var _a;
                    this.log(`Reconnect virtual connector failed: ${err.message}`);
                    if (this.virtualConnectorDisconnectedTimestamp + 180000 <= moment().valueOf()) {
                        (_a = this.connector) === null || _a === void 0 ? void 0 : _a.emit('disconnect');
                    }
                    else {
                        this.virtualConnectorReconnectTimer = setTimeout(() => __awaiter(this, void 0, void 0, function* () {
                            yield this.attemptReconnectVirtualConnector();
                        }), 5000);
                    }
                };
                try {
                    yield this.socket.emitWithAckTimeout('notify_connector_disconnected', () => void 0);
                    yield this.socket.emitWithAckTimeout('reboot_virtual_connector', (err, connectorInfo) => __awaiter(this, void 0, void 0, function* () {
                        try {
                            if (err) {
                                throw new Error(err.message);
                            }
                            else {
                                yield this.connector.connect(connectorInfo, () => __awaiter(this, void 0, void 0, function* () {
                                    yield this.socket.emitWithAckTimeout('notify_connector_reconnected', () => void 0);
                                }));
                                this.virtualConnectorDisconnectedTimestamp = null;
                            }
                        }
                        catch (e) {
                            retry(e);
                        }
                    }));
                }
                catch (e) {
                    retry(e);
                }
            }
        });
    }
    /**
     * デストラクタ
     */
    destructor(resolve = () => void 0) {
        var _a, _b, _c;
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.destructed) {
                this.destructed = true;
                // ダウンロード監視停止
                yield ((_a = this.downloadWatcher) === null || _a === void 0 ? void 0 : _a.close());
                try {
                    const deletePaths = ['downloads', 'copied'];
                    for (const deletePath of deletePaths) {
                        yield fs.promises.rmdir(`${config.storage.path}/${this.event.contractId}/${deletePath}/${this.event.uuid}`, {
                            recursive: true
                        });
                    }
                }
                catch (e) {
                    console.log(e.message);
                }
                // ブラウザ停止
                yield ((_b = this.browser) === null || _b === void 0 ? void 0 : _b.stop());
                this.browser = null;
                // コネクタ切断
                (_c = this.connector) === null || _c === void 0 ? void 0 : _c.disconnect();
                this.connector = null;
                yield this.stopBrowserNotification();
            }
            if (this.mode == 'instant') {
                resolve();
            }
        });
    }
    /**
     * イベント設定
     */
    setEvents(resolve, reject) {
        /**
         * 接続エラー
         */
        this.socket.on('connect_error', (err) => __awaiter(this, void 0, void 0, function* () {
            this.log(`connect_error. [${err.message}]`);
            reject(new Error('connect error'));
        }));
        /**
         * 切断
         */
        this.socket.on('disconnect', () => __awaiter(this, void 0, void 0, function* () {
            this.log('disconnect.');
            this.unready();
            if (this.browser === null) {
                resolve();
            }
            else {
                if (this.reconnectId) {
                    this.disconnectedTimestamp = moment().valueOf();
                    this.attemptReconnect(resolve, reject);
                }
                else {
                    yield this.destructor(resolve);
                }
            }
        }));
        /**
         * エレメント監視登録
         */
        this.socket.on('registWatchElement', (data, ack) => __awaiter(this, void 0, void 0, function* () {
            try {
                this.log(`registWatchElement(${data})`);
                this.browser.registWatchElement(data);
                yield ack({
                    result: 0,
                });
            }
            catch (e) {
                this.log('Regist watch element failed.');
                this.log(e.message);
                yield ack({
                    result: -1,
                    message: e.message,
                });
            }
        }));
        /**
         * エレメント監視
         */
        this.socket.on('watchElements', (data, ack) => __awaiter(this, void 0, void 0, function* () {
            try {
                this.log(`WatchElements(${data.append})`);
                const watch = yield this.browser.watchElements(data);
                yield ack({
                    result: 0,
                    watch: watch
                });
            }
            catch (e) {
                this.log('Watch elements failed.');
                this.log(e.stack);
                yield ack({
                    result: -1,
                    message: e.message,
                });
            }
        }));
        /**
         * 再接続ID通知
         */
        this.socket.on('notifyReconnectId', (data, ack) => {
            console.log('onNotifyReconnectId', data);
            this.reconnectId = data.reconnectId;
            ack();
        });
        /**
         * コマンド実行
         */
        this.socket.on('command', (data, ack) => __awaiter(this, void 0, void 0, function* () {
            try {
                if (!data.params) {
                    data.params = {};
                }
                this.log(`Command('${data.command}',script)`);
                const return_var = yield this.browser.command(data);
                yield ack({
                    result: 0,
                    data: return_var,
                });
            }
            catch (e) {
                this.log(`Command failed. [${e.message}]`);
                yield ack({
                    result: -1,
                    message: e.message,
                });
            }
        }));
        /**
         * コマンド実行(stream)
         */
        socketIoStream(this.socket).on('command_stream', (stream, data, ack) => __awaiter(this, void 0, void 0, function* () {
            try {
                if (!data.params) {
                    data.params = {};
                }
                this.log(`Command('${data.command}',script)`);
                yield new Promise((_resolve, _reject) => __awaiter(this, void 0, void 0, function* () {
                    try {
                        const return_var = yield this.browser.command(data);
                        if (return_var) {
                            let timeout;
                            const disconnect = () => {
                                const message = 'socket.io-stream socket disconnected';
                                // this.log(message);
                                if (stream.listenerCount('drain')) {
                                    stream.removeAllListeners('drain');
                                    stream.end();
                                }
                                clearTimeout(timeout);
                                _reject(new Error(message));
                            };
                            const resolve = (value) => {
                                this.socket.off('disconnect', disconnect);
                                clearTimeout(timeout);
                                _resolve(value);
                            };
                            const reject = (reason) => {
                                if (stream.listenerCount('drain')) {
                                    stream.removeAllListeners('drain');
                                    stream.end();
                                }
                                this.socket.off('disconnect', disconnect);
                                clearTimeout(timeout);
                                _reject(reason);
                            };
                            this.socket.once('disconnect', disconnect);
                            const ready = stream.write(return_var);
                            if (!ready) {
                                stream.once('drain', () => {
                                    stream.end();
                                    resolve();
                                });
                                timeout = setTimeout(() => {
                                    const message = 'socket.io-stream stream-write timed out';
                                    this.log(message);
                                    reject(new Error(message));
                                }, 60000);
                            }
                            else {
                                stream.end();
                                resolve();
                            }
                        }
                        else {
                            throw new Error('buffer is empty');
                        }
                    }
                    catch (e) {
                        _reject(e);
                    }
                }));
                ack({
                    result: 0,
                });
            }
            catch (e) {
                this.log(`Command failed. [${e.message}]`);
                yield ack({
                    result: -1,
                    message: e.message,
                });
            }
        }));
        /**
         * アラート操作
         */
        this.socket.on('alert_command', (data, ack) => __awaiter(this, void 0, void 0, function* () {
            this.log(`Alert_command('${data.command}','${data.value}')`);
            yield this.browser.alertCommand(data);
            ack();
        }));
        /**
         * FileChooser操作
         */
        this.socket.on('file_chooser_command', (data, ack) => {
            this.log(`File_chooser_command('${data.tokenId}', ${JSON.stringify(data.targets)})`);
            const fileManager = new fileManager_1.FileManager(data.tokenId, this.event.contractId);
            fileManager.on('progress', (progress) => {
                this.socket.emit('file_chooser_progress', progress);
            });
            this.browser.fileChooserCommand(fileManager, data.targets);
            ack();
        });
        /**
         * FileChooserキャンセル
         */
        this.socket.on('file_chooser_cancel', (ack) => __awaiter(this, void 0, void 0, function* () {
            this.log('File_chooser_cancel()');
            yield this.browser.fileChooserCancel();
            ack();
        }));
        /**
         * unknown namespace
         */
        this.socket.on('$__unknown_namespace', () => {
            this.reconnectId = null;
            this.socket.disconnect();
        });
    }
    /**
     * ブラウザイベント設定
     */
    setBrowserEvents() {
        /**
         * エラー通知
         */
        this.browser.on('error', (e) => __awaiter(this, void 0, void 0, function* () {
            this.log(e.message);
            yield this.disconnect();
        }));
        /**
         * devicePixelRatio更新
         */
        this.browser.on('update_device_pixel_ratio', (devicePixelRatio) => {
            if (this.eventEmitter) {
                this.eventEmitter.emit('update_device_pixel_ratio', devicePixelRatio);
            }
        });
        /**
         * ウインドウリスト通知
         */
        this.browser.on('windows', (windows, ack) => __awaiter(this, void 0, void 0, function* () {
            try {
                yield this.socket.emitWithAckTimeout('windows', windows, () => void 0);
                ack(null);
            }
            catch (e) {
                ack(e);
            }
        }));
        /**
         * 遷移通知
         */
        this.browser.on('transition', (windowState, ack) => __awaiter(this, void 0, void 0, function* () {
            try {
                yield this.socket.emitWithAckTimeout('transition', windowState, () => void 0);
                ack(null);
            }
            catch (e) {
                ack(e);
            }
        }));
        /**
         * スクリーンショット通知
         */
        this.browser.on('screenshot', (grid, ack) => __awaiter(this, void 0, void 0, function* () {
            try {
                yield new Promise((_resolve, _reject) => {
                    if (this.socket.disconnected) {
                        _reject(new Error('socket disconnected'));
                    }
                    else {
                        let timeout;
                        const stream = socketIoStream.createStream({
                            highWaterMark: 1024,
                        });
                        const disconnect = () => {
                            const message = 'socket.io-stream socket disconnected';
                            // this.log(message);
                            if (stream.listenerCount('drain')) {
                                stream.removeAllListeners('drain');
                                stream.end();
                            }
                            clearTimeout(timeout);
                            _reject(new Error(message));
                        };
                        const resolve = (value) => {
                            this.socket.off('disconnect', disconnect);
                            clearTimeout(timeout);
                            _resolve(value);
                        };
                        const reject = (reason) => {
                            if (stream.listenerCount('drain')) {
                                stream.removeAllListeners('drain');
                                stream.end();
                            }
                            this.socket.off('disconnect', disconnect);
                            clearTimeout(timeout);
                            _reject(reason);
                        };
                        this.socket.once('disconnect', disconnect);
                        socketIoStream(this.socket).emit('screenshot', stream, {
                            windowNumber: grid.windowNumber,
                            gridWidth: grid.width,
                            gridHeight: grid.height,
                            gridLeft: grid.left,
                            gridTop: grid.top,
                            timestamp: moment().valueOf()
                        }, () => {
                            resolve();
                        });
                        const ready = stream.write(grid.binary);
                        if (!ready) {
                            stream.once('drain', () => {
                                stream.end();
                            });
                        }
                        else {
                            stream.end();
                        }
                        timeout = setTimeout(() => {
                            const message = 'socket.io-stream stream-write timed out';
                            this.log(message);
                            reject(new Error(message));
                        }, 60000);
                    }
                });
                ack(null);
            }
            catch (e) {
                ack(e);
            }
        }));
        /**
         * エレメント監視通知
         */
        this.browser.on('watchElements', (data, ack) => __awaiter(this, void 0, void 0, function* () {
            try {
                yield this.socket.emitWithAckTimeout('watchElements', {
                    windowNumber: data.windowNumber,
                    elements: data.elements
                }, () => void 0);
                ack(null);
            }
            catch (e) {
                ack(e);
            }
        }));
        /**
         * ダイアログ通知
         */
        this.browser.on('dialog', (text, type) => {
            const alert = () => __awaiter(this, void 0, void 0, function* () {
                try {
                    yield new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
                        try {
                            yield this.socket.emitWithAckTimeout('alert', {
                                text: text,
                                type: type,
                            }, (result, e) => {
                                if (result) {
                                    resolve();
                                }
                                else {
                                    reject(e);
                                }
                            });
                        }
                        catch (e) {
                            reject(e);
                        }
                    }));
                }
                catch (e) {
                    this.log(e.message);
                    if (this.isReady) {
                        setTimeout(() => {
                            this.log('Retry notify alert');
                            alert();
                        }, 1000);
                    }
                }
            });
            alert();
        });
        /**
         * FileChooser通知
         */
        this.browser.on('file_chooser', (isMultiple) => {
            const fileChooser = () => __awaiter(this, void 0, void 0, function* () {
                try {
                    yield new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
                        try {
                            yield this.socket.emitWithAckTimeout('file_chooser', {
                                isMultiple: isMultiple
                            }, (result, e) => {
                                if (result) {
                                    resolve();
                                }
                                else {
                                    reject(e);
                                }
                            });
                        }
                        catch (e) {
                            reject(e);
                        }
                    }));
                }
                catch (e) {
                    this.log(e.message);
                    if (this.isReady) {
                        setTimeout(() => {
                            this.log('Retry notify file chooser');
                            fileChooser();
                        }, 1000);
                    }
                }
            });
            fileChooser();
        });
        /**
         * FileChooser完了通知
         */
        this.browser.on('file_chooser_ended', (data, ack) => __awaiter(this, void 0, void 0, function* () {
            try {
                yield this.socket.emitWithAckTimeout('file_chooser_ended', data, () => void 0);
                ack(null);
            }
            catch (e) {
                ack(e);
            }
        }));
        /**
         * ファイルコピー通知
         */
        this.browser.on('copy_files', (files) => __awaiter(this, void 0, void 0, function* () {
            yield this.checkStorageQuota(files);
        }));
    }
    /**
     * ダウンロードファイル監視
     */
    handleDownloadWatcher() {
        this.downloadWatcher = new downloadWatcher_1.DownloadWatcher(this.event.uuid, this.event.contractId);
        this.downloadWatcher.on('files', (files) => __awaiter(this, void 0, void 0, function* () {
            if (yield this.checkStorageQuota(files)) {
                yield this.socket.emitWithAckTimeout('download_files', files, () => void 0);
            }
        }));
    }
    /**
     * ストレージクォータチェック
     */
    checkStorageQuota(files) {
        return __awaiter(this, void 0, void 0, function* () {
            if (config.browser.storageQuota) {
                for (const file of files) {
                    if (this.storageUsageSize[file.ino] === undefined || this.storageUsageSize[file.ino] < file.size) {
                        this.storageUsageSize[file.ino] = file.size;
                    }
                }
                let totalSize = 0;
                for (const ino in this.storageUsageSize) {
                    totalSize += this.storageUsageSize[ino];
                }
                if (totalSize > config.browser.storageQuota) {
                    this.log('The storage quota has been exceeded');
                    yield this.stop();
                    return false;
                }
            }
            return true;
        });
    }
}
exports.Client = Client;
