"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Downloader = void 0;
const config = require("config");
const contentDisposition = require("content-disposition");
const fs = require("fs");
const mime = require("mime-types");
const url = require("url");
const httpClient_1 = require("../../httpClient");
const utils_1 = require("../../utils");
class Downloader {
    constructor(uuid, contractId, cookies) {
        this.outputPath = `${config.storage.path}/${contractId}/downloads/${uuid}`;
        this.cookies = cookies.map(cookie => cookie.name + '=' + cookie.value).join(';');
    }
    /**
     * ダウンロード
     */
    download(src, originUrl, options) {
        return __awaiter(this, void 0, void 0, function* () {
            // srcがbase64データの場合はバイナリにして保存
            let base64Data;
            if (base64Data = this.extractBase64Data(src)) {
                yield this.saveBase64Data(base64Data);
            }
            else {
                yield this.downloadSrc(src, originUrl, options);
            }
        });
    }
    /**
     * srcをダウンロードする
     */
    downloadSrc(src, originUrl, options) {
        var _a, _b;
        return __awaiter(this, void 0, void 0, function* () {
            // 保存ファイル名
            const parsedSrc = url.parse(src);
            const path = (_a = parsedSrc.pathname) === null || _a === void 0 ? void 0 : _a.replace(/^\//, '').split('/');
            let fileName;
            if (path && path[path.length - 1]) {
                fileName = path[path.length - 1];
            }
            else {
                fileName = 'download.html';
            }
            // srcがパスのみの場合はURLにする
            if (!parsedSrc.protocol) {
                const parsedOriginUrl = url.parse(originUrl);
                const baseUrl = `${parsedOriginUrl.protocol}//${parsedOriginUrl.auth ? parsedOriginUrl.auth + '@' : ''}${parsedOriginUrl.host}`;
                // パスが無い場合
                if (!parsedSrc.pathname) {
                    src = originUrl;
                }
                // 絶対パスの場合
                else if (parsedSrc.pathname.indexOf('/') === 0) {
                    src = `${baseUrl}${src}`;
                }
                // 相対パスの場合
                else {
                    const basePath = parsedOriginUrl.pathname.replace(/[^\/]+$/, '');
                    src = `${baseUrl}${basePath}${src}`;
                }
            }
            // ダウンロード
            const httpClient = httpClient_1.default.getNewInstance();
            httpClient.init(() => __awaiter(this, void 0, void 0, function* () {
                if (!options.proxy.host) {
                    return {
                        enable: false
                    };
                }
                else {
                    return {
                        enable: true,
                        url: utils_1.createProxyUrl(options.proxy.host, options.proxy.port, options.proxy.user, options.proxy.password)
                    };
                }
            }));
            const response = yield httpClient.get(encodeURI(src), {
                headers: {
                    'Accept-Language': options.lang,
                    'Cookie': this.cookies,
                    'User-Agent': options.userAgent
                },
                responseType: 'stream',
                auth: options.auth
            });
            if (response.headers['content-disposition']) {
                const parsedDisposition = contentDisposition.parse(response.headers['content-disposition']);
                if ((_b = parsedDisposition.parameters) === null || _b === void 0 ? void 0 : _b.filename) {
                    fileName = parsedDisposition.parameters.filename;
                }
            }
            const readStream = response.data;
            const writeStream = fs.createWriteStream(`${this.outputPath}/${fileName}.crdownload`);
            writeStream.on('finish', () => __awaiter(this, void 0, void 0, function* () {
                yield fs.promises.rename(`${this.outputPath}/${fileName}.crdownload`, `${this.outputPath}/${fileName}`);
            }));
            readStream.pipe(writeStream);
        });
    }
    /**
     * base64データを保存する
     */
    saveBase64Data(base64Data) {
        return __awaiter(this, void 0, void 0, function* () {
            const ext = mime.extension(base64Data.mimetype);
            const fileName = `download.${ext}`;
            yield fs.promises.writeFile(`${this.outputPath}/${fileName}`, base64Data.data);
        });
    }
    /**
     * base64データを抽出する
     */
    extractBase64Data(src) {
        const matches = src.match(/^data:([^;]+);base64,(.*)$/);
        if (matches) {
            const base64 = matches[2];
            const buf = Buffer.from(base64, 'base64');
            return {
                mimetype: matches[1],
                data: buf
            };
        }
        return null;
    }
}
exports.Downloader = Downloader;
