"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const config = require("config");
const http = require("http");
const url = require("url");
const net = require("net");
const utils_1 = require("../../utils");
class WebProxy {
    constructor() {
        this.usings = {};
        this.host = '127.0.0.1';
        this.port = null;
        this.locked = false;
        // プロキシサーバーの作成
        this.server = http.createServer((req, res) => {
            const clientSocket = req.socket;
            const requestUrl = url.parse(req.url);
            const headers = req.headers;
            delete headers['proxy-connection'];
            const requestToServer = http.request({
                host: requestUrl.hostname,
                port: requestUrl.port || 80,
                path: requestUrl.path,
                method: req.method,
                headers: this.uppercaseHeaderNames(headers),
                agent: clientSocket.httpAgent
            }, (serverResponse) => {
                res.writeHead(serverResponse.statusCode, serverResponse.headers);
                serverResponse.pipe(res);
            });
            req.pipe(requestToServer);
            requestToServer.on('error', (err) => {
                res.writeHead(400, err.message, {
                    'content-type': 'text/html'
                });
                res.end('<h1>Error</h1><div>' + err.message + '</div><div>' + req.url + '</div>');
            });
        });
        this.registConnectEvent();
        this.registConnectionEvent();
        this.registClientErrorEvent();
    }
    /**
     * コネクトメソッドの処理を登録
     */
    registConnectEvent() {
        this.server.on('connect', (req, clientSocket, cliHead) => {
            const requestUrl = url.parse('https://' + req.url);
            // コネクタ用のループバックホスト名を通常のホスト名に戻す
            if (requestUrl.hostname == 'localhost.agent.local' || requestUrl.hostname == '127.0.0.1.agent.local') {
                requestUrl.hostname = requestUrl.hostname.replace('.agent.local', '');
            }
            const serverSocket = net.connect(parseInt(requestUrl.port) || 443, requestUrl.hostname, () => {
                clientSocket.write('HTTP/1.1 200 Connection established\r\n\r\n');
                if (cliHead && cliHead.length) {
                    serverSocket.write(cliHead);
                }
                clientSocket.pipe(serverSocket);
            });
            serverSocket.pipe(clientSocket);
            serverSocket.on('error', () => {
                clientSocket.end();
            });
            clientSocket.on('error', () => {
                if (serverSocket) {
                    serverSocket.end();
                }
            });
        });
    }
    /**
     * connectionイベント発生時の処理を登録
     */
    registConnectionEvent() {
        this.server.on('connection', (clientSocket) => {
            if (!this.isLoopbackRequest(clientSocket)) {
                clientSocket.end();
                return;
            }
            clientSocket.httpAgent = new http.Agent({
                keepAlive: true
            });
        });
    }
    /**
     * clientErrorイベント発生時の処理を登録
     */
    registClientErrorEvent() {
        this.server.on('clientError', (err, clientSocket) => {
            clientSocket.end();
        });
    }
    /**
     * ループバックからのリクエストか
     */
    isLoopbackRequest(socket) {
        const ip = socket.remoteAddress.split(':').pop();
        return ip == '127.0.0.1';
    }
    /**
     * ヘッダ名の頭文字を大文字に変換する
     */
    uppercaseHeaderNames(headers) {
        for (const headerName in headers) {
            headers[this.uppercaseKebabInitials(headerName)] = headers[headerName];
            delete headers[headerName];
        }
        return headers;
    }
    /**
     * ケバブケース文字列の各単語の頭文字を大文字に変換する
     */
    uppercaseKebabInitials(words) {
        return words.split('-').map((word) => {
            return word.charAt(0).toUpperCase() + word.slice(1);
        }).join('-');
    }
    /**
     * Listen開始
     */
    listen() {
        return new Promise((resolve, reject) => {
            let [startPort, endPort] = config.port.webProxy.split('-').map((port) => {
                return parseInt(port);
            });
            if (!endPort) {
                endPort = startPort;
            }
            const attempt = (port) => {
                const listeningHandler = () => {
                    this.server.off('error', errorHandler);
                    resolve(port);
                };
                const errorHandler = (error) => {
                    this.server.off('listening', listeningHandler);
                    console.log(error.message);
                    if (error.message.match('EADDRINUSE') && port < endPort) {
                        attempt(++port);
                    }
                    else {
                        reject(error);
                    }
                };
                this.server.once('listening', listeningHandler);
                this.server.once('error', errorHandler);
                this.server.listen(port);
            };
            attempt(startPort);
        });
    }
    /**
     * Close
     */
    close() {
        return __awaiter(this, void 0, void 0, function* () {
            return new Promise((resolve, reject) => {
                this.server.close(() => {
                    resolve();
                });
            });
        });
    }
    /**
     * ロックが解除されるまで待機
     */
    waitUntilUnlocked() {
        return __awaiter(this, void 0, void 0, function* () {
            while (true) {
                if (!this.locked) {
                    break;
                }
                yield utils_1.sleep(100);
            }
        });
    }
    /**
     * 起動
     */
    start(id) {
        return new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
            yield this.waitUntilUnlocked();
            this.locked = true;
            try {
                this.usings[id] = true;
                if (Object.keys(this.usings).length > 1) {
                    resolve({
                        host: this.host,
                        port: this.port
                    });
                }
                else {
                    this.port = yield this.listen();
                    console.log(`Web proxy listening at http://127.0.0.1:${this.port}`);
                    resolve({
                        host: this.host,
                        port: this.port
                    });
                }
            }
            catch (e) {
                delete this.usings[id];
                reject(e);
            }
            finally {
                this.locked = false;
            }
        }));
    }
    /**
     * 停止
     */
    stop(id) {
        return new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
            if (id in this.usings) {
                yield this.waitUntilUnlocked();
                this.locked = true;
                try {
                    delete this.usings[id];
                    if (Object.keys(this.usings).length >= 1) {
                        resolve();
                    }
                    else {
                        yield this.close();
                        console.log(`Web proxy closed. http://127.0.0.1:${this.port}`);
                        resolve();
                    }
                }
                catch (e) {
                    this.usings[id] = true;
                    reject(e);
                }
                finally {
                    this.locked = false;
                }
            }
            else {
                resolve();
            }
        }));
    }
}
exports.default = new WebProxy();
