"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const events_1 = require("events");
const Tunnel_1 = require("./Tunnel");
const types_1 = require("./types");
const logger_1 = require("./logger");
const util_1 = require("./util");
const ss = require("socket.io-stream");
var TunnelBuilderEvent;
(function (TunnelBuilderEvent) {
    // トンネル登録申請
    TunnelBuilderEvent["RequestTunnel"] = "request_tunnel";
    // トンネルの完了
    TunnelBuilderEvent["CloseTunnel"] = "close_tunnel";
    // トンネルリスニングスタート
    TunnelBuilderEvent["ListenTunnel"] = "listen_tunnnel";
    // TCPストリームからデータの受信
    TunnelBuilderEvent["Data"] = "data";
    // トンネルからTCPストリームの生成
    TunnelBuilderEvent["OpenStream"] = "open_stream";
    // TCPストリームの完了
    TunnelBuilderEvent["CloseStream"] = "close_stream";
    // トンネルエラー
    TunnelBuilderEvent["Error"] = "error_tunnel";
})(TunnelBuilderEvent || (TunnelBuilderEvent = {}));
class TunnelBuilder extends events_1.EventEmitter {
    constructor(wsSocket, options = {
        // authイベント未定義時に許可する
        approveOnNoSetAuthHandler: true
    }) {
        super();
        this.tunnels = {};
        this.wsSocket = wsSocket;
        this._policy = {
            whiteList: true,
            items: []
        };
        this._options = options;
        this.setWsSocketHandler();
    }
    emit(event, ...datas) {
        if (typeof event === 'string' && datas.length === 2) {
            // トンネル申請
            const tunnelName = event;
            const setting = datas[0];
            const ackCallback = datas[1];
            if (this.isExistTunnel(tunnelName)) {
                // すでに登録されたトンネル
                logger_1.Logger.systemError(`[${tunnelName}] Already exist`);
                return false;
            }
            if (setting.tcpHost === undefined || setting.tcpPort === undefined) {
                logger_1.Logger.systemError(`[${tunnelName}] Invalid setting`);
                return false;
            }
            // ローカルトンネルを登録
            const tunnel = this.openTunnel(types_1.TunnelType.Local, tunnelName, setting);
            const listenHandler = () => {
                tunnel.off(Tunnel_1.TunnelEvent.Error, errorHandler);
                logger_1.Logger.systemInfo(`[${tunnelName}] Requesting tunnel`);
                // リモートトンネルの登録申請
                this.wsSocket.emit(TunnelBuilderEvent.RequestTunnel, tunnelName, setting, (success, error) => {
                    if (success) {
                        ackCallback(tunnel);
                        logger_1.Logger.systemInfo(`[${tunnelName}] Opened tunnel`);
                    }
                    else {
                        const error = `[${tunnelName}] Failed to request tunnel`;
                        tunnel.throwError(error);
                        tunnel.close();
                    }
                });
            };
            const errorHandler = () => {
                tunnel.off(Tunnel_1.TunnelEvent.Listen, listenHandler);
                ackCallback(null);
            };
            tunnel.once(Tunnel_1.TunnelEvent.Listen, listenHandler);
            tunnel.once(Tunnel_1.TunnelEvent.Error, errorHandler);
            return true;
        }
        return false;
    }
    on(event, listener) {
        super.on(event, listener);
        return this;
    }
    once(event, listener) {
        super.once(event, listener);
        return this;
    }
    off(event, listener) {
        super.off(event, listener);
        return this;
    }
    isExistTunnel(tunnelName) {
        return tunnelName in this.tunnels;
    }
    // トンネル用イベントハンドラ
    setTunnelHandler(tunnel) {
        const tunnelName = tunnel.name;
        if (tunnel.tunnelType === types_1.TunnelType.Local) {
            // ローカルトンネルからTCPストリームの生成
            tunnel.on(Tunnel_1.TunnelEvent.OpenStream, (stream) => {
                logger_1.Logger.accessInfo(`[${tunnelName}] LOCAL tunnel >> Open stream(${stream.id})`);
                // リモートトンネルにTCPストリームを申請
                this.wsSocket.emit(TunnelBuilderEvent.OpenStream, tunnelName, stream.id, stream.tcpClientHost, stream.tcpClientPort);
            });
        }
        // ローカルTCPサーバーがリスニングスタート
        tunnel.on(Tunnel_1.TunnelEvent.Listen, (listenPort) => {
            this.tunnels[tunnelName] = tunnel;
            this.wsSocket.emit(TunnelBuilderEvent.ListenTunnel, tunnelName, listenPort);
        });
        // ローカルTCPストリームからデータの受信
        tunnel.on(Tunnel_1.TunnelEvent.sendStreamFromTCP, (streamId, stream) => {
            // Logger.systemDebug(`[${tunnelName}/${streamId}] LOCAL tunnel >> REMOTE tunnel`, data);
            // ローカルTCPストリームから受信したデータをリモートトンネルへ送信する
            const ioStream = ss.createStream({
                highWaterMark: 1024,
            });
            ss(this.wsSocket).emit(TunnelBuilderEvent.Data, tunnelName, streamId, ioStream);
            stream.pipe(ioStream);
            ioStream.on('end', () => {
                logger_1.Logger.accessInfo(`[${tunnelName}] LOCAL tunnel >> Close stream(${streamId})`);
                this.wsSocket.emit(TunnelBuilderEvent.CloseStream, tunnelName, streamId);
            });
        });
        // ローカルトンネル完了
        tunnel.on(Tunnel_1.TunnelEvent.CloseTunnel, () => {
            logger_1.Logger.accessInfo(`[${tunnelName}] LOCAL >> Close tunnel`);
            this.deleteTunnel(tunnel.name, false);
        });
        tunnel.on(Tunnel_1.TunnelEvent.Error, (err) => {
            this.wsSocket.emit(TunnelBuilderEvent.Error, tunnelName, err);
        });
        tunnel.on(Tunnel_1.TunnelEvent.ErrorLocalTCPServer, (err) => {
            this.throwError(err);
        });
    }
    openTunnel(tunnelType, tunnelName, setting) {
        const tunnel = new Tunnel_1.Tunnel(tunnelType, tunnelName, setting, this._options.approveOnNoSetAuthHandler);
        if (tunnel.tunnelType == types_1.TunnelType.Remote) {
            this.tunnels[tunnelName] = tunnel;
        }
        logger_1.Logger.systemInfo(`[${tunnelName}] New Tunnel info: type=${tunnelType} name=${tunnelName} tcpHost=${setting.tcpHost} tcpPort=${setting.tcpPort} listenPort=${setting.listenPort}`);
        this.setTunnelHandler(tunnel);
        return tunnel;
    }
    deleteTunnel(tunnelName, fromRemote = false) {
        if (this.tunnels[tunnelName]) {
            delete this.tunnels[tunnelName];
            if (!fromRemote && this.wsSocket.connected) {
                this.wsSocket.emit(TunnelBuilderEvent.CloseTunnel, tunnelName);
            }
        }
    }
    throwError(error) {
        super.emit(TunnelBuilderEvent.Error, error);
    }
    // ポリシーチェック
    checkPolicy(setting) {
        if (this.policy.items.length > 0) {
            if (this.policy.whiteList) {
                return this.policy.items.map((item) => util_1.isAddrInSubNet(setting.tcpHost, item.address) &&
                    util_1.isPortInRange(setting.tcpPort, item.fromPort, item.toPort))
                    .filter((valid) => valid).length > 0;
            }
            else {
                // blacklist
                return this.policy.items.map((item) => util_1.isAddrInSubNet(setting.tcpHost, item.address) &&
                    util_1.isPortInRange(setting.tcpPort, item.fromPort, item.toPort))
                    .filter((valid) => !valid).length == 0;
            }
        }
        return true;
    }
    // Webソケット用イベントハンドラ
    setWsSocketHandler() {
        // トンネル登録申請
        this.wsSocket.on(TunnelBuilderEvent.RequestTunnel, (tunnelName, setting, ackCallback) => {
            if (this.isExistTunnel(tunnelName)) {
                // エラー：トンネルがすでに登録されている場合
                const error = `[${tunnelName}] REMOTE >> Failed to request tunnel: Already exist tunnel`;
                logger_1.Logger.systemError(error);
                ackCallback(false, error);
                return;
            }
            if (!this.checkPolicy(setting)) {
                // エラー：トンネルのアドレスとポートがポリシーに許可され無い場合
                const error = `[${tunnelName}] REMOTE >> Failed to request tunnel: Denied by policy`;
                logger_1.Logger.systemError(error);
                ackCallback(false, error);
                return;
            }
            // 新のトンネルの登録
            const tunnel = this.openTunnel(types_1.TunnelType.Remote, tunnelName, setting);
            logger_1.Logger.systemInfo(`[${tunnelName}] REMOTE >> Opened tunnel`);
            super.emit(tunnelName, tunnel);
            ackCallback(true, '');
        });
        // トンネルリスニングスタート
        this.wsSocket.on(TunnelBuilderEvent.ListenTunnel, (tunnelName, listenPort) => {
            const tunnel = this.tunnels[tunnelName];
            if (tunnel !== undefined) {
                // 新のTCPストリームの登録
                logger_1.Logger.systemInfo(`[${tunnelName}] REMOTE >> Start listening(:${listenPort})`);
                tunnel.listenPort = listenPort;
            }
            else {
                logger_1.Logger.systemError(`[${tunnelName}] Not found tunnel.`);
            }
        });
        // リモートトンネルからTCPストリームの生成
        this.wsSocket.on(TunnelBuilderEvent.OpenStream, (tunnelName, streamId, tcpClientHost, tcpClientPort) => {
            logger_1.Logger.accessInfo(`[${tunnelName}] REMOTE tunnel >> Open stream(${streamId})`);
            const tunnel = this.tunnels[tunnelName];
            if (tunnel !== undefined) {
                // 新のTCPストリームの登録
                tunnel.openStream(streamId, tcpClientHost, tcpClientPort);
            }
            else {
                logger_1.Logger.systemError(`[${tunnelName}] Not found tunnel.`);
            }
        });
        // リモートTCPストリームからデータの受信
        ss(this.wsSocket).on(TunnelBuilderEvent.Data, (tunnelName, streamId, ioStream) => {
            const tunnel = this.tunnels[tunnelName];
            if (tunnel !== undefined) {
                // リモートTCPストリームから受信したデーターをローカルTCPストリームへ送信する
                // Logger.systemDebug(`[${tunnelName}/${streamId}] REMOTE tunnel >> LOCAL tunnel`, data);
                tunnel.sendDataToTcp(streamId, ioStream);
            }
            else {
                logger_1.Logger.systemError(`[${tunnelName}] Not found tunnel.`);
            }
        });
        // リモートTCPストリームの完了
        this.wsSocket.on(TunnelBuilderEvent.CloseStream, (tunnelName, streamId) => {
            logger_1.Logger.accessInfo(`[${tunnelName}] REMOTE >> Close stream(${streamId})`);
            const tunnel = this.tunnels[tunnelName];
            if (tunnel !== undefined) {
                tunnel.closeStream(streamId, true);
            }
            else {
                logger_1.Logger.systemError(`[${tunnelName}] Not found tunnel.`);
            }
        });
        // リモートからトンネルの完了
        this.wsSocket.on(TunnelBuilderEvent.CloseTunnel, (tunnelName) => {
            logger_1.Logger.accessInfo(`[${tunnelName}] REMOTE >> Close tunnel`);
            const tunnel = this.tunnels[tunnelName];
            if (tunnel !== undefined) {
                // ローカルトンネル削除
                tunnel.close();
                this.deleteTunnel(tunnelName, true);
            }
        });
        // エラー処理
        this.wsSocket.on(TunnelBuilderEvent.Error, (tunnelName, err) => {
            logger_1.Logger.systemError(`[${tunnelName}] REMOTE >> Error`);
            logger_1.Logger.systemError(err);
        });
        // Webソケットの接続切れ
        this.wsSocket.on('disconnect', () => {
            logger_1.Logger.accessInfo(`Closed ws socket`);
            // close all tunnel
            const tunnelNames = Object.keys(this.tunnels);
            tunnelNames.forEach((tunnelName) => {
                this.tunnels[tunnelName].close();
                super.removeAllListeners(tunnelName);
            });
            super.emit('disconnect');
        });
    }
    get policy() {
        return this._policy;
    }
    set policy(val) {
        this._policy = val;
    }
}
exports.TunnelBuilder = TunnelBuilder;
//# sourceMappingURL=TunnelBuilder.js.map