"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const events_1 = require("events");
const types_1 = require("./types");
const TCPStream_1 = require("./TCPStream");
const logger_1 = require("./logger");
const net = require("net");
const nanoid = require("nanoid");
var TunnelEvent;
(function (TunnelEvent) {
    TunnelEvent["OpenStream"] = "open";
    TunnelEvent["CloseStream"] = "close";
    TunnelEvent["Listen"] = "listen";
    TunnelEvent["ReceiveDataFromTCP"] = "receive_data_from_tcp";
    TunnelEvent["Auth"] = "auth";
    TunnelEvent["CloseTunnel"] = "close_tunnel";
    TunnelEvent["Error"] = "error";
    TunnelEvent["ErrorLocalTCPServer"] = "error_local_tcp_server";
})(TunnelEvent = exports.TunnelEvent || (exports.TunnelEvent = {}));
class Tunnel extends events_1.EventEmitter {
    constructor(tunnelType, tunnelName, setting, approveOnNoSetAuthHandler = false) {
        super();
        this.streams = {};
        this._tunnelId = nanoid();
        this._tunnelType = tunnelType;
        this._tunnelName = tunnelName;
        this._tcpHost = setting.tcpHost;
        this._tcpPort = setting.tcpPort;
        this._listenPort = setting.listenPort || 0;
        this._authId = setting.authId || '';
        this._authKey = setting.authKey || '';
        this._status = types_1.TunnelStatus.Create;
        this._receiveSize = 0;
        this._sendSize = 0;
        if (this.tunnelType === types_1.TunnelType.Local) {
            this.startLocalTCPServer(setting);
        }
        else {
            setTimeout(() => {
                if (super.listenerCount(TunnelEvent.Auth) === 0) {
                    if (!approveOnNoSetAuthHandler) {
                        // authイベント未定義時に拒否にする
                        this.throwError(`[${this.name}] Failed to auth tunnel`);
                        this.close();
                    }
                }
                super.emit(TunnelEvent.Auth, (success) => {
                    if (!success) {
                        // failed auth
                        this.throwError(`[${this.name}] Failed to auth tunnel`);
                        this.close();
                    }
                    else {
                        logger_1.Logger.accessInfo(`[${this.name}] Authoring OK`);
                    }
                });
            }, 1000);
        }
    }
    // TCPストリームをトンネルに登録
    registerStream(socket, streamId, tcpClientHost, tcpClientPort) {
        const stream = new TCPStream_1.TCPStream(this.name, socket, streamId, tcpClientHost, tcpClientPort);
        this.streams[stream.id] = stream;
        super.emit(TunnelEvent.OpenStream, stream);
        // ストリームからデータ受信
        stream.on(TCPStream_1.TCPStreamEvent.ReceiveData, (data) => {
            super.emit(TunnelEvent.ReceiveDataFromTCP, stream.id, data);
            this._receiveSize += data.length;
        });
        stream.on(TCPStream_1.TCPStreamEvent.Error, (err) => {
            stream.close(false);
            super.emit(TunnelEvent.CloseStream, stream.id);
            this.deleteStream(stream.id);
        });
        // ストリーム完了
        stream.on(TCPStream_1.TCPStreamEvent.Close, () => {
            if (!stream.closedRemote)
                super.emit(TunnelEvent.CloseStream, stream.id);
            this.deleteStream(stream.id);
        });
        return stream;
    }
    // ローカルTCPソケットサーバーを起動
    startLocalTCPServer(setting) {
        if (setting.listenServer) {
            this._localTcpServerInjected = true;
            this._localTcpServer = setting.listenServer;
            const addr = this._localTcpServer.address();
            this._listenPort = addr.port;
            setting.listenPort = addr.port;
            setTimeout(() => {
                super.emit(TunnelEvent.Listen, this._listenPort);
            });
        }
        else {
            this._localTcpServerInjected = false;
            this._localTcpServer = net.createServer();
            this._localTcpServer.listen(this._listenPort);
            this._localTcpServer.on('error', (e) => {
                this._localTcpServer.close();
                super.emit(TunnelEvent.ErrorLocalTCPServer, e.message);
            });
            this._localTcpServer.on('listening', () => {
                const addr = this._localTcpServer.address();
                // 自動割り当てた場合
                this._listenPort = addr.port;
                logger_1.Logger.systemInfo(`[${this.name}] TCP Server(:${this._listenPort}) is listening!`);
                super.emit(TunnelEvent.Listen, this._listenPort);
            });
        }
        this._localTcpServer.on('connection', (socket) => {
            logger_1.Logger.systemInfo(`[${this.name}] Connected new TCP client`);
            // 新のストリームを登録
            const newStreamId = nanoid();
            this.registerStream(socket, newStreamId, socket.remoteAddress, socket.remotePort);
        });
    }
    openStream(streamId, tcpClientHost, tcpClientPort) {
        let stream = this.streams[streamId];
        if (stream !== undefined) {
            return stream;
        }
        if (this.tunnelType === types_1.TunnelType.Remote) {
            // リモートTCPサーバーに接続する
            const socket = new net.Socket();
            socket.connect(this._tcpPort, this._tcpHost, () => {
                logger_1.Logger.accessInfo(`[${this.name}] Connected to TCP server(${this._tcpHost}:${this._tcpPort})`);
            });
            stream = this.registerStream(socket, streamId, tcpClientHost, tcpClientPort);
            return stream;
        }
        return undefined;
    }
    sendDataToTcp(streamId, data) {
        let stream = this.streams[streamId];
        if (stream === undefined) {
            // not found stream
            this.throwError(`[${this.name}/${streamId}] Failed to send data to stream: not found stream`);
            return;
        }
        stream.sendData(data);
        this._sendSize += data.length;
    }
    deleteStream(streamId) {
        if (this.streams[streamId])
            delete this.streams[streamId];
    }
    closeStream(streamId, fromRemote) {
        const stream = this.streams[streamId];
        if (stream === undefined) {
            // not found stream
            logger_1.Logger.systemWarning(`[${this.name}/${streamId}] Failed to close stream: not found stream`);
            return;
        }
        if (fromRemote) {
            logger_1.Logger.accessInfo(`[${this.name}/${streamId}] REMOTE >> Closed the stream`);
        }
        stream.close(fromRemote);
        this.deleteStream(streamId);
    }
    close() {
        if (this._status === types_1.TunnelStatus.Close) {
            // avoid unlimited close event
            logger_1.Logger.accessInfo(`[${this.name}] Already closed`);
            return;
        }
        // close all stream
        const streamIds = Object.keys(this.streams);
        streamIds.forEach((streamId) => {
            this.streams[streamId].close();
            this.deleteStream(streamId);
        });
        if (this.tunnelType === types_1.TunnelType.Local) {
            this._localTcpServer.removeAllListeners('connection');
            if (!this._localTcpServerInjected) {
                this._localTcpServer.close();
            }
            delete this._localTcpServer;
        }
        this._status = types_1.TunnelStatus.Close;
        this.emit(TunnelEvent.CloseTunnel);
    }
    throwError(error) {
        logger_1.Logger.systemError(error);
        super.emit(TunnelEvent.Error, error);
    }
    get id() {
        return this._tunnelId;
    }
    get tunnelType() {
        return this._tunnelType;
    }
    get name() {
        return this._tunnelName;
    }
    get authId() {
        return this._authId;
    }
    get authKey() {
        return this._authKey;
    }
    get status() {
        return this._status;
    }
    get tcpHost() {
        return this._tcpHost;
    }
    get tcpPort() {
        return this._tcpPort;
    }
    get listenPort() {
        return this._listenPort;
    }
    set listenPort(val) {
        this._listenPort = val;
    }
    get receiveSize() {
        return this._receiveSize;
    }
    get sendSize() {
        return this._sendSize;
    }
    get streamCount() {
        return Object.keys(this.streams).length;
    }
}
exports.Tunnel = Tunnel;
//# sourceMappingURL=Tunnel.js.map