"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const axios_1 = require("axios");
const commandError_1 = require("../errors/commandError");
const agentStatusError_1 = require("../errors/agentStatusError");
class default_1 {
    constructor(agent) {
        this.agent = agent;
    }
    /**
     * エージェントが起動しているか
     */
    isRunningAgent() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const requestUrl = `http://${this.agent.agent_endpoint}:${this.agent.agent_port}/api/healthcheck`;
                yield axios_1.default.get(requestUrl);
                return true;
            }
            catch (error) {
                return false;
            }
        });
    }
    /**
     * エージェントを初期化する
     */
    initialize() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const requestUrl = `http://${this.agent.agent_endpoint}:${this.agent.agent_port}/api/agent/initialize`;
                yield axios_1.default.post(requestUrl, null, {
                    'headers': {
                        'content-type': 'application/json',
                        'x-agent-id': this.agent.agent_id,
                        'x-secret': this.agent.agent_secret
                    }
                });
            }
            catch (error) {
                console.log(error.message);
                throw new commandError_1.CommandError('FAILED_AGENT_INIT_API');
            }
        });
    }
    /**
     * 言語初期化
     */
    initLang(lang) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const requestUrl = `http://${this.agent.agent_endpoint}:${this.agent.agent_port}/api/agent/lang/initialize`;
                yield axios_1.default.post(requestUrl, {
                    'lang': lang
                }, {
                    'headers': {
                        'content-type': 'application/json',
                        'x-agent-id': this.agent.agent_id,
                        'x-secret': this.agent.agent_secret
                    }
                });
            }
            catch (error) {
                console.log(error.message);
                throw new commandError_1.CommandError('FAILED_AGENT_INIT_LANG_API');
            }
        });
    }
    /**
     * エージェントプロフィール更新
     */
    updateProfile(name, max_boots, allow_file_transfer, temp) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const requestUrl = `http://${this.agent.agent_endpoint}:${this.agent.agent_port}/api/agent/profile`;
                yield axios_1.default.post(requestUrl, {
                    'name': name,
                    'max_values': {
                        '1': Number(max_boots.browser),
                        '2': Number(max_boots.connector)
                    },
                    'allow_file_transfer': {
                        'inbound': allow_file_transfer.inbound,
                        'outbound': allow_file_transfer.outbound
                    },
                    'temp': temp
                }, {
                    'headers': {
                        'content-type': 'application/json',
                        'x-agent-id': this.agent.agent_id,
                        'x-secret': this.agent.agent_secret
                    }
                });
            }
            catch (error) {
                console.log(error.message);
                throw new commandError_1.CommandError('FAILED_AGENT_UPDATE_PROFILE_API');
            }
        });
    }
    /**
     * エージェントプロキシ更新
     */
    reloadProxy() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const requestUrl = `http://${this.agent.agent_endpoint}:${this.agent.agent_port}/api/agent/proxy/reload`;
                yield axios_1.default.post(requestUrl, null, {
                    'headers': {
                        'content-type': 'application/json',
                        'x-agent-id': this.agent.agent_id,
                        'x-secret': this.agent.agent_secret
                    }
                });
            }
            catch (error) {
                console.log(error.message);
                throw new commandError_1.CommandError('FAILED_AGENT_RELOAD_PROXY_API');
            }
        });
    }
    /**
     * エージェント接続
     */
    connect() {
        var _a;
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const requestUrl = `http://${this.agent.agent_endpoint}:${this.agent.agent_port}/api/agent/connect`;
                const response = yield axios_1.default.post(requestUrl, null, {
                    'headers': {
                        'content-type': 'application/json',
                        'x-agent-id': this.agent.agent_id,
                        'x-secret': this.agent.agent_secret
                    }
                });
                // エージェントがオンライン以外はステータスエラー
                if (((_a = response.data) === null || _a === void 0 ? void 0 : _a.code) != 201) {
                    throw new agentStatusError_1.AgentStatusError(response.data.code);
                }
            }
            catch (error) {
                if (error instanceof agentStatusError_1.AgentStatusError) {
                    throw error;
                }
                else {
                    console.log(error.message);
                    throw new commandError_1.CommandError('FAILED_AGENT_CONNECT_API');
                }
            }
        });
    }
    /**
     * エージェントPIN受付APIをリクエスト
     *
     * @params pin string
     */
    inputPin(pin) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const requestUrl = `http://${this.agent.agent_endpoint}:${this.agent.agent_port}/api/agent/pin`;
                yield axios_1.default.post(requestUrl, {
                    "pin": pin
                }, {
                    'headers': {
                        'content-type': 'application/json',
                        'x-agent-id': this.agent.agent_id,
                        'x-secret': this.agent.agent_secret
                    }
                });
            }
            catch (error) {
                console.log(error.message);
                throw new commandError_1.CommandError('FAILED_AGENT_INPUT_PIN_API');
            }
        });
    }
    /**
     * エージェント終了
     */
    exitProcess() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const requestUrl = `http://${this.agent.agent_endpoint}:${this.agent.agent_port}/api/process`;
                yield axios_1.default.delete(requestUrl, {
                    'headers': {
                        'content-type': 'application/json',
                        'x-agent-id': this.agent.agent_id,
                        'x-secret': this.agent.agent_secret
                    }
                });
            }
            catch (error) {
                console.log(error.message);
                throw new commandError_1.CommandError('FAILED_AGENT_PROCESS_EXIT_API');
            }
        });
    }
}
exports.default = default_1;
