#!/usr/bin/env node
"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const commander_1 = require("commander");
const envNameRequest_1 = require("../requests/envNameRequest");
const langRequest_1 = require("../requests/langRequest");
const profileRequest_1 = require("../requests/profileRequest");
const pinRequest_1 = require("../requests/pinRequest");
const agent_1 = require("./agent");
const device_1 = require("../device");
const apiClient_1 = require("./apiClient");
const operation_1 = require("./operation");
const app_1 = require("../app");
const environment_1 = require("../environment");
const manager_1 = require("../aliveMonitor/manager");
const commandError_1 = require("../errors/commandError");
const agentStatusError_1 = require("../errors/agentStatusError");
(() => __awaiter(void 0, void 0, void 0, function* () {
    var _a, _b, _c, _d, _e;
    try {
        commander_1.program
            .version((new app_1.default()).version, '-v, --version')
            .description('Connect the agent.')
            .option('--env-name <env name>', 'Add the agent environment variable name.')
            .option('--lang <language>', 'Add ja or en (default ja).')
            .option('--agent-name <name>', 'Add agent name (Enter less than 64 characters).')
            .option('--browser-max-boot <browser max boot>', 'Add maximum number of simultaneous browser launches.')
            .option('--connector-max-boot <connector max boot>', 'Add maximum number of simultaneous connector launches.')
            .option('--pin <pin code>', 'Add pin code (Enter in 6 digit integer).')
            .parse(process.argv);
        // 引数
        const options = commander_1.program.opts();
        const env_name = (_a = options.envName) !== null && _a !== void 0 ? _a : 'default';
        // 環境変数名のバリデーション
        const envNameRequest = new envNameRequest_1.default(env_name);
        envNameRequest.validate();
        // 環境情報
        const env = new environment_1.default(env_name);
        // デバイス情報
        const device = new device_1.default(env);
        // エージェント情報
        const agent = new agent_1.default(env);
        // APIクライアント
        const apiClient = new apiClient_1.default(agent);
        // エージェント操作
        const agentOperation = new operation_1.default(agent, apiClient);
        // エージェント死活監視の管理
        const aliveMonitor = new manager_1.default(env);
        const lang = (_b = options.lang) !== null && _b !== void 0 ? _b : 'ja';
        const agent_name = (_c = options.agentName) !== null && _c !== void 0 ? _c : `${device.model}-${device.mac}`;
        const max_boots = {
            browser: (_d = options.browserMaxBoot) !== null && _d !== void 0 ? _d : env.agent_config.browser.default_max_boot,
            connector: (_e = options.connectorMaxBoot) !== null && _e !== void 0 ? _e : env.agent_config.connector.default_max_boot
        };
        const pin = options.pin;
        // クラウドBOT用エージェントのバリデーション
        if (env.env_name == 'cbot') {
            // エージェントプロフィールのバリデーション
            const profileRequest = new profileRequest_1.default(agent_name, max_boots);
            profileRequest.validate();
        }
        // サポート用エージェントのバリデーション
        else if (env.env_name == 'support') {
            // 言語のバリデーション
            const langRequest = new langRequest_1.default(lang);
            langRequest.validate();
            // エージェントプロフィールのバリデーション
            const profileRequest = new profileRequest_1.default(agent_name, max_boots);
            profileRequest.validate();
            // PINコードのバリデーション
            const pinRequest = new pinRequest_1.default(pin);
            pinRequest.validate();
        }
        // 識別情報の設定
        console.log('Set agent serial number...');
        device.setHardware();
        console.log('done');
        // クラウドBOT用エージェントの場合、死活監視プロセスを終了する
        if (env.env_name == 'cbot') {
            yield aliveMonitor.kill();
        }
        // エージェント終了
        try {
            yield agentOperation.exitProcess();
        }
        catch (error) {
            // エージェントが立ち上がってない時もコールしてエラーとなるが無視する
        }
        try {
            // クラウドBOT用エージェントの接続
            if (env.env_name == 'cbot') {
                // エージェントを起動
                const pid = yield agentOperation.bootProcess();
                // エージェントPIDを保存する
                env.setAgentPid(pid);
                // エージェントのレディ状態を確認
                yield agentOperation.checkProcess();
                // エージェントプロフィール更新
                const allow_file_transfer = { inbound: true, outbound: true };
                yield agentOperation.updateProfile(agent_name, max_boots, allow_file_transfer, true);
                // エージェント死活監視のリセット
                yield aliveMonitor.boot();
            }
            // サポート用エージェントの接続
            else if (env.env_name == 'support') {
                // statusを0（オフラインにする）
                agentOperation.setAgentStatusOffline();
                // エージェントを起動
                yield agentOperation.bootProcess();
                // エージェントのレディ状態を確認
                yield agentOperation.checkProcess();
                // エージェントの初期化
                yield agentOperation.initialize();
                // 言語設定
                yield agentOperation.initLang(lang);
                // エージェントプロフィール更新
                const allow_file_transfer = { inbound: false, outbound: false };
                yield agentOperation.updateProfile(agent_name, max_boots, allow_file_transfer);
                // エージェント接続
                yield agentOperation.connect();
                // エージェントPIN受付APIのリクエスト
                yield agentOperation.inputPin(pin);
            }
        }
        catch (error) {
            // クラウドBOT用エージェントの場合、死活監視プロセスを終了する
            if (env.env_name == 'cbot') {
                yield aliveMonitor.kill();
            }
            // エージェント終了
            try {
                yield agentOperation.exitProcess();
            }
            catch (error) {
                // 例外は無視する
                console.log(error.message);
            }
            throw error;
        }
        console.log('Successful connection to the agent!!!');
        process.exit(0);
    }
    catch (error) {
        console.log('Failed to connect to the agent!!!');
        if (error instanceof commandError_1.CommandError || error instanceof agentStatusError_1.AgentStatusError) {
            console.log(`error: ${error.message}`);
            console.error(error.getJsonError());
            process.exit(error.getExitcode());
        }
        else {
            console.log('error: undefined error');
            console.error(JSON.stringify({
                exitcode: 999,
                message: error.message
            }));
            process.exit(999);
        }
    }
}))();
