"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const config = require("config");
const path = require("path");
const fs = require("fs");
const commandError_1 = require("./errors/commandError");
const child_process_1 = require("child_process");
const utils_1 = require("./utils");
class Environment {
    constructor(env_name) {
        this.pid = {
            agent: null,
            alive_monitor: null
        };
        this.env_name = env_name;
        this.userDir = this.getHomeDir();
        this.readAgentConfig();
        this.optimazeConfig();
        this.readPid();
        this.makeStorageDir();
    }
    /**
     * エージェント設定ファイルの読み込み
     */
    readAgentConfig() {
        try {
            const data = fs.readFileSync(`${config.agent.configPath}/${this.env_name}.json`);
            this.agent_config = JSON.parse(data.toString());
        }
        catch (error) {
            console.log(error.message);
            throw new commandError_1.CommandError('FAILED_READ_AGENT_CONFIG_FILE');
        }
    }
    /**
     * プロセスIDファイルの読み込み
     */
    readPid() {
        try {
            const data = fs.readFileSync(config.pid.fileName, {
                encoding: config.pid.encoding
            });
            this.pid = JSON.parse(data.toString());
        }
        catch (error) {
            // pidが存在しない場合もあるため例外を無視する
        }
    }
    /**
     * プロセスIDファイルの書き込み
     */
    setAgentPid(agent_pid) {
        try {
            this.pid.agent = agent_pid;
            utils_1.writeFile(config.pid.fileName, JSON.stringify(this.pid), config.pid.encoding);
        }
        catch (error) {
            console.log(error.message);
            throw new commandError_1.CommandError('FAILED_WRITE_PID_FILE');
        }
    }
    /**
     * エージェントプロセスが存在するかどうか
     */
    existAgentPid() {
        try {
            // pidからプロセス名を取得する
            const get_process_name_command = config.commands.getProcessNameByPid
                .replace('{pid}', this.pid.agent);
            const process_name = child_process_1.execSync(get_process_name_command);
            // agent起動コマンド
            const confirm_command_name = config.commands.agentBoot
                .replace('{user}', config.agent.owner.user)
                .replace('{env_name}', this.env_name);
            // pidで取得したプロセス名がエージェント起動コマンドと同じかどうか
            return process_name.toString().indexOf(confirm_command_name) != -1;
        }
        catch (error) {
            // エージェントプロセスが取得できない場合は、存在しないものとする
            return false;
        }
    }
    /**
     * プロセスIDファイルの書き込み
     */
    setAliveMonitorPid(alive_monitor_pid) {
        try {
            this.pid.alive_monitor = alive_monitor_pid;
            utils_1.writeFile(config.pid.fileName, JSON.stringify(this.pid), config.pid.encoding);
        }
        catch (error) {
            console.log(error.message);
            throw new commandError_1.CommandError('FAILED_WRITE_PID_FILE');
        }
    }
    /**
     * ホームディレクトリの取得
     */
    getHomeDir() {
        try {
            const command = config.commands.getHome.replace('{user}', config.agent.owner.user);
            return child_process_1.execSync(command).toString().replace(/\r?\n/g, '');
        }
        catch (error) {
            console.log(error.message);
            throw new commandError_1.CommandError('FAILED_GET_HOME_DIRECTORY');
        }
    }
    /**
     * 設定ファイルの最適化
     */
    optimazeConfig() {
        config.proxy.configXml = this.convertRelativePathToUserPath(config.proxy.configXml);
        config.pid.fileName = this.convertRelativePathToUserPath(config.pid.fileName);
        this.agent_config.storage.path = this.convertRelativePathToUserPath(this.agent_config.storage.path);
        this.agent_config.hardware.fileName = this.convertRelativePathToUserPath(this.agent_config.hardware.fileName);
        this.agent_config.profile.fileName = this.convertRelativePathToUserPath(this.agent_config.profile.fileName);
        this.agent_config.apiRelayClient.agent.fileName = this.convertRelativePathToUserPath(this.agent_config.apiRelayClient.agent.fileName);
        this.agent_config.apiRelayClient.status.fileName = this.convertRelativePathToUserPath(this.agent_config.apiRelayClient.status.fileName);
    }
    /**
     * 相対パスをユーザパスに変更
     */
    convertRelativePathToUserPath(targetPath) {
        if (targetPath.indexOf('/') !== 0) {
            return path.join(this.userDir, `.${this.agent_config.productId}`, targetPath);
        }
        return targetPath;
    }
    /**
     * ストレージディレクトリを作成
     */
    makeStorageDir() {
        utils_1.mkdirAndChownRecursive(this.agent_config.storage.path);
    }
}
exports.default = Environment;
